// Header file for HpObjects
/*!  \mainpage HPObjects
HPObjects is a library of C functions to access and create RPL objects
for the HP 49-series calculators using the HPGCC compiler.  As
distributed, the package works with HPGCC 2.0, but it can be
recompiled with HPGCC 1.1 to work on the 48gii calculator.
*/

///  \file HPObjects.h
#ifndef OBJECTS_H
#define OBJECTS_H

#include <stdarg.h>

/// A Saturn address
typedef unsigned int SatAddr;		/* saturn address */

#define bool int
#define false 0
#define true 1

#define SAT_DOXLIB  0x02e92	/* not sure why this isn't in saturn.h */
#define SAT_DOMINIFONT 0x026FE	/* Mini font */
#define SAT_DOFONT SAT_DOEXT2	/* Font prolog */

#define SAT_PRLG 0x28FC		/* common prolog */

// The definition for SAT_DOLNKARRY in saturn.h wrongly shows
// it as the same as SAT_DOARRY
#ifdef SAT_DOLNKARRY
#undef SAT_DOLNKARRY
#define SAT_DOLNKARRY 0x2A0A
#endif

/// \defgroup BINT Binary Integers (BINT)
/// A BINT is a 5 nibble unsigned binary number
/// @{
///////////////////////////////////////////////////////////////////////
// BINT routines
//

/// \brief is object a BINT?
/// \param src Saturn address of an object
/// \return true if the object at src is a BINT
bool isBINT(SatAddr src);

/// \brief Convert BINT to C integer.
/// \param src The BINT to decode
/// \param dst Where to store the BINT's value
/// \return true on success, false if src doesn't contain a BINT
bool BINTdecode(SatAddr src, int *dst);

/// \brief Convert a C integer to a BINT.
/// \param src The integer to convert. It must be between 0x0 and 0xFFFFF
/// \param dst Where to store the BINT. If zero then a new BINT is created.
/// \return The BINT's Saturn address, or 0 if i is out of range
SatAddr BINTencode(int src, SatAddr dst);

/// \brief Create a new BINT in tempOb. It's value is undefined.
SatAddr makeBINT();

/// \brief The size of a BINT in nibbles
/// \param src Saturn address of the BINT. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 10
#define BINTnibbles(src) (10)

/// @}

///////////////////////////////////////////////////////////////////////
/// \defgroup CHAR Characters (CHAR)
/// A CHAR is a single ASCII character
/// @{

/*! \brief is object a CHAR?
 * \param src Saturn address of an object
 * \return true if the object at src is a CHAR
 */
bool isCHAR(SatAddr src);

/// \brief Convert CHAR to C character.
/// \param src The CHAR to decode
/// \param dst Where to store the CHAR's value
/// \return true on success, false if src doesn't contain a CHAR
bool CHARdecode(SatAddr src, char *dst);

/// \brief Convert a C char to an CHAR object
/// \param src The ASCII character to convert.
/// \param dst Where to store the CHAR. If zero then a new CHAR is created.
/// \return The CHAR's Saturn address.
SatAddr CHARencode(unsigned char ch, SatAddr dst);

/// \brief Create a new CHAR in tempOb. It's value is undefined.
SatAddr makeCHAR();

/// \brief The size of a CHAR in nibbles
/// \param src Saturn address of the CHAR. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 7
#define CHARnibbles(src) (7)
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup REAL Real numbers (REAL)
/// A REAL is a 64 bit number with a 12 digit mantissa and an exponent
/// from -500 to 499
/// @{

/// \brief is object a real "approximate mode" number (REAL)?
/// \param src Saturn address of an object
/// \return true if the object at src is a REAL
bool isREAL(SatAddr src);

/// \brief Convert REAL to C double
/// \param src The REAL to decode
/// \param dst Where to store the REAL's value
/// \return true on success, false if src doesn't contain a REAL
///
/// Some HP REALs are too large or too small to be represented in a C
/// double and will result in a very large or very small number.
bool REALdecode(SatAddr src, double *dst);

/// \brief Convert a C double to an HP REAL
/// \param src The double to convert.
/// \param dst Where to store the REAL. If zero then a new REAL is created.
/// \return The REAL's Saturn address, or zero on error
SatAddr REALencode(double d, SatAddr dst);

/// \brief Create a new REAL in tempOb. It's value is undefined.
SatAddr makeREAL();

/// \brief The size of a REAL in nibbles
/// \param src Saturn address of the REAL. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 21
#define REALnibbles(src) (21)
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup EREL Extended Real numbers (EREL)
/// An EREL has 15 digits and an exponent from -50000 to 49999
/// @{

/*! \brief is object an extended real?
 * \param src Saturn address of an object
 * \return true if the object at src is an EREL
 */
bool isEREL(SatAddr src);

/// \brief Convert EREL to C double
/// \param src The EREL to decode
/// \param dst Where to store the EREL's value
/// \return true on success, false if src doesn't contain a EREL
///
/// Some HP ERELs are too large or too small to be represented in a C
/// double and will result in a very large or very small number.
bool ERELdecode(SatAddr src, double *dst);

/// \brief Convert a C double to an HP EREL
/// \param src The double to convert.
/// \param dst Where to store the EREL. If zero then a new EREL is created.
/// \return The EREL's Saturn address, or zero on error
SatAddr ERELencode(double d, SatAddr dst);

/// \brief Create a new EREL in tempOb. It's value is undefined.
SatAddr makeEREL();

/// \brief The size of a EREL in nibbles
/// \param src Saturn address of the EREL. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 26
#define ERELnibbles(src) (26)

// Convert a 15 nibble extended real number (i.e. the body of an EREL
// object ) to a double.  This is used by several different objects.
double ereal2double(SatAddr src);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup CMP Complex Numbers (CMP)
/// A complex number contains a real part and an imaginary part.
/// @{

/// \brief is object a Complex number (CMP)?
/// \param src Saturn address of an object
/// \return true if the object at src is a CMP
bool isCMP(SatAddr src);

/// \brief Decode CMP into it's real and imaginary components
/// \param src The CMP to decode
/// \param re If non-null, the real part of src will be placed here
/// \param im If non-null, the imaginary part of src will be placed here
/// \return true on success, false if src doesn't contain a CMP
bool CMPdecode(SatAddr src, double *re, double *im);

/// \brief Convert two C doubles into an HP CMP object
/// \param re The real part of the complex number
/// \param im The imaginary part of the complex number
/// \param dst Where to store the CMP.  If zero, then a new CMP is
/// created in tempOb.
/// \return The CMP's Saturn address, or zero on error
SatAddr CMPencode(double re, double im, SatAddr dst);

/// \brief Create a new CMP in tempOb. It's value is undefined.
SatAddr makeCMP();

/// \brief The size of a CMP in nibbles
/// \param src Saturn address of the CMP. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 37
#define CMPnibbles(src) (37)

/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup ZINT Infinite Precision Integers (ZINT)
/// @{

/// \brief is object an infinite precision integer (ZINT)?
/// \param src Saturn address of an object
/// \return true if the object at src is a ZINT
bool isZINT(SatAddr src);

/// \brief Convert a ZINT to C long long
/// \param src The ZINT to decode
/// \param dst Where to store the ZINT's value
/// \return true on success, false if src doesn't contain a ZINT or if
/// src won't fit in a long long
bool ZINTdecodell(SatAddr src, long long *dst);

/// \brief Convert a C long long to an HP ZINT
/// \param src The long long to convert.
/// \param dst If non-zero then dst must contain exactly enough room
/// for the ZINT. If zero then a new ZINT is created.
/// \return The ZINT's Saturn address, or zero on error
SatAddr ZINTencodell(long long src, SatAddr dst);

/// \brief Convert a ZINT to a NULL-terminated C string of ASCII digits.
/// \param src The ZINT to decode
/// \param dst Where to store the ZINT's value. If set, then dst must
/// contain enough space for the number, which you can get from
/// ZINTbytes(). IF dst is zero, then sufficient space is allocated
/// for the string and the caller must free the space.
/// \return the decoded string, or zero if src isn't a ZINT or another
/// error occurs.
char *ZINTdecodeStr(SatAddr src, char *str);


/// \brief Convert a NULL-terminated C string if ASCII digits to an HP ZINT
/// \param src The string to convert.  For example, "12345"
/// \param dst Where to store the ZINT. If non-zero then dst must
/// contain exactly the right amount of space for the resulting ZINT.
/// If zero, then the space will be allocated in TEMPOB.
/// \return The ZINT's Saturn address, or zero on error
SatAddr ZINTencodeStr(const char *src, SatAddr dst);

/// \brief Make a ZINT in TEMPOB with room for "digits" digits.
SatAddr makeZINT(int digits);

/// \brief The size of a ZINT in nibbles
/// \param src Saturn address of the ZINT.
/// \return The number of nibbles occupied by the ZINT
int ZINTnibbles(SatAddr src);

/// \brief Get the number of digits in a ZINT
/// \param src The Saturn address of the ZINT
/// \return The number of digits in the zint. Note that this returns 1
/// when the ZINT is zero, even though the ZINT doesn't store any
/// digits for zero.
int ZINTdigits(SatAddr src);

/// \brief Return the number of bytes needed to store a ZINT as a
/// string
/// \param src The Saturn address of the ZINT
/// \return The number of bytes needed, or -1 on error
int ZINTbytes(SatAddr src);


/// \private
// Clean up a string representation of an integer.  This skips +/-
// signs and leading zeros, and figures out how many digits are
// present.  "src" is the source string.  "digits" is an output param
// that is set to the number of digits in the string (which might be
// zero) the scan stops at the first non-digit character.  isPostive
// will be set to true/false depending on whether the number is
// positive or negative.  Returns a pointer to the first non-zero
// digit of the number.
//
// The following examples show the result of
// int digits;
// bool isPositive
// const char *result = cleanUpZINTstr(input, &digits, &isPositive);
//
// Examples:
// input               digits       isPositive        return value
// "1234"              4            1                 "1234"
// "+1234"             4            1                 "1234"
// "-1234"             4            1                 "1234"
// "---+1234ab"        4            0                 "1234"
// "-0"                0            1                 ""
const char *cleanupZINTstr(const char *src, int *digits, bool *isPositive);

/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup FLASHP Flash Pointer (FLASHP)
/// A Flash Pointer contains two binary numbers, a 12 bit "Pointer to
/// table A" and a 16 bit "pointer to table B."
/// @{

/// \brief is object a FLASHP?
/// \param src Saturn address of an object
/// \return true if the object at src is a FLASHP
bool isFLASHP(SatAddr src);

/// \brief Decode FLASHP into it's components
/// \param src The FLASHP to decode
/// \param a If non-null, the pointer to table A will be stored here
/// \param b If non-null, the pointer to table B will be stored here
/// \return true on success, false if src doesn't contain a FLASHP
bool FLASHPdecode(SatAddr src, int *a, int *b);

/// \brief Convert two C integers into an HP FLASHP object
/// \param a The pointer to table A. This must be between 0x0 and 0xfff
/// \param b The pointer to table B. This must be between 0x0 and 0xffff
/// \param dst Where to store the FLASHP.  If zero, then a new FLASHP is
/// created in tempOb.
/// \return The FLASHP's Saturn address, or zero if a or b is out of
/// range, or if it can't allocate a new FLASHP
SatAddr FLASHPencode(int a, int b, SatAddr dst);

/// \brief Create a new FLASHP in tempOb. It's value is undefined.
SatAddr makeFLASHP();

/// \brief The size of a FLASHP in nibbles
/// \param src Saturn address of the FLASHP. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 12
#define FLASHPnibbles(src) (12)
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup MINIFONT Mini font (MINIFONT)
/// A MINIFONT maps ASCII characters to 6-nibble graphical
/// representations. See chapter 56.16 in "Saturn Assembly Language
/// Programming" for details.
/// @{

/// \brief is object a Mini font (MINIFONT)?
/// \param src Saturn address of an object
/// \return true if the object at src is a MINIFONT
bool isMINIFONT(SatAddr src);

/// \brief Return the Saturn address of a character's MINIFONT entry.
///
/// This returns the address of the 6-nibble minifont entry corresponding to
/// ASCII char ch.  Once you have the entry, you can read or change it
/// with sat_peek() and sat_poke()
/// \param src The Saturn address of the MINIFONT
/// \param ch The ASCII character
/// \returns Saturn address of the 6-nibble minifont entry. If src
/// isn't a MINIFONT or if ch is not an ASCII character, then it
/// returns 0
SatAddr MINIFONTitem(SatAddr src, int ch);

/// \brief Get the minifont ID number
/// \param src The Saturn address of a minifont
/// \return The ID number for the minifont, or -1 if src isn't a MINIFONT
int MINIFONTgetId(SatAddr src);

/// \brief Set the MINIFONT id.
/// \param src The Saturn address of a minifont
/// \param id The new ID number. id must be between 0 and 0xff
/// \return true on success, or false if src isn't a MINIFONT or if id
/// is out of range.
bool MINIFONTsetId(SatAddr src, int id);

/// \brief create a new MINIFONT object in TEMPOB. The contents of the minifont
/// are undefined
/// \return the Saturn address of the new MINIFONT
SatAddr makeMINIFONT();

/// \brief The size of a MINIFONT in nibbles
/// \param src Saturn address of the MINIFONT. Although unnecessary,
/// the parameter is included for consistency
/// \return Always returns 1558
#define MINIFONTnibbles(src) (1548)
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup FONT FONT
/// @{

/// \brief is object a FONT?
/// \param src Saturn address of an object
/// \return true if the object at src is a FONT
bool isFONT(SatAddr src);

/// \brief The size of a FONT in nibbles
/// \param src Saturn address of the FONT.
/// \return The size of the FONT in nibbles.
int FONTnibbles(SatAddr src);

/// \brief Return the Saturn address of a character's FONT entry.
///
/// This returns the address of the font entry corresponding to
/// ASCII char ch.  Once you have the entry, you can read or change it
/// with sat_peek() and sat_poke(). See section 56.17 in "Saturn
/// Assembly Language Programming" for details of a FONT entry.
/// \param src The Saturn address of the FONT
/// \param ch The ASCII character
/// \return Saturn address of the font entry. If src
/// isn't a FONT or if ch is not an ASCII character, then it
/// returns 0
SatAddr FONTitem(SatAddr src, int idx);

/// \brief Get the font ID
/// \param src The Saturn Address of a FONT object
/// \return The FONT'd id number, or -1 if src isn't a FONT
int FONTgetId(SatAddr src);

/// \brief Set the FONT id.
/// \param src The Saturn Address of a FONT object
/// \param id The new ID, which must be between 0 and 255.
/// \return true on success, false if src isn't a FONT or if id is out
/// of range.
bool FONTsetId(SatAddr src, int id);

/// \brief Get the number of characters in the font's name
/// \param src The Saturn Address of a FONT object
/// \return The number of characters in the fon't name, or -1 if src
/// isn't a FONT
int FONTgetNameLen(SatAddr src);

/// \brief Change the name of a FONT. Note that new name must be the
/// same length as existing name
/// \param src The Saturn Address of a FONT object
/// \param name a null-terminated string that is the new name.  This
/// must be the same length as the old name. In other words
/// strlen(name) must equal FONTgetNameLen(src)
/// \return true on success, or false if src isn't a FONT or name
/// doesn't point to a null terminated string with the same length as
/// the existing FONT name.
bool FONTsetName(SatAddr src, const char *name);

/// \brief Decode a FONT into it's parts.
/// \param src The Saturn Address of a FONT object
/// \param id if non-null, then the FONT's id number will be stored
/// here.
/// \param name if non-null, then the font name will be copied here as a
/// null-terminated string.  Name must point to enough space to hold
/// the name. You can get the number of characters in
/// the name with FONTgetNameLen() ahead of time and pass this value
/// plus 1 (for the terminating NULL byte.  ALternatively, since the name
/// can never exceed 255 characters, just pass a 256 character buffer.
/// \param height If non-null, the FONT's height in pixels will be
/// stored here.
/// \param graphics If non-null, the FONT's graphic data will be
/// copied here. Graphics must point to sufficient space to hold the
/// data, which can be up to 2kb.
/// \return true on success, false if src isn't a FONT.
bool FONTdecode(SatAddr src, int *id, char *name, int *height, char *graphics);

/// \brief Create a new FONT.
/// \param id The FONT id, which must be between 0 and 255
/// \param name A null-terminated string containing the font's
/// name. Must be less than 255 characters.
/// \param height The font's height in pixels.
/// \\param graphics. The graphical data.  See "Saturn Assembly
/// Language Programming" for details on the format of FONT graphics.
/// \param dst Where to store the resulting FONT. If zero then a new
/// FONT will be created in TEMPOB.  Otherwise dst must point to the
/// right amount of space to hold the FONT object.
/// \returns The address of the new font, or 0 if the id, name length,
/// or height is out of range, or if dst is zero and it can't allocate
/// sufficient space for the FONT.
SatAddr FONTencode(int id, const char *name, int height, const char *graphics, SatAddr dst);

/// \brief Make a blank font with room for "nameLen" characters in the
/// name
/// \param nameLen The length of the name in characters
/// \return the address of the new font, or zero if nameLen is too
/// large or if no space can be allocated.
SatAddr makeFONT(int nameLen);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup STRING ASCII strings (STRING)
/// @{

/// \brief is the object at src a STRING?
/// \param src Saturn address of an object
/// \return true if the object at src is a STRING
bool isSTRING(SatAddr src);

/// \brief decode a STRING to a null-terminated C string.
/// \param src The Saturn Address of a STRING object
/// \param dst Where to store the resulting C string.  If dst is NULL
/// then a new string will be malloc'ed. Otherwise, dst must have
/// sufficient space for the string and it's null terminator. 
/// \return the C string, or NULL if src is not a STRING or an error occurs.
char *STRINGdecode(SatAddr src, char *dst);

/// \brief convert a null-terminated C string into an HP STRING object
/// \param src a null-terminated C string to convert
/// \param dst Where to store the resulting STRING. If zero then a new
/// STRING will be created in TEMPOB.  Otherwise dst must point to the
/// right amount of space to hold the STRING.
/// \return the Saturn address of the resulting STRING, or zero if an
/// error occurs.
SatAddr STRINGencode(const char *src, SatAddr dst);

/// brief append a null-terminated string to a STRING object
/// \param src a null-terminated C string to append
/// \param dst Saturn Address of an existing STRING object. There must
/// be sufficient space after the object to hold the appended string
/// \return true on success, false if dst isn't a STRING.
/// \warning Whatever space dst was allocated in must be completely
/// filled before your program returns or your will corrupt the memory
/// in your calculator!  This is because all objects in the TEMPOB
/// space must be contiguous with no empty space between them.  The
/// garbage collector relies on this property.
/// 
/// In retrospect, this function was a bad idea because it's so dangerous.
bool STRINGadd(const char *src, SatAddr dst);

/// \brief Allocate a STRING with numChars characters
/// The initial contents of the string are undefined.
/// \param size The number of characters that the string must hold.
/// \return the Saturn address of the new STRING, or zero if
/// sufficient space could not be allocated.
SatAddr makeSTRING(unsigned size);

/// \brief Return the size in nibbles of the STRING object at src.
/// \param src the Saturn address of a STRING object
/// \return the size of the STRING in nibbles, or zero if src is not a STRING.
int STRINGnibbles(SatAddr src);

/// \brief return the number of characters in a STRING
/// \param src the Saturn address of a STRING object
/// \return The number of characters in the STRING, or -1 if src is
/// not a STRING.
int STRINGnumChars(SatAddr src);
/// @}

//////////////////////////////////////////////////////////////////////
/// \defgroup LIST List of objects (LIST)
/// @{

/// \brief is the object at src a LIST?
/// \param src Saturn address of an object
/// \return true if the object at src is a LIST
bool isLIST(SatAddr src);

/*!
  \brief Allocate space for a list with nibbles data.

  This allocates space for a list with nibbles data in it and puts an
  empty list at the beginning of the allocated space.

  \param nibbles the number of nibbles to allocate
  \return Saturn address of the empty list

  \warning Because of the way space is allocated on the calculator,
  you must fill the list up so its size is exactly the number of
  nibbles allocated.  For this reason, you should use this function
  only when absolutely necessary. Use LISTencode() or LISTencodeN()
  to create lists instead.
*/
SatAddr makeLIST(unsigned nibbles);


/// \brief Return the size in nibbles of the LIST at src.
/// \param src the Saturn address of a LIST object
/// \return the size of the LIST in nibbles, or zero if src is not a LIST
int LISTnibbles(SatAddr list);

/// \brief Returns the number of elements in a LIST
/// \param src the Saturn address of a LIST object
/// \return the number of elements in the LIST, or -1 if src is not a LIST
int LISTsize(SatAddr src);

/// \brief Make a copy of obj and append it to the list at dst
/// \param dst The LIST to append to
/// \param obj The object to append
/// \return true on success, false if dst isn't a list or obj isn't an
/// object.
/// \warning See the warning on makeLIST()
bool LISTadd(SatAddr dst, SatAddr obj);

/// \brief Concatenate an existing list and an object to create a new list
///
/// LISTcat creates a new list from an existing list src and an object
/// obj. The original list and object remain unchanged. This is
/// equivalent to the RPL "+" opertor applied to a list and an object.
/// \param src the existing list
/// \param dst the existing object
/// \return The Saturn address of a new list formed by adding obj to
/// src
SatAddr LISTcat(SatAddr src, SatAddr obj);

/// \brief Merge two lists to create a third. Equivalent to RPL "+"
/// operator aplied to two lists.
/// \param list1 the list that will apear first
/// \param list2 the list that will appear second
/// \return a new list that combines list1 and list 2. Returns 0 if
/// list1 or list2 isn't a LIST object, or if space for a new list
/// cannot be allocated.
/// For example if list1 is {1 2 3} and list2 is {4 5 6} then it returns
/// {1 2 3 4 5 6}
SatAddr LISTmerge(SatAddr list1, SatAddr list2);

/// \brief create a list containing the specified objects
/// \param dst The Saturn address where the list should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// list.  Otherwise dst must point to sufficient space for the list.
/// See the warning under makeLIST
/// \param ob1, ob2, ... The Saturn addresses of the objects to put in
/// the list.  If the parameter is an indirect pointer (i.e. the
/// Saturn address of a pointer to an object), then the pointer is put
/// in the list, not the object.  If the parameter points to an object
/// in ROM (addresses 0-0x40000) then the pointer to the object is
/// added to the list, not the object itself.
/// \param 0 The parameter list must end with 0! Otherwise the
/// function won't know when to stop adding objects.
/// \return the Saturn address of the new list.  Returns 0 on error
/// (can't allocate space, a parameter isn't an object or a pointer to
/// an object, etc.
SatAddr LISTencode(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ );

/// \brief create a list from an array of Saturn Addresses. This is
/// like the RPL ->LIST command
/// \param n the number of items to put in the list
/// \param obs an array of n Saturn Addresses (SatAddrs). The list
/// will contain these N objects.  If an entry in the array is an
/// address between 0 and 0x40000, then it points to ROM and the
/// address is copied to the list instead of the object itself.
/// Otherwise if
/// the array entry is the address of an object then the object is
/// copied to the list, othewise if the array entry is the address of
/// a pointer to an object, then the pointer is copied to the list.
/// \param dst The Saturn address where the list should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// list.  Otherwise dst must point to sufficient space for the list.
/// See the warning under makeLIST
SatAddr LISTencodeN(int n, SatAddr obs[], SatAddr dst);

//
// Iteration
//

/*!
\brief Get the first object in a list and update the list
iterator

Use LISTfirstOb() and LISTnextOb() to iterate through the objects of a
list. Note that lists can contain objects or pointers to objects
(indirect objects). These functions always return the Saturn address
of the object itself.

For example:
\code
SatAddr iter, obj, myList;
...
for (obj = LISTfirstOb(myList, &iter); obj; obj = LISTnextOb(&iter)) {
    // do something with obj
}
\endcode

\param list the list to iterate through
\param iter pointer to a SatAddr variable. The variable will be used to
store the state of the iteration.
\return the Saturn address of the first object in the list, or zero
if the list is empty.
\sa LISTnextOb(), LISTfirstEntry(), LISTnextEntry() */
SatAddr LISTfirstOb(SatAddr list, SatAddr *iter);

/*!
\brief return the next object in a list and update the list iterator.
\param iter pointer to a SatAddr variable that was previously
passed to LISTfirstOb() or LISTnextOb().
\return the Saturn address of the next object in the list, or
zero if there are no more entries in the list.
\sa LISTfirstOb(), LISTfirstEntry(), LISTnextEntry() */
SatAddr LISTnextOb(SatAddr *iter);

/*!
\brief Get the first entry in a list

Use LISTfirstEntry() and LISTnextEntry() to iterate through the contents of a
list. Note that lists can contain objects or pointers to objects
(indirect objects). These functions always return the Saturn address
of what is actually stored in the list. In other words, the address
returned may point to an object or a pointer.

For example:
\code
SatAddr addr, myList;
...
for (addr = LISTfirstEntry(myList); addr; addr = LISTnextEntry(addr)) {
    // do something with addr
}
\endcode

\param list the list to iterate through
\return the Saturn address of the first entry in the list, or zero
if the list is empty.
\sa LISTnextEntry(), LISTfirstOb(), LISTnextOb(), directPtr(), isOb() */
SatAddr LISTfirstEntry(SatAddr list);

/*!
\brief return the next entry in a list or zero if the end has been reached
\param entry the value returned by a previous call to LISTfirstEntry()
or LISTnextEntry()
\return the Saturn address of the next entry in the list, or
zero if there are no more entries in the list.
\sa LISTfirstEntry(), LISTfirstOb(), LISTnextOb(), directPtr(), isOb() */
SatAddr LISTnextEntry(SatAddr entry);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup IDNT Global Identifiers (IDNT)
/// @{

/// \brief is the object at src an IDNT?
/// \param src Saturn address of an object
/// \return true if the object at src is an IDNT
bool isIDNT(SatAddr src);

/// \brief Convert an IDNT to a null-terminated string
/// \param src The Saturn address of an IDNT
/// \param dst A pointer to a buffer with sufficient space to hold the
/// name, or zero.  If the value is zero then space will be malloc'ed
/// to hold the name.
/// \return The null-terminated string.
/// \sa IDNTnumChars()
char * IDNTdecode(SatAddr src, char *dst);

/// \brief Convert a null-terminated C string into an IDNT object
/// \param src The null-terminated string. It must be less than 256
/// characters in length.
/// \param dst The Saturn address where the IDNT should be stored. If
/// dst is zero then space is allocated for the
/// IDNT  Otherwise dst must point to sufficient space for the IDNT
/// \return The Saturn address of the IDNT, or zero on error.  Errors
/// occur if the IDNT can't be allocated, or if src points to a string
/// that is longer than 255 characters.
SatAddr IDNTencode(const char *src, SatAddr dst);

/// \brief Allocate space for an IDNT with size characters
/// \param size The number of characters in the IDNT
/// \return The Saturn address of a new empty IDNT, or zero if the IDNT
/// can't be allocated or size is larger than 255.
SatAddr makeIDNT(unsigned size);


/// \brief Return the size in nibbles of the IDNT at src.
/// \param src the Saturn address of an IDNT object
/// \return the size of the IDNT in nibbles, or zero if src is not an IDNT
int IDNTnibbles(SatAddr src);

/// \brief return the number of characters in an IDNT
/// \param src The Saturn address of an IDNT
/// \return the number of characters in the IDNT, or -1 if src isn't
/// an IDNT.
int IDNTnumChars(SatAddr src);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup LAM Local Identifiers (LAM)
/// @{
/// \brief is the object at src a LAM?
/// \param src Saturn address of an object
/// \return true if the object at src is a LAM
bool isLAM(SatAddr src);

/// \brief Convert a LAM to a null-terminated string
/// \param src The Saturn address of a LAM
/// \param dst A pointer to a buffer with sufficient space to hold the
/// name, or zero.  If the value is zero then space will be malloc'ed
/// to hold the name.
/// \return The null-terminated string.
/// \sa LAMnumChars()
char * LAMdecode(SatAddr src, char *dst);

/// \brief Convert a null-terminated C string into a LAM object
/// \param src The null-terminated string. It must be less than 256
/// characters in length.
/// \param dst The Saturn address where the LAM should be stored. If
/// dst is zero then space is allocated for the
/// LAM  Otherwise dst must point to sufficient space for the LAM
/// \return The Saturn address of the LAM, or zero on error.  Errors
/// occur if the LAM can't be allocated, or if src points to a string
/// that is longer than 255 characters.
SatAddr LAMencode(const char *src, SatAddr dst);

/// \brief Allocate space for a LAM with size characters
/// \param size The number of characters in the LAM
/// \return The Saturn address of a new empty LAM, or zero if the LAM
/// can't be allocated or size is larger than 255.
SatAddr makeLAM(unsigned size);


/// \brief Return the size in nibbles of the LAM at src.
/// \param src the Saturn address of a LAM object
/// \return the size of the LAM in nibbles, or zero if src is not a LAM
int LAMnibbles(SatAddr src);

/// \brief return the number of characters in a LAM
/// \param src The Saturn address of a LAM
/// \return the number of characters in the LAM, or -1 if src isn't
/// a LAM.
int LAMnumChars(SatAddr src);

/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup HXS Hex String (HXS)
///
/// A Hex String is simply a block of data.  These can be used to
/// store binary numbers or unspecified data. Functions are provided
/// to convert to/from long long integers or buffers of any data.
/// @{

/// \brief is the object at src a HXS
/// \param src Saturn address of an object
/// \return true if the object at src is a HXS
bool isHXS(SatAddr src);	/* is it an HXS? */

/// \brief decode a HXS into an unsigned long long.
/// \param src The Saturn address of a HXS
/// \param dst pointer to where the result will be stored.
/// \return true on success, false if src isn't a HXS or if the HXS
/// won't fit in an unsigned long long, or if dst is NULL.
/// \sa HXSencodell()
bool HXSdecodell(SatAddr src, unsigned long long *dst);

/// \brief Convert an unsigned long long to a HXS object
/// \param i the unsigned long long to convert
/// \param dst Saturn address where the HXS will be stored, or zero
/// to allocate a new one.
/// \return The address of the HXS, or zero on error
SatAddr HXSencodell(unsigned long long i, SatAddr  dst);


/// \brief Extract the data and/or size of an HXS.
/// \param src The Saturn address of a HXS
/// \param buf A charcter buffer to receive the data, or zero if
/// you don't want to copy the data
/// \param nibbles If non zero, the number of data nibbles in the HXS
/// will be copied here.
/// \return true on success, false if src doesn't point to an HXS.
bool HXSdecode(SatAddr src, char *buf, int *nibbles);

/// \brief Create an HXS object from a data buffer
/// \param buf pointer to the data to put in the HXS
/// \param nibbles number of nibbles in buf to copy.
/// \param dst The Saturn address of the buffer to recieve the HXS, or
/// zero to create a new buffer.
/// \return Saturn address of the resulting HXS, or zero on error.
SatAddr HXSencode(char *buf, int nibbles, SatAddr dst);


/// \brief Return the size in nibbles of the HXS object at src
/// \param src Saturn address of an HXS
/// \return The size in nibbles of the HXS object, or -1 if src
/// doesn't point to an HXS.
int HXSnibbles(SatAddr src);

/// \brief Return the number of bytes required to store the data in an
/// HXS
/// \param src Saturn address of an HXS
/// \return The number of bytes, or -1 on error
int HXSbytes(SatAddr src);


/// \brief Create an HXS
/// \param nibbles the number of data nibbles in the HXS
/// \return the Saturn address of the new HSX, or 0 on error
SatAddr makeHXS(int nibbles);

/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup GROB Graphical Object (GROB)
/// @{

/// \brief is the object at src a GROB
/// \param src Saturn address of an object
/// \return true if the object at src is a GROB
bool isGROB(SatAddr src);


/// \brief Return the size in nibbles of the GROB object at src
/// \param src Saturn address of an GROB
/// \return The size in nibbles of the GROB object, or -1 if src
/// doesn't point to an GROB.
int GROBnibbles(SatAddr src);

/// \brief Extract the information about a GROB
/// \param src Saturn address of a GROB
/// \param rows If non-null, then number of rows in the GROB will be
/// stored here.
/// \param columns If non-null, then number of columns in the GROB
/// will be stored here.
/// \param pixels If non-null, then GROB data will be stored here.
/// The buffer must contain enough space for the data. You can get the
/// space required with GROBbytes()
/// \return true on success, false if src doesn't point to a GROB.
bool GROBdecode(SatAddr src, int *rows, int *columns, char *pixels);

/// \brief Create a GROB from the parts
/// \param rows The number of rows in the GROB
/// \param columns The number of columns in the GROB
/// \param pixels. The pixel data for the grob, or NULL to leave the
/// data unspecified
/// \param dst The Saturn address of the buffer to recieve the GROB, or
/// zero to create a new buffer.
/// \return the Saturn address of the grob
SatAddr GROBencode(int rows, int columns, char *pixels, SatAddr dst);

/// brief Return the number of bytes required to store the graphical
/// data in a GROB
/// \param src The Saturn address of the GROB
/// \return teh number of bytes, or -1 if src doesn't point to a GROB
int GROBbytes(SatAddr src);

/// \brief Create a blank GROB in TEMPOB
/// \param rows The number of rows
/// \param columns The number of columns
/// \return the Saturnd address of the new GROB, or 0 on error
SatAddr makeGROB(int rows, int columns);

/// \brief Set or clear a pixel in a GROB
/// \param src The Saturn address of the GROB
/// \param row The row of the pixel.  Row numbers start at 0.
/// \param column The column of the pixel. Column numbers start at 0.
/// \param val The value to set the pixel to.  zero makes the pixel
/// white, non-zero makes it black.
/// \return true on success, false if src isn't a GROB, or if row or
/// column is out of range.
bool GROBsetPixel(SatAddr src, int row, int column, bool val);
/// @}



/////////////////////////////////////////////////////////////////////
/// \defgroup XLIB External Library Command (XLIB)
/// An XLIB specifies a command in a library. It contains a 3 nibble
/// library number and a 3 nibble command number.
/// @{

/// \brief is the object at src an XLIB?
/// \param src Saturn address of an object
/// \return true if the object at src is an XLIB
bool isXLIB(SatAddr src);

/// \brief extract an XLIB's library and command numbers.
/// \param src Saturn address of an XLIB
/// \param lib if non-null, the library number will be stored here.
/// \param cmd if non-null, the command number will be stored here.
/// \return true on success, false if src doesn't point to an XLIB
bool XLIBdecode(SatAddr src, int *lib, int *cmd);


/// \brief Encode an XLIB object with the given library and command
/// numbers
/// \param lib The library number. Must be between 0 and 0xfff
/// \param cmd The command number. Must be between 0 and 0xfff
/// \param dst The Saturn address where the XLIB should be stored, or
/// zero to create a new XLIB
/// \return the Saturn address of the XLIB, or zero if lib or cmd are
/// out of range, or if a new XLIB cannot be allocated.
SatAddr XLIBencode(int lib, int cmd, SatAddr dst);


/// \brief Create an empty XLIB in tempOb. Contents are undefined.
SatAddr makeXLIB();


/// \brief Return the number of nibbles in an XLIB
/// \param src Saturn address of an XLIB. This parameter is unused but
/// is included for consistency with the other XYZnibbles() functions.
/// \return Always returns 11
#define XLIBnibbles(src) (5+3+3)
/// @}



/////////////////////////////////////////////////////////////////////
/// \defgroup CODE Saturn Assembly Code (CODE)
/// @{

/// \brief is the object at src a CODE object?
/// \param src Saturn address of an object
/// \return true if the object at src is a CODE object
bool isCODE(SatAddr src);

/// \brief extract the code
/// \param src Saturn address of a CODE object
/// \param dst If non-null, then the contents of the code object will
/// be copied here. Dst must point to sufficient space to store the
/// code. See CODEbytes().
/// \param nibbles If non-null, then the number of nibbles of CODE
/// data is stored here.
/// \return true on success. False if src isn't a CODE object
/// \sa CODEbytes()
bool CODEdecode(SatAddr src, char *dst, int *nibbles);


/// \brief construct a CODE object
/// \param src The data to be stored in the CODE object
/// \param nibbles The number of nibbles in src to store.
/// \param dst If non-null, then this is the Saturn address where the CODE
/// object should be stored.  If null, the a new CODE object is
/// allocated in tempOb.
/// \return The Saturn address of the new CODE object, or zero if src
/// is NULL or a new CODE object can't be allocated.
SatAddr CODEencode(const char *src, const int nibbles, SatAddr dst);


/// \brief Allocate a new CODE object
/// \param nibbles The number of nibbles of code data in the object
/// \return The address of a new CODE object. The contents are
/// undefined.
SatAddr makeCODE(unsigned nibbles);

/// \brief Return the number of nibbles in a CODE object
/// \param src Saturn address of a CODE object.
/// \return The number of nibbles, or -1 if src doesn't point to a
/// CODE object.
int CODEnibbles(SatAddr src);


/// brief Return the number of bytes required to store the
/// data in a CODE object
/// \param src The Saturn address of a CODE object
/// \return the number of bytes, or -1 if src doesn't point to a CODE object
int CODEbytes(SatAddr src);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup LIBDATA Library Data (LIBDATA, also known as EXT0)
/// @{
/// Library data is free-formed data, so there isn't much to do with it.


/// \brief LIBDATA uses the EXT0 prolog.
#define SAT_DOLIBDATA SAT_DOEXT0

/// \brief is the object at src a LIBDATA object?
/// \param src Saturn address of an object
/// \return true if the object at src is a LIBDATA object
bool isLIBDATA(SatAddr src);

/// \brief Return the number of nibbles in a LIBDATA object
/// \param src Saturn address of a LIBDATA object.
/// \return The number of nibbles, or -1 if src doesn't point to a
/// LIBDATA object.
int LIBDATAnibbles(SatAddr src);

/// brief Return the number of bytes required to store the
/// data in a LIBDATA object
/// \param src The Saturn address of a LIBDATA object
/// \return the number of bytes, or -1 if src doesn't point to a LIBDATA object
int LIBDATAbytes(SatAddr src);

/// \brief extract the code
/// \param src Saturn address of a CODE object
/// \param dst If non-null, then the contents of the code object will
/// be copied here. Dst must point to sufficient space to store the
/// code. See CODEbytes().
/// \param nibbles If non-null, then the number of nibbles of CODE
/// data is stored here.
/// \return true on success. False if src isn't a CODE object
/// \sa CODEbytes()

/// \brief Extract a LIBDATA's data and its size.
/// \param src Saturn address of a LIBDATA object
/// \param data If non-null, then the contents of the LIBDATA object will
/// be copied here. Data must point to sufficient space to store the
/// data. See LIBDATAbytes().
/// \param nibbles If non-null, then the number of nibbles of LIBDATA
/// data is stored here.
/// \return true on success. False if src isn't a LIBDATA object
/// \sa LIBDATAbytes()
bool LIBDATAdecode(SatAddr src, char *data, int *nibbles);

/// \brief Construct a LIBDATA object
/// \param data The data to be stored in the LIBDATA object
/// \param nibbles The number of nibbles in data to store.
/// \param dst If non-null, then this is the Saturn address where the LIBDATA
/// object will be stored.  If null, then a new LIBDATA object is
/// allocated in tempOb.
/// \return The Saturn address of the new LIBDATA object, or zero if data
/// is NULL or a new LIBDATA object can't be allocated.
SatAddr LIBDATAencode(const char *data, int nibbles, SatAddr dst);

/// \brief Allocate a new LIBDATA object
/// \param nibbles The number of nibbles of LIBDATA data in the object
/// \return The address of a new LIBDATA object. The contents are
/// undefined.
SatAddr makeLIBDATA(unsigned nibbles);
/// @}


/////////////////////////////////////////////////////////////////////
// \defgroup ACPTR Extended Pointer (ACPTR)
/// @{
/// I believe that Exteneded Pointers are used to access "covered"
/// memory. Memory is covered when the MMU maps another page onto the
/// original memory's address.  See section 56.13 in "Introduction to
/// Saturn Assembly Language" for details.
///
/// An Extended Pointer contains a 5-nibble object address and a
/// 5-nibble program address.

/// \brief is the object at src an ACPTR object?
/// \param src Saturn address of an object
/// \return true if the object at src is an ACPTR object.
bool isACPTR(SatAddr src);

/// \brief extract an ACPTR's object and program numbers.
/// \param src Saturn address of an ACPTR
/// \param obj if non-null, the Saturn address of the object will be stored here.
/// \param program if non-null, the Saturn address of the program will be stored here.
/// \return true on success, false if src doesn't point to an ACPTR
bool ACPTRdecode(SatAddr src, SatAddr *obj, SatAddr *program);

/// \brief Encode an ACPTR object with the given object and program numbers
/// \param obj Saturn address of the object
/// \param program Saturn address of the program
/// \param dst The Saturn address where the ACPTR should be stored, or
/// zero to create a new ACPTR
/// \return the Saturn address of the ACPTR, or zero if obj or program
/// are out of range, or if a new ACPTR cannot be allocated.
SatAddr ACPTRencode(SatAddr obj, SatAddr program, SatAddr dst);

/// \brief Create an empty ACPTR. Contents are undefined.
/// \return the Saturn address of the new ACPTR
SatAddr makeACPTR();

/// \brief Return the number of nibbles in a ACPTR object
/// \param src Saturn address of a ACPTR object. Although not
/// necessary, this parameter is included for consistency with the
/// other objects.
/// \return Always returns 15
#define ACPTRnibbles(src) (15)
/// @}


/////////////////////////////////////////////////////////////////////
/*!
\defgroup ARRAY Array of 1 type of object (ARRY)

Array's store many objects of a single type by storing one copy of
the prolog and then storing each object's body without its prolog.  Thus
to access an object within the array, we need to copy the prolog
and body somewhere else.  Functions are provided to do this.

For lower level access, functions are also provided to access the
object bodies directly.

\warning Note that in the C routines, array indexes start at zero.
Also, note that dimension numbers start at zero.
*/
/// @{

/// \brief is the object at src an ARRAY object?
/// \param src Saturn address of an object
/// \return true if the object at src is an ARRAY object.
bool isARRAY(SatAddr src);

/// \brief Return the number of nibbles in a ARRAY object
/// \param src Saturn address of a ARRAY object.
/// \return The size of the ARRAY
int ARRAYnibbles(SatAddr src);

/// \brief Create an N-dimensional array
///
/// Creates an N-dimensional array and initializes each element with a
/// copy of the "element" parameter
/// \param element The object to initialize the array with.  Element's
/// type defines the type of all items in the array.
/// \param numDimensions The number of dimensions in the array.
/// \param dim1 The size of the first dimension.  The size of each
/// subsequent dimension follows.
/// \return Saturn address of the new array.
///
/**
Example: make a 4x5 array of real numbers. Initialize each value with 1.0
\code
SatAddr myArray = makeARRAY(REALencode(1.0, 0), 2, 4, 5);
\endcode

Example: make a 1-dimensional array of 10 lists where each list is {
1. 2. 3. }
\code
SatAddr theList = LISTencode(0, REALencode(1., 0), REALencode(2., 0), REALencode(3.,0), 0));
SatAddr myArray = makeARRAY(theList, 1, 10);
\endcode
*/
SatAddr makeARRAY(SatAddr element,
		  int numDimensions,
		  int dim1,
		  ... /* dim2,dim3 etc */ );

/// \brief Get a copy of an item in an array
/// \param array The ARRAY object
/// \param dst If non-null, then the array item, including its prolog will be copied here.
/// Otherwise space will be allocated in TEMPOb for the item and the
/// it will be copied there.
/// \param idx1 Index value of the array's first dimension.  NOTE!
/// In keeping with C conventions, the index values start at zero, not
/// 1!!
/// \param idxN Index values for the rest of the dimensions. You must
/// pass as many indices as there are dimensions.
/// \return The Saturn address of the item, or zero on error
///
/// To avoid filling up TempOb, you should pass a non-null value to
/// dst if at all possible.
/// 
/// Because arrays store object bodies separate from object prologs,
/// we must copy the object and prolog to a separate place to have
/// meaningful access to it.
///
/// This function's speed depends on greatly on whether the array
/// contains fixed-sized objects.  If the objects are all the same
/// size (like REAL numbers), then it uses arithemetic to compute the
/// address of the appropriate item.  Otherwise the funtion is forced
/// to scan through the array to find the right one.
SatAddr ARRAYgetItem(SatAddr array, SatAddr dst, int idx1,
			... /* idx2, idx3 etc */ );


/// \brief Put an item into an array.
///
/// This puts an item into an array. The new item must be the same
/// size as the existing item, and the same type as the array contents.

/// \param array The ARRAY object
/// \param item The item to copy into the array.
/// \param idx1 Index value of the array's first dimension.  NOTE!
/// In keeping with C conventions, the index values start at zero, not
/// 1!!
/// \param idxN Index values for the rest of the dimensions. You must
/// pass as many indices as there are dimensions.
/// \return Returns 0 on success. Returns -1 if "item" is not the same
/// type as the objects in the array. Returns -2 if "item" is not the
/// size as the item that it is replacing. Returns -3 if "array" isn't
/// an ARRAY object. Returns -10-n if the n-th index is out of range.
int ARRAYputItem(SatAddr array,
		 SatAddr item,
		 int idx1,
		 ... /* idx2, idx3, ... */ );

//
// Iteration
//

/// \brief Get the first item in an array.
/// \param array The ARRAY object
/// \param dst If non-null, then the array item, including its
/// prolog will be copied here. Otherwise space will be allocated in TEMPOb for the item and
/// it will be copied there.
/// \param iter A pointer to a SatAddr variable. This will be set to a
/// value that can be passed to ARRAYnextItem() to get the next item.
/// \return The Saturn address of the item, or zero on error
SatAddr ARRAYfirstItem(SatAddr array, SatAddr dst, SatAddr *iter);


/// \brief Return the next item in an array.
/// \param array The ARRAY object
/// \param dst If non-null, then the array item's prolog and body
/// will be copied here. Otherwise space will be allocated in TEMPOb
/// for the item and it will be copied there.
/// \param iter A pointer to a SatAddr variable that was previously
/// passed to ARRAYfirstItem() or ARRAYnextItem(). This will be
/// updated to a value that can be passed to ARRAYnextItem() again to
/// get the following item.
/// \return The Saturn address of the next item in the array, or zero
/// if no more items are available, or if an error occurs.  In
/// addition, *iter is updated.
SatAddr ARRAYnextItem(SatAddr array, SatAddr dst, SatAddr *iter);


//
// Utility functions
//

/// \brief Return the number of dimensions in an array
/// \param array The ARRAY object
/// \return The number of dimensions, or -1 on error.
int ARRAYnumDimensions(SatAddr array);

/// \brief Get the size of the i'th dimension.
///
/// Dimensions are numbered from zero, so valid dimensions
/// are 0, 1, 2, 3, etc).
/// \param array The ARRAY object
/// \param dimensionNumber The dimension number (0, 1, 2, etc)
/// \return The size of the dimension, or -1 on error
///
/// For example, if arr is a 2x4 array, then ARRAYgetDimension(arr,0)
/// return 2 and ARRAYgetDimension(arr,1) return 4.
int ARRAYgetDimension(SatAddr array, int dimensionNumber);

/// \brief get the prolog value of the items stored in the array.
/// \param array The ARRAY object
/// \return the Prolog value or -1 on error.
int ARRAYgetElProlog(SatAddr array);


//
// Low level functions.
//

/// \brief Get the Saturn address of an element body in an array.
///
/// Arrays store object bodies, without their prologs. This function
/// gets the address of one such body.
/// \param array The ARRAY object
/// \param elementBody Address of a SatAddr variable. The Saturn
/// address of the element's body will be stored here.
/// \param idx1 Index value of the array's first dimension.  NOTE!
/// In keeping with C conventions, the index values start at zero, not
/// 1!!
/// \param idxN Index value of the n'th index. You must
/// pass as many indices as there are dimensions.
/// \return Returns zero on success. Returns -3 if "array" isn't an
/// ARRAY. Returns -10-n if the n'th index value is out of range.
/// \sa ARRAYgetElement() to get an actual HP Object constructed from
/// the body and prolog.
int ARRAYgetElementBody(SatAddr array,
			SatAddr *elementBody,
			int idx1,
			... /* idx2, idx3, etc */ );

/// \brief Get the size of an element body in an array.
///
/// Given an array and a pointer to an element body in an array, return the
/// size of the element in the array.  Since the array contains the body
/// of an object, the size needed to reconstitute this object is
/// 5+ARRAYelementSize().
/// \param array The ARRAY object
/// \param element The element body 
/// \return The element's size, or -1 on error.
/// \sa ARRAYgetElementBody(), ARRAYfirstElement(), ARRAYnextElement()
int ARRAYelementSize(SatAddr array, SatAddr element);

/// \brief Get the first element body in an array.
/// \param array The ARRAY object
/// \return The Saturn address of the first element body in the array,
/// or zero on error or if no elements are in the array.
/// \warning Note that there is no way to distinguish between an empty
/// array and an error.  If in doubt, use ISARRAY() to make sure it is
/// an ARRAY object first.
SatAddr ARRAYfirstElement(SatAddr array);

/// \brief Return the next element body in an array.
/// \param array The ARRAY object
/// \param prev The Saturn address of an existing item in the array.

/// \return The Saturn address of the next element body following
/// "prev" in the array, or zero if no more elements are present, or if
/// an error occurs.
SatAddr ARRAYnextElement(SatAddr array, SatAddr prev);
/// @}

/////////////////////////////////////////////////////////////////////
//
/// \defgroup LNKARRY Linked Array (LNKARRY)
/// @{
///  Linked arrays are faster than
/// regular arrays because they have a table of pointers to each item.
///
/// The API is identical to the ARRAY API, except that the function
/// names begin with LNKARRY instead of ARRAY.  Please see the ARRAY
/// object documentation for details.
/// \sa Array of 1 type of object (ARRY)

/// \brief Same as isARRAY() but works on a LNKARRY object
bool isLNKARRY(SatAddr src);

/// \brief Same as ARRAYnibbles() but works on a LNKARRY object
int LNKARRYnibbles(SatAddr src);

/// \brief Same as makeARRAY() but works on a LNKARRY object
SatAddr makeLNKARRY(SatAddr element,
		  int numDimensions,
		  int dim1,
		  ... /* dim2,dim3 etc */ );

/// \brief Same as ARRAYgetItem() but works on a LNKARRY object
SatAddr LNKARRYgetItem(SatAddr lnkarry, SatAddr dst, int idx1,
			... /* idx2, idx3 etc */ );

/// \brief Same as ARRAYputItem() but works on a LNKARRY object
int LNKARRYputItem(SatAddr lnkarry,
		 SatAddr item,
		 int idx1,
		 ... /* idx2, idx3, ... */ );

/// \brief Same as ARRAYfirstItem() but works on a LNKARRY object
SatAddr LNKARRYfirstItem(SatAddr lnkarry, SatAddr dst, SatAddr *iter);

/// \brief Same as ARRAYnextItem() but works on a LNKARRY object
SatAddr LNKARRYnextItem(SatAddr lnkarry, SatAddr dst, SatAddr *iter);

/// \brief Same as ARRAYnumDimensions() but works on a LNKARRY object
int LNKARRYnumDimensions(SatAddr lnkarry);

/// \brief Same as ARRAYgetDimension() but works on a LNKARRY object
int LNKARRYgetDimension(SatAddr lnkarry, int dimensionNumber);

/// \brief Same as ARRAYgetElProlog() but works on a LNKARRY object
int LNKARRYgetElProlog(SatAddr lnkarry); /* get prolog of elements in lnkarry */

/// \brief Same as ARRAYelementSize() but works on a LNKARRY object
int LNKARRYelementSize(SatAddr lnkarry, SatAddr item);

/// \brief Same as ARRAYgetElementBody() but works on a LNKARRY object
int LNKARRYgetElementBody(SatAddr lnkarry,
			SatAddr *elementBody,
			int idx1,
			... /* idx2, idx3, etc */ );

/// \brief Same as ARRAYfirstElement() but works on a LNKARRY object
SatAddr LNKARRYfirstElement(SatAddr lnkarry);

/// \brief Same as ARRAYnextElement() but works on a LNKARRY object
SatAddr LNKARRYnextElement(SatAddr lnkarry, SatAddr item);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup COL RPL program (AKA secondary) (COL)
/// @{
/// An RPL program is probably called a COL because its source code begins with
/// "::".  Except for its prolog, an RPL program is structured
/// identically to a list.

/// \brief is the object at src a COL?
/// \param src Saturn address of an object
/// \return true if the object at src is a COL
bool isCOL(SatAddr src);

/// \brief Extract the code from a COL
/// \param src Saturn address of a COL object
/// \param dst If non-null, then the contents of the COL object will
/// be copied here (everything after the prolog and before the
/// terminating SEMI. Dst must point to sufficient space to store the
/// code. See COLbytes().
/// \param nibbles If non-null then the number of nibbles copied will
/// be stored here.
/// \sa COLbytes()
/// \return true on success, 0 if src isn't a COL or an error occurs.
bool COLdecode(SatAddr src, char *dst, int *nibbles);

/// \brief construct a CODE object
/// \param src The data to be stored in the CODE object
/// \param nibbles The number of nibbles in src to store.
/// \param dst If non-null, then this is the Saturn address where the CODE
/// object will be stored.  If null, then a new CODE object is
/// allocated in tempOb.
/// \return The Saturn address of the new CODE object, or zero if src
/// is NULL or a new CODE object can't be allocated.
SatAddr COLencode(const char *src, const int nibbles, SatAddr dst);

/// \brief create a COL containing the specified objects
/// \param dst The Saturn address where the COL should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// COL.  Otherwise dst must point to sufficient space for the COL.
/// See the warning under makeCOL
/// \param ob1, ob2, ... The Saturn addresses of the objects to put in
/// the COL.  If the parameter is an indirect pointer (i.e. the
/// Saturn address of a pointer to an object), then the pointer is put
/// in the COL, not the object.  If the parameter is
/// in ROM (addresses 0-0x40000) then the pointer to the object is
/// added to the COL, not the object itself.
/// \param 0 The parameter list must end with 0! Otherwise the
/// function won't know when to stop adding objects.
/// \return the Saturn address of the new COL.  Returns 0 on error
/// (can't allocate space, a parameter isn't an object or a pointer to
/// an object, etc.
/// \warning Note that a SEMI is automatically appended to the end of
/// the COL, so you don't need to include it in the parameter list.
/// Indeed, if you include it, you will confuse the garbage collector
/// and probably crash your calculator.
SatAddr COLencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ );

/// \brief create a COL containing the objects specified in an array
/// \param n The number of objects to put in the COL.
/// \param obs An array of n Saturn Addresses. At each address is an
/// object or a pointer to an object.  If an item in the array is an
/// indirect pointer (i.e. the Saturn address of a pointer to an
/// object), then the pointer is put in the COL, not the object.  If
/// the parameter is an in ROM (addresses 0-0x40000) then a pointer to
/// the object is added to the COL, not the object itself.
/// \param dst The Saturn address where the COL should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// COL.  Otherwise dst must point to sufficient space for the COL.
/// See the warning under makeCOL
/// \return the Saturn address of the new COL.  Returns 0 on error
/// (can't allocate space, a parameter isn't an object or a pointer to
/// an object, etc.
/// \warning Note that a SEMI is automatically appended to the end of
/// the COL, so you don't need to include it in the obs array.
/// Indeed, if you include it, you will confuse the garbage collector
/// and probably crash your calculator.
SatAddr COLencodeN(int n, SatAddr obs[], SatAddr dst);

/*!
  \brief Allocate space for a COL with nibbles data.

  This allocates space for a COL with nibbles data in it and puts an
  empty COL at the beginning of the allocated space.

  \param nibbles the number of nibbles to allocate
  \return Saturn address of the empty COL

  \warning Because of the way space is allocated on the calculator,
  you must fill the COL up so its size is exactly the number of
  nibbles allocated.  For this reason, you should use this function
  only when absolutely necessary. Use COLencode(), COLencodeV() or COLencodeN()
  to create COL's instead.
*/
SatAddr makeCOL(unsigned nibbles);

/// \brief Return the size in nibbles of the COL at src.
/// \param src the Saturn address of a COL object
/// \return the size of the COL in nibbles, or zero if src is not a COL
int COLnibbles(SatAddr src);

/// brief Return the number of bytes required to store the
/// data in a COL object
/// \param src The Saturn address of a COL object
/// \return the number of bytes, or -1 if src doesn't point to a COL object
int COLbytes(SatAddr src);

//
// Iteration
//
/// \brief Get the first object in a COL
///
/// This has the same semantics as LISTfirstOb(), but works on a COL
/// object
/// \sa LISTfirstOb(), LISTnextOb(), COLnextOb()
SatAddr COLfirstOb(SatAddr col, SatAddr *iter);

/// \brief Get the next object in a COL
///
/// This has the same semantics as LISTnextOb(), but works on a COL
/// object
/// \sa LISTfirstOb(), LISTnextOb(), COLfirstOb()
SatAddr COLnextOb(SatAddr *iter);

/// \brief Get the first entry in a COL.
///
/// The difference between COLfirstEntry() and COLfirstOb() is that
/// COLfirstEntry() returns whatever is in the list, whether it's an
/// object or an indirect pointer to an object.  COLfirstOb() always
/// returns the address of the object.
///
/// This has the same semantics as LISTfirstEntry(), but works on a COL
/// object.
/// \sa LISTfirstEntry(), LISTnextEntry(), COLnextEntry()
SatAddr COLfirstEntry(SatAddr col);

/// \brief Get the next entry in a COL.
///
/// This has the same semantics as LISTnextEntry(), but works on a COL
/// object.
/// \sa LISTfirstEntry, LISTnextEntry(), COLfirstEntry()
SatAddr COLnextEntry(SatAddr entry);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup SYMB Symbolic expression (SYMB)
/// @{
/// These functions have identical meeting to the equivalent ones for
/// COL objects or LIST objects.  In other words, SYMBxyz() has the same
/// symantics as COLxyz(), except that it operates on a SYMB object


/// \brief Same as isCOL() but works on a SYMB object
bool isSYMB(SatAddr src);

/// \brief Same as COLdecode() but works on a SYMB object
bool SYMBdecode(SatAddr src, char *dst, int *nibbles);

/// \brief Same as COLencode() but works on a SYMB object
SatAddr SYMBencode(const char *src, const int nibbles, SatAddr dst);

/// \brief Same as makeCOL() but works on a SYMB object
SatAddr makeSYMB(unsigned nibbles);

/// \brief Same as COLnibbles() but works on a SYMB object
int SYMBnibbles(SatAddr src);

/// \brief Same as COLbytes() but works on a SYMB object
int SYMBbytes(SatAddr src);

/// \brief Same as LISTfirstOb() but works on a SYMB object
SatAddr SYMBfirstOb(SatAddr src, SatAddr *iter);

/// \brief Same as LISTnextOb() but works on a SYMB object
SatAddr SYMBnextOb(SatAddr *iter);

/// \brief Same as LISTfirstEntry() but works on a SYMB object
SatAddr SYMBfirstEntry(SatAddr src);

/// \brief Same as LISTnextEntry() but works on a SYMB object
SatAddr SYMBnextEntry(SatAddr entry);

/// \brief Same as COLencodeV() but works on a SYMB object
SatAddr SYMBencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ );

/// \brief Same as COLencodeN() but works on a SYMB object
SatAddr SYMBencodeN(int n, SatAddr obs[], SatAddr dst);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup TAG Tagged object (TAG)
/// @{
/// A TAG object consists of an ASCII tag string and an arbitrary
/// object.

/// \brief is the object at src a TAG?
/// \param src Saturn address of an object
/// \return true if the object at src is a TAG.
bool isTAG(SatAddr src);

/// \brief Make a tagged object
/// \param tag A null-terminated string that is the tag.
/// \param obj The Saturn address of the object to be tagged
/// \param dst The Saturn address where the TAG should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// TAG.  Otherwise dst must point to sufficient space for the TAG.
/// \return The Saturn address of the TAG object, or zero on error.
SatAddr TAGencode(const char *tag, SatAddr obj, SatAddr dst);

/// \brief Return the number of tag characters in a TAG object
/// \param src the Saturn address of a TAG object
/// \return the number of characters in the tag, or -1 if src is not a
/// TAG object.
int TAGnumChars(SatAddr src);

/// \brief Decode a TAG object.
/// \param src The Saturn address of a TAG object
/// \param tag If non-null, then the tag will be stored here as a
/// null-terminated string.  In this case tag must point to sufficient
/// space to store it.  256 characters is enough, or you can create a
/// buf with TAGnumChars(src)+1 bytes.
/// \param obj If non-null, then the address of the object within the
/// TAG will be written here.
/// \return true on success, false if src isn't a TAG object.
bool TAGdecode(SatAddr src, char *tag, SatAddr *obj);

/// \brief Create an empty tag object with "nibbles" space.
///
/// This isn't very useful since you need to populate the tag object with
/// something. Use TAGencode() instead. This function is included
/// for consistency with other objects.
/// \param nibbles The number of nibbles in the TAG object
/// \return The new empty TAG object.
SatAddr makeTAG(int nibbles);

/// \brief Return the size in nibbles of the TAG at src.
/// \param src the Saturn address of a TAG object
/// \return the size of the TAG in nibbles, or zero if src is not a TAG
int TAGnibbles(SatAddr src);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup UNIT Unit object (UNIT)
/// @{
/// A UNIT object is structured a lot like a LIST, COL, or SYMB object,
/// but the contents of a UNIT object are pretty rigidly defined. 
/// These functions don't enforce the structure of the contents. See Programming in System
/// RPL chapter 13 for more details about units.  I'm only providing
/// rudimentary support for creating them here. Maybe in a later
/// version I'll provide more.
///
/// These functions have identical meeting to the equivalent ones for
/// COL objects or LIST objects.  In other words, UNITxyz() has the same
/// symantics as COLxyz(), except that it operates on a UNIT object
///
/// Most of the functions in this library use the name of the prolog
/// as a prefix for the name of the function (e.g., real numbers use
/// DOREAL for the prolog and the functions are all named REALxyz()).
/// The UNIT object's prolog is DOEXT, but I'll use UNIT as the prefix
/// instead. For convenience, I'll define SAT_DOUNIT also


// See Programming in SystemRPL for the meaning of these values
/// \brief The um* ROM marker.
#define UNITumMULT 0x2D74F	/* um* marker */

/// \brief The um/ ROM marker.
#define UNITumDIV  0x2D759 	/* um/ marker */

/// \brief The um^ ROM marker.
#define UNITumPOW  0x2D763	/* um^ marker */

/// \brief The umP ROM marker.
#define UNITumP    0x2D76D	/* umP marker */

/// \brief The umEND ROM marker.
#define UNITumEND  0x2D777	/* umEND marker */

/// \brief Alias for the prolog used for UNIT objects.
#define SAT_DOUNIT SAT_DOEXT

/// \brief Same as isCOL() but works on a UNIT object
bool isUNIT(SatAddr src);
/// \brief Same as COLdecode() but works on a UNIT object
bool UNITdecode(SatAddr src, char *dst, int *nibbles);
/// \brief Same as COLencode() but works on a UNIT object
SatAddr UNITencode(const char *src, const int nibbles, SatAddr dst);
/// \brief Same as makeCOL() but works on a UNIT object
SatAddr makeUNIT(unsigned nibbles);
/// \brief Same as COLnibbles() but works on a UNIT object
int UNITnibbles(SatAddr src);
/// \brief Same as COLbytes() but works on a UNIT object
int UNITbytes(SatAddr src);
/// \brief Same as LISTfirstOb() but works on a UNIT object
SatAddr UNITfirstOb(SatAddr src, SatAddr *iter);
/// \brief Same as LISTnextOb() but works on a UNIT object
SatAddr UNITnextOb(SatAddr *iter);
/// \brief Same as LISTfirstEntry() but works on a UNIT object
SatAddr UNITfirstEntry(SatAddr src);
/// \brief Same as LISTnextEntry() but works on a UNIT object
SatAddr UNITnextEntry(SatAddr entry);
/// \brief Same as COLencodeV() but works on a UNIT object
SatAddr UNITencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ );
/// \brief Same as COLencodeN() but works on a UNIT object
SatAddr UNITencodeN(int n, SatAddr obs[], SatAddr dst);

/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup DIR Directory (DIR)
/// @{
/// A directory contains a bunch of "entries" that consist
/// of name/value pairs.  Use DIRfirstEntry() and DIRnextEntry() to go
/// through the entries.  Then you can use DIRdecode() to extract the
/// name and object within each entry.

/// \brief Get the first entry in a directory
/// \param dir Saturn address of a DIR object
/// \return Saturn address of the first DIR entry or zero if the DIR
/// is empty or dir is not a DIR object
/// \sa DIRnextEntry(), DIRdecode()
SatAddr DIRfirstEntry(SatAddr dir);

/// \brief Get the next entry in a directory
/// \param entry The return value from a previous call to DIRfirstEntry() or
/// DIRnextEntry()
/// \return Saturn address of the next directory entry, or zero if
/// there are no more or if an error occurs. 
SatAddr DIRnextEntry(SatAddr entry);

/// \brief Decode a directory entry
/// \param entry The directory entry, which is the return value from a
/// previous call to DIRfirstEntry() or DIRnextEntry()
/// \param name If non-null then the name of the directory entry will
/// be stored here as a null-terminated string.  Name must point to
/// enough space to store the name.  256 bytes is always enough.  You
/// can also use DIRentryNameLen() to get the exact length
/// \param obj If non-null, then the Saturn address of the DIR entry's
/// object will be stored here.
/// \return true on success, false if entry is not a directory entry.
bool DIRdecodeEntry(SatAddr entry, char *name, SatAddr *obj);

/// Get the number of characters in a directory entry name
/// \param entry The directory entry, which is the return value from a
/// previous call to DIRfirstEntry() or DIRnextEntry()
/// \return The number of characters in the directory entry's name.
/// You can create a buffer with this many bytes, plus 1 for the
/// terminating null, and pass the buffer as the name parameter in
/// DIRdecodeEntry().
int DIRentryNameLen(SatAddr entry);

//
// Searching
//

/// \brief Find a directory entry by name.
/// \param dir A DIR object
/// \param name A null-terminated string that is the name to find.
/// \return The Saturn address of the directory entry or 0 if no entry
/// was found.  Unlike RCL, this searches in the dir specified only,
/// not the parent directories.
/// \sa DIRfindEntryByIdnt() DIRfindObByName(), DIRfindObByIdnt()
SatAddr DIRfindEntryByName(SatAddr dir, const char *name);

/// \brief Find a directory entry from an IDNT object
/// \param dir A DIR object
/// \param idnt Saturn address of an IDNT object (a global identifier)
/// \return The Saturn address of the directory entry or 0 if no entry
/// was found.  Unlike RCL, this searches in the dir specified only,
/// not the parent directories.
/// \sa DIRfindEntryByName()
SatAddr DIRfindEntryByIdnt(SatAddr dir, SatAddr idnt);

/// \brief find a directory object by name.
///
/// This is like DIRfindEntryByName, except it returns the Saturn
/// address of the object in the entry, not just the entry.
/// \sa DIRfindEntryByName(), DIRfindEntryByIdnt(), DIRfindObByIdnt()
SatAddr DIRfindObByName(SatAddr dir, const char *name);

/// \brief find a directory object from an IDNT object
///
/// This is like DIRfindEntryByIdnt, except it returns the Saturn
/// address of the object in the entry, not just the entry.
/// \sa DIRfindEntryByName(), DIRfindEntryByIdnt(), DIRfindObByName()
SatAddr DIRfindObByIdnt(SatAddr dir, SatAddr idnt);

// Creating

/// \brief Create an empty directory
/// \param dst Saturn address where the directory should go.
/// \warning If dst is in TEMPOB then you must fill the space
/// allocated completely.  Otherwise the garbage collector will fail
/// the next time it runs and you calculator will probably hang or reboot.
void DIRinit(SatAddr dst);

/// \brief Make an empty directory with "nibbles" nibbles of free
/// space.
/// \param nibbles the number of nibbles to allocate
/// \return The empty directory.
SatAddr makeDIR(int nibbles);

/// \brief Add an entry to a DIR.
///
/// The DIR must have enough space for the new entry.
/// \param src The DIR object to add to
/// \param name Null-terminated string of the new entry's name
/// \param obj Saturn address if the object. The object will be copied
/// to the directory.
bool DIRadd(SatAddr src, const char *name, SatAddr obj);

/// \brief A C structure for creating directories
struct DirItem {
    struct DirItem *next;	///< pointer to next item
    char *name;			///< must point to heap if you use freeDirItem */
    SatAddr obj;		///< Saturn address of the object.
};

/// \brief Decode an entire directory
/// \param src The DIR object to decode
/// \return a DirItem list of the objects in the directory.  This is
/// allocated on the heap and the caller must free it.
struct DirItem *DIRdecode(SatAddr src);

/// \brief Encode an entire directory
/// \param head Pointer to a DirItem list that represents the items to
/// go in the directory.
/// \param dst The Saturn address where the DIR should be stored. If
/// dst is zero (the expected case) then space is allocated for the
/// DIR in tempOb.  Otherwise dst must point to sufficient space for the DIR.
/// \return Saturn address of the directory, or zero if an error
/// occurred.
///
/// DIRencode() does not modify or free the DirItem list.
SatAddr DIRencode(struct DirItem *head, SatAddr dst);

/// \brief Overwrite the variable in a directory with an object of the
/// same size.
/// \param dir Saturn address of a DIR object.
/// \param name null-terminated string containing the name of the
/// directory entry to change.
/// \param obj Saturn address of the replacement object.  This object
/// must be the same size as the existing object associated with
/// "name."
/// \return 0 on success.  -1 if there is no entry for "name" in dir
/// or dir isn't a DIR object, -2 if "obj" is not the same size as the
/// existing object for "name".
int DIRchangeByName(SatAddr dir, const char *name, SatAddr obj);

/// \brief Same as DIRchangeByName(), but uses a global identifier
/// object (idnt) to lookup the name
int DIRchangeByIdnt(SatAddr dir, SatAddr idnt, SatAddr obj);


// Miscellaneous

/// \brief Return the size in nibbles of the DIR at src.
/// \param src the Saturn address of a DIR object
/// \return the size of the DIR in nibbles, or zero if src is not a DIR
int DIRnibbles(SatAddr dir);

/// \brief Get the DIR object that a encloses a directory entry.
/// \param entry Saturn address of a directory entry returned by
/// DIRfirstEntry(), DIRnextEntry(), DIRfindEntryByIdnt(), etc
/// \return The directory that the entry is in, or 0 on error.
/// \warning This won't work properly if "entry" is located in a copy
/// of the home directory and the copy has multiple libraries attached
/// to it.
SatAddr DIRdirForEntry(SatAddr entry);

/// \brief Get the parent of a directory
/// \param dir Saturn address of a DIR object
/// \return Saturn address of dir's parent DIR object, or zero if
/// "dir" is the HOME directory or if an error occurrs, such as when
/// "dir" isn't a DIR object.
SatAddr DIRparentDir(SatAddr dir);


/// \brief Get the Saturn address of the HOME directory
SatAddr getHOME();

/// \brief Get the Saturn address of the current directory
SatAddr getCWD();


/// \brief is the object at src a DIR?
/// \param src Saturn address of an object
/// \return true if the object at src is a DIR.
bool isDIR(SatAddr src);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup BACKUP BACKUP object
/// @{
///
/// I've included code for BACKUP objects, but I'm not sure how to
/// access a BACKUP so I haven't been able to test these.  USE AT YOUR
/// OWN RISK.
///
/// I don't know how to create the CRC for a BACKUP object so there
/// are currently no methods to create them.

/// \brief is the object at src a BACKUP?
/// \param src Saturn address of an object
/// \return true if the object at src is a BACKUP.
bool isBAK(SatAddr src);

/// \brief Return the size in nibbles of the BACKUP at src.
/// \param src the Saturn address of a BACKUP object
/// \return the size of the BACKUP in nibbles, or zero if src is not a BACKUP
int BAKnibbles(SatAddr src);

/// \brief Get the number of characters in a BACKUP's name.
/// \param src Saturn address of a BACKUP object.
/// \return The number of characters in the BACKUP's name, or -1
/// if "src" isn't a BACKUP object
int BAKnameLen(SatAddr src);

/// \brief Decode a BACKUP object
/// \param src Saturn address of a BACKUP object.
/// \param name If non-null, then the BACKUP's name will be stored
/// here as a null-terminated string.  In this case "name" must point
/// to sufficient space for the name and the null terminator.  256
/// bytes is always enough.  BAKnameLen(src)+1 is exactly the right
/// size.
/// \param obj If non-null then the Saturn address of the object in
/// the backup will be put here.
/// \param crc. If non-null then the address of the backup's CRC will
/// be put here. The CRC is BINT object.
/// \return true on success, false if src isn't a BACKUP object or
/// another error occurs.
bool BAKdecode(SatAddr src, char *name, SatAddr *obj, SatAddr *crc);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup ECMP Extended COMPLEX number (ECMP)
/// @{
///
/// Extended complex numbers use extended real (15 digit) numbers
/// for the real and complex parts.  The functions are identical to
/// the CMP functions, but they operate on ECMP objects

/// \brief Same as isCMP() but uses an ECMP object.
bool isECMP(SatAddr src);
/// \brief Same as CMPdecode() but uses an ECMP object.
bool ECMPdecode(SatAddr src, double *re, double *im);
/// \brief Same as CMPencode() but uses an ECMP object.
SatAddr ECMPencode(double re, double im, SatAddr dst);
/// \brief Same as makeCMP() but uses an ECMP object.
SatAddr makeECMP();
/// \brief Same as CMPnibbles() but uses an ECMP object.
int ECMPnibbles(SatAddr src);

/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup LIB Library (LIB)
/// @{
/// Libraries are very complex objects and I didn't think anyone would
/// be manipulating them in C code so I've included the bare minimum
/// of support for them.  If you need support, let me know and I'll
/// see if I can add it.

/// \brief is the object at src a LIB?
/// \param src Saturn address of an object
/// \return true if the object at src is a LIB.
bool isLIB(SatAddr src);

/// \brief Return the size in nibbles of the LIB at src.
/// \param src the Saturn address of a LIB object
/// \return the size of the LIB in nibbles, or zero if src is not a LIB
int LIBnibbles(SatAddr src);

/// @}
/////////////////////////////////////////////////////////////////////
/// \defgroup LNGREAL Infinite precision Real number (LNGREAL)
/// @{
///
/// A LNGREAL is like an infinite precision integer (INT), but it has
/// infinite precision mantissa and exponent.  Note that this is
/// different from the 15 digit extended-real object (EREL) built in
/// to the calculator.
///
/// I've done limited testing on these because there is no software
/// built-in to the calculator that handles them (except to display
/// them). "Programming in Saturn Assembly" says that there are
/// libraries that can manipulate them.
///
/// Note that unlike REAL numbers, the decimal point always comes
/// after the LAST mantissa digit, not the first.  So if the mantissa
/// is 123 and the exponent is 0, that means 123.0, not 1.23.


/// \brief is the object a LNGREAL
/// \param src Saturn address of an object
/// \return true if the object at src is a LNGREAL
bool isLNGREAL(SatAddr src);

/// \brief Convert LNGREAL to C double
/// \param src The LNGREAL to decode
/// \param dst Where to store the LNGREAL's value
/// \return true on success, false if src doesn't contain a LNGREAL
///
/// Some HP LNGREALs are too large or too small to be represented in a C
/// double and will result in a very large or very small number.
bool LNGREALdecoded(SatAddr src, double *dst);

/// \brief Convert a C double to an HP LNGREAL
/// \param src The double to convert.
/// \param dst Where to store the LNGREAL. If zero then a new LNGREAL is created.
/// \return The LNGREAL's Saturn address, or zero on error
SatAddr LNGREALencoded(double src, SatAddr dst);

/// \brief Decode a LNGREAL to its mantissa and exponent
/// \param src The LNGREAL to decode
/// \param mantissa If non-null, then the mantissa will be stored
/// here.
/// \param exponent If non-null then the exponent will be stored here.
/// \return true on success, false if the mantissa or exponent won't
/// fit in a long long.
bool LNGREALdecodell(SatAddr src, long long *mantissa, long long *exponent);

/// \brief Encode a LNGREAL from its mantissa and exponent.
/// \param mantissa The mantissa part.
/// \param exponent The exponent part.
/// \param dst Where to store the LNGREAL. If zero then a new LNGREAL is created.
/// \return The LNGREAL's Saturn address, or zero on error
SatAddr LNGREALencodell(long long mantissa, long long exponent, SatAddr dst);

/// \brief Decode a LNGREAL to its mantissa and exponent expressed as
/// null-terminated ASCII strings.
/// \param src The LNGREAL to decode
/// \param mantissa If non-null, then the mantissa will be stored
/// here as a null-terminated string.  mantissa must point to
/// sufficient space for the string.  You can get the required size
/// with LNGREALbytes()
/// \param exponent If non-null, then the exponent will be stored
/// here as a null-terminated string. Exponent must point to
/// sufficient space for the string.  You can get the required size
/// with LNGREALbytes()
/// \return true on success, false if src isn't a LNGREAL.
/// \sa LNGREALbytes()
bool LNGREALdecodeStr(SatAddr src, char *mantissa, char *exponent);

/// \brief Encode a LNGREAL from its mantissa and exponent expressed as
/// null-terminated ASCII strings.
/// \param mantissa ASCII string of the mantissa
/// \param exponent ASCII string of the exponent part.
/// \param dst Where to store the LNGREAL. If zero then a new LNGREAL is created.
/// \return The LNGREAL's Saturn address, or zero on error
SatAddr LNGREALencodeStr(const char *mantissa, const char *exponent, SatAddr dst);

/// \brief Create a new LNGREAL in tempOb. Populate it with zero.
SatAddr makeLNGREAL(int numMdigits, int numEdigits);

/// \brief The size of a LNGREAL in nibbles.
/// \param src Saturn address of the LNGREAL.
/// \return The size of the LNGREAL in nibbles.
int LNGREALnibbles(SatAddr src);

/// \brief Get the number of digits in a LNGREAL's mantissa and
/// exponent.
/// \param src Saturn address of a LNGREAL object.
/// \param mDigits If non-null then the number of digits in the
/// mantissa will be stored here.
/// \parram eDigits If non-null then the number of digits in the
/// exponent will be stored here.
/// \return true on success, false if src is not a LNGREAL
bool LNGREALdigits(SatAddr src, int *mDigits, int *eDigits);

/// \brief  Get the number of bytes required to store a LNGREAL's mantissa and
/// exponent as strings.
/// \param src Saturn address of a LNGREAL object.
/// \param mBytes If non-null then the number of digits required to
/// store the mantissa as a null-terminated ASCII string, including a
/// possible leading minus sign ("-") and a terminating null byte.
/// \param eBytes If non-null then the number of digits required to
/// store the exponent as a null-terminated ASCII string, including a
/// possible leading minus sign ("-") and a terminating null byte.
/// \return true on success, false if src isn't a LNGREAL
bool LNGREALbytes(SatAddr src, int *mBytes, int *eBytes);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup LNGCOMP Infinite precision Complex number (LNGCOMP)
/// @{
/// This is basically a complex number formed from the body of two
/// LNGREALs.

/// \brief is the object a LNGCOMP
/// \param src Saturn address of an object
/// \return true if the object at src is a LNGCOMP
bool isLNGCOMP(SatAddr src);


/// \brief Convert LNGCOMP to C doubles
///
/// Some HP LNGCOMPs are too large or too small to be represented by C
/// doubles and will result in a very large or very small number.
/// \param src The LNGCOMP to decode
/// \param re If non-null, the real part will be stored here.
/// \param im If non-null, the imaginary part will be stored here.
/// \return true on success, false if src doesn't contain a LNGCOMP
bool LNGCOMPdecoded(SatAddr src, double *re, double *im);

/// \brief Convert two C doubles to an HP LNGCOMP
/// \param re The real part.
/// \param im The imaginary part.
/// \param dst Where to store the LNGCOMP. If zero then a new
/// LNGCOMP is created.
/// \return The LNGCOMP's Saturn address, or zero on error
SatAddr LNGCOMPencoded(double re, double im, SatAddr dst);

/// \brief Decode a LNGCOMP into its mantissa and exponent expressed
/// as C long longs.
/// \param src The LNGCOMP to decode
/// \param re_mantissa If non-null, then the real part of the mantissa
/// will be stored here.
/// \param re_exponent If non-null, then the real part of the exponent
/// will be stored here.
/// \param im_mantissa If non-null, then the imaginary part of the mantissa
/// will be stored here.
/// \param im_exponent If non-null, then the imaginary part of the exponent
/// will be stored here.
/// \return true on success, false if any of the components won't
/// fit in a long long.
bool LNGCOMPdecodell(SatAddr src,
		     long long *re_mantissa, long long *re_exponent,
		     long long *im_mantissa, long long *im_exponent);

/// \brief Encode a LNGCOMP from its real and imaginary mantissa and
/// exponent expressed as C long longs.
/// \param re_mantissa Mantissa of real part
/// \param re_exponent Exponent of real part
/// \param im_mantissa Mantissa of imaginary part
/// \param im_exponent Exponent of imaginary part
/// \param dst Where to store the LNGCOMP. If zero then a new LNGCOMP
/// is created.
/// \return The LNGCOMP's Saturn address, or zero on error
SatAddr LNGCOMPencodell(long long re_mantissa, long long re_exponent,
			long long im_mantissa, long long im_exponent,
			SatAddr dst);


/// \brief Encode a LNGCOMP from its real and imaginary mantissa and
/// exponent expressed as null-terminated ASCII strings
/// \param re_mantissa Mantissa of real part
/// \param re_exponent Exponent of real part
/// \param im_mantissa Mantissa of imaginary part
/// \param im_exponent Exponent of imaginary part
/// \param dst Where to store the LNGCOMP. If zero then a new LNGCOMP
/// is created.
/// \return The LNGCOMP's Saturn address, or zero on error
SatAddr LNGCOMPencodeStr(const char *re_mantissa, const char *re_exponent,
			 const char *im_mantissa, const char *im_exponent,
			 SatAddr dst);

/// \brief Decode a LNGCOMP to its components expressed as
/// null-terminated ASCII strings.
///
/// If non-null, the re_mantissa, re_exponent, im_mantissa and
/// im_exponent parameters must point to sufficient space to hold
/// their respective components expressed as ASCII strings.  You can
/// getthe space required through the LNGCOMPbytes() function.
/// \param src The LNGCOMP to decode
/// \param re_mantissa If non-null, then the mantissa of the real part
/// will be stored here as a null-terminated ASCII string.
/// \param re_exponent If non-null, then the exponent of the real part
/// will be stored here as a null-terminated ASCII string.
/// \param im_mantissa If non-null, then the mantissa of the imaginary part
/// will be stored here as a null-terminated ASCII string.
/// \param im_exponent If non-null, then the exponent of the imaginary
/// part will be stored here as a null-terminated ASCII string.
/// \return true on success, false if src isn't a LNGCOMP.
/// \sa LNGCOMPbytes()
SatAddr LNGCOMPdecodeStr(SatAddr src,
			 char *re_mantissa, char *re_exponent,
			 char *im_mantissa, char *im_exponent);

// Make a LNGCOMP with room for the number of digits in the different parts
// given in the parameters. Populate it with zero

/// \brief Create a new LNGCOMP in tempOb. Withe space for the
/// specified number of digits in each component.
/// \return The Saturn address of the LNGCOMP.
SatAddr makeLNGCOMP(int numReMdigits, int numReEdigits,
		    int numImMdigits, int numImEdigits);


/// \brief The size of a LNGCOMP in nibbles.
/// \param src Saturn address of the LNGCOMP.
/// \return The size of the LNGCOMP in nibbles.
int LNGCOMPnibbles(SatAddr src);

/// \brief Get the number of digits in a LNGCOMP's real/imaginary
/// mantissa and exponent.
/// \param src Saturn address of a LNGCOMP object.
/// \param mReDigits If non-null then the number of digits in the
/// mantissa of the real part will be stored here.
/// \param eReDigits If non-null then the number of digits in the
/// exponent of the real part will be stored here.
/// \param mImDigits If non-null then the number of digits in the
/// mantissa of the imaginary part will be stored here.
/// \param eImDigits If non-null then the number of digits in the
/// exponent of the imaginary part will be stored here.
/// \return true on success, false if src is not a LNGCOMP.
bool LNGCOMPdigits(SatAddr src, int *mReDigits, int *eReDigits,
		   int *mImDigits, int *eImDigits);

/// \brief Get the number of bytes required to store the components of
/// a LNGCOMP as null-terminated ASCII strings.
/// \param src Saturn address of a LNGCOMP object.
/// \param mReDigits If non-null then the number of bytes required by the
/// mantissa of the real part will be stored here.
/// \param eReDigits If non-null then the number of bytes required by the
/// exponent of the real part will be stored here.
/// \param mImDigits If non-null then the number of bytes required by the
/// mantissa of the imaginary part will be stored here.
/// \param eImDigits If non-null then the number of bytes required by the
/// exponent of the imaginary part will be stored here.
/// \return true on success, false if src is not a LNGCOMP.
bool LNGCOMPbytes(SatAddr src, int *mReBytes, int *eReBytes,
		   int *mImBytes, int *eImBytes);
/// @}



//////////////////////////////////////////////////////////////////////
/// \defgroup MATRIX Symbolic MATRIX
/// @{
/// This is like a list, but with a different prolog and it must have
/// at least one object in it.  The routines within the calculator only
/// support including objects of type real, complex, integer, or
/// algebraic, so don't put anything else into a MATRIX.
///
/// Most of the functions for a MATRIX work the same as the equivalent
/// LIST functions.


/// \brief is the object a MATRIX
/// \param src Saturn address of an object.
/// \return true if the object at src is a MATRIX.
bool isMATRIX(SatAddr src);

/*!
  \brief Allocate space for a MATRIX with "nibbles" data and put a
  copy of "obj" in it.

  This allocates space for a MATRIC with nibbles data in it and puts an
  empty list at the beginning of the allocated space.

  \param nibbles the number of nibbles to allocate
  \param obj Saturn address of the object (or pointer to object) to
  copy into the MATRIX
  \return Saturn address of the MATRIX, or zero on error

  \warning Because of the way space is allocated on the calculator,
  you must fill the MATRIX up so its size is exactly the number of
  nibbles allocated.  For this reason, you should use this function
  only when absolutely necessary. Use MATRIXencode() or MATRIXencodeN()
  to create MATRIXes instead.
*/
SatAddr makeMATRIX(SatAddr obj, unsigned nibbles);

/// \brief Write a new matrix at "dst" and insert a copy of "obj" into
/// it.
/// \param obj Saturn address of an object or pointer to an object
/// \param dst Saturn address of the destination.
void MATRIXinit(SatAddr obj, SatAddr dst);

/// \brief Return the size in nibbles of the MATRIX at src.
/// \param src the Saturn address of a MATRIX object
/// \return the size of the MATRIX in nibbles, or zero if src is not a MATRIX
int MATRIXnibbles(SatAddr src);

/// \brief Same as LISTsize() but operates on a MATRIX
int MATRIXsize(SatAddr src);

/// \brief Same as LISTadd() but operates on a MATRIX
bool MATRIXadd(SatAddr dst, SatAddr obj);

/// \brief Same as LISTcat() but operates on a MATRIX
SatAddr MATRIXcat(SatAddr MATRIX, SatAddr obj);

/// \brief Same as LISTmerge() but operates on a MATRIX
SatAddr MATRIXmerge(SatAddr MATRIX1, SatAddr MATRIX2);

/// \brief Same as LISTencode() but operates on a MATRIX
SatAddr MATRIXencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ );

/// \brief Same as LISTencodeN() but operates on a MATRIX
SatAddr MATRIXencodeN(int n, SatAddr obs[], SatAddr dst);

/// \brief Same as LISTfirstOb() but operates on a MATRIX
SatAddr MATRIXfirstOb(SatAddr matrix, SatAddr *iter);

/// \brief Same as LISTnextOb() but operates on a MATRIX
SatAddr MATRIXnextOb(SatAddr *iter);

/// \brief Same as LISTfirstEntry() but operates on a MATRIX
SatAddr MATRIXfirstEntry(SatAddr matrix);

/// \brief Same as LISTnextEntry() but operates on a MATRIX
SatAddr MATRIXnextEntry(SatAddr entry);

/// @}


///////////////////////////////////////////////////////////////////////
/// \defgroup STACK Stack routines.
/// @{
/// Although very similar to the routines that are included with
/// HPGCC, there were enough small differences that I decided to
/// create my own functions to manipulate the stack.

/// \brief Pop an address off the stack.
/// \return The Saturn address of the object on level 1, or 0 if stack
/// is empty.
SatAddr STACKpop();

/// \brief push an address on the stack
/// \param src Saturn address of an object to push
/// \return void
#define STACKpush(src) sat_pushptr(src)

/// \brief pick an object on the stack from an arbitrary level.
/// \param the stack level (1, 2, 3, etc).
/// \return Saturn address of the object at the given level,
/// or zero if no object is at that level.
SatAddr STACKpick(int level);

/// \brief Get the stack depth.
///
/// This is faster and more accurate than sat_stack_depth(), which
/// will return the wrong value if you push zero on the stack.
/// \return the number of items on the stack.
int STACKdepth();

/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup generic Functions that work with any object
/// @{

/// \brief Get an object's prolog
/// \param src Saturn address of an object
/// \return The object's prolog.
#define obProlog(src) (sat_peek((src), 5))

/// \brief Are all objects with this prolog fixed in size
/// \param prolog An object prolog
/// \return true if all objects with this prolog have the same size,
/// otherwise false.
bool fixedSize(int prolog);

/// \brief Is there a recognized HP object at src?
///
/// This function is very fast but not 100% accurate. It works by
/// assuming that src contains an object and inspecting the first
/// nibble at the object's prolog address.  For real objects, this
/// value is always the same.  Note however  that it's possible to get
/// a false positive result. That is, isOb() may return true when in
/// fact "src" doesn't contain an object.
///
/// If you need to be certain that an object exists at the address,
/// call obSize() and see if the result is greater than zero.
/// \param src A Saturn address
/// \return true if there appears to be an object at Src. Otherwise
/// false.
/// \sa obSize()
bool isOb(SatAddr src);

/// \brief Return the size in nibbles of object or pointer at src.
/// \param src Saturn address of an object or indirect pointer
/// \return The size of the object or indirect pointer.
int obNibbles(SatAddr src);

/// \brief Return the address of the object after src, which can point
/// to an object or an indirect pointer. This is basically the C
/// version of skipOb.
/// \param src Saturn address of an object, or a pointer to an object
/// \return The address right after the object or pointer.  This just
/// returns src+obSize(src).
/// \sa obSize()
SatAddr skipOb(SatAddr src);

/// \brief Copy an object or pointer from src to dst, but only if the
/// object at dst is the same size as the one at src.
/// \param src Saturn address of the source object or pointer
/// \param dst Saturn address of the desination
/// \return true what was at src and dst were the same sizes and src
/// was copied to the dst.  False otherwise.
/// \warning If anything in the calculator
// points to the dst object (like a value on the stack), then you'll
// change that also!
bool obCopy(SatAddr src, SatAddr dst);

/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup Indirect Functions to deal with indirect pointers
/// @{
/// sys RPL and UserRPL programs can contain objects or pointers to
/// objects. Indeed, this is a key property of RPL programs.  The same
/// property is true of other composite objects like lists and
/// matrixes.  This section contains support functions for dealing
/// with indirect pointers.
/// 
/// You probably won't need to use these. The library already contains
/// functions that will tunnel through indirect pointer.  See
/// LISTfirstOb() and LISTnextOb() for example.

/// \brief Get the direct pointer from a possible indirect pointer
///
/// Given an address that contains either an object or a pointer to an
/// object, get the address of the object.  directPtr() is used in
/// much of the code that processes composite objects like lists and
/// matrixes.
/// \param src Saturn address of an object or a pointer to an object
/// \return Saturn address of the object.
SatAddr directPtr(SatAddr src);

/// \brief Can you store an object as an indirect pointer?
///
/// If you add a ROM object to a composite then there's no need to
/// store the object itself, you can store a pointer to it instead.
/// This function tells you whether you can store the pointer.  Its
/// function is currently quite simple. Any object below 0x40000 is in
/// ROM and be stored as a pointer.
/// \brief obj Saturn address of an object
/// \return true if you can store the pointer.
bool obStorePtr(SatAddr obj);

/// \brief how many nibbles are in an object or pointer?
/// \param obj Saturn address of an object or indirect pointer
/// \return the size of the object or pointer.
int objOrPtrNibbles(SatAddr obj);

/// \brief Copy an object or indirect pointer.
///
/// This decides whether the object can be copied as a pointer and
/// then copies the object or the pointer.
/// \param src Saturn address of an object
/// \param dst Saturn address of the destination
/// \return The number of nibbles copied (5 if its a pointer,
/// obSize(src) otherwise).  Returns -1 on error.
int copyObjOrPtr(SatAddr src, SatAddr dst);
/// @}


/////////////////////////////////////////////////////////////////////
/// \defgroup MISC Miscellaneous routines
/// @{

/// \brief Convert a C double to an HP 64 bit real number.
///
/// Note that this does NOT convert it to the 21 nibble DOREAL object,
/// it just converts the binary formats.  This is the inverse of
/// sat_real2double(), except it always writes to the Saturn address
/// space.
/// \param number The double to convert
/// \param dst Saturn address where the HP REAL body should be written
void double2hp(double number, SatAddr dst);

/// \brief Convert a C double to an HP 21 nibble extended real number.
///
/// Note that this does NOT convert it to the 26 nibble DOEREL object,
/// it just converts the binary formats.  This is the inverse of
/// ereal2double(), except it always writes to the Saturn address
/// space.
/// \param number The double to convert
/// \param dst Saturn address where the HP EREL body should be written
void double2hpereal(double number, SatAddr dst);

/// \brief Convert a long long to Binary Coded Decimal (BCD), one digit per
/// nibble, LSB stored first.
/// \param src The input number.
/// \param dst Saturn address where the BCD should be written
/// \param numDigits The number of digits/nibbles to write. -1 means
/// as many as required.
/// \return The number of digits actually written.
int ll2bcd(unsigned long long src, SatAddr dst, int numDigits);

/// \brief Convert a null terminated string of digits to BCD, one digit per
/// nibble, LSB stored first.
/// \param src The input number.
/// \param dst Where to store the BCD.
/// \param numDigits The number of digits to write. -1 means as many
/// as required).
/// \return The number of digits actually written.
int str2bcd(const char *src, SatAddr dst, int numDigits);

/// \brief Return the number of digits in a long long
int llDigits(unsigned long long src);

/// @brief Extract the base-10 exponent and mantissa from a double.
/// Store the exponent in ex.  1 <= abs(mantissa) < 10.
void doubledecode(double number, int *ex, double *mantissa);

/// @brief decode a C double into mantissa and exponent.
/// @param number The number to decode
/// @param mantissa Pointer to a long long where the mantissa will be
/// stored
/// @param exponent Pointer to a long long where the exponent will be
/// stored. Although this could have been an int or even a short, I'm
/// using a long long so the programmer doesn't have to worry about
/// which type is needed for which param.
void doubledecodell(double number, long long *mantissa, long long *exponent);

/// @brief Absolute value of a long long.
long long llabs(long long src);

/// @brief Like memset(), but works in the saturn domain and sets a specified
/// number of nibbles, not bytes
void sat_memset(SatAddr dst, int val, int nibbles);


/// @brief implementation of LNGREALnibbles() but doesn't check prolog
/// This function is also used LNGCOMPnibbles().
int lngrealNibbles(SatAddr src);

/// @brief implementation of LNGREALdigits() but doesn't check prolog
/// This function is also used LNGCOMPdigits().
bool lngrealDigits(SatAddr src, int *mDigits, int *eDigits);

/// @brief Given a prolog, return its name, or "unknown".
/// @param prolog the Prolog value
/// @return null-terminated string like "DOBINT" or "DOREAL"
char *prolog2Name(SatAddr prolog);

/// @brief Return a prolog from a name.

/// Given a prolog name like "DOBINT", return it's prolog or zero if
/// unknown.  The name is case insensitive and may optionally omit the
/// "DO" prefix.  I.E. "DOBINT", "dobint", "bint" and "BINT" will all
/// return SAT_DOBINT
SatAddr name2Prolog(const char *name);
/// @}

/////////////////////////////////////////////////////////////////////
/// \defgroup Composites Generic Functions for Composite Objects
/// @{
/// 
/// Composite routines.
/// A composite object consists of a prolog, a bunch of objects, or
/// pointers to objects, and finally a 5 nibble SEMI.  The most
/// familiar composite is the list, but RPL programs, symbolic matrices
/// and others also follow the form.
///
/// The functions that follow implement common functionality for any
/// composite object.  Type-specific methods exist for each composite
/// type also. The type-specific versions check (or set) the
/// appropriate prolog value.

/// \brief Implementation for LISTnibbles() etc. Doesn't check the prolog.
int compositeObNibbles(SatAddr src);

/// \brief Implementation for LISTadd() etc. Doesn't check the prolog.
bool compositeAdd(SatAddr dst, SatAddr obj);

/// \brief Implementation for LISTcat() etc. Doesn't check the prolog.
SatAddr compositeCat(SatAddr src, SatAddr obj);

/// \brief Implementation for LISTmerge() etc. Doesn't check the prolog.
SatAddr compositeMerge(SatAddr comp1, SatAddr comp2);

/// \brief Implementation for LISTfirstEntry() etc. Doesn't check the prolog.
SatAddr compositeFirstEntry(SatAddr col);

/// \brief Implementation for LISTnextEntry() etc. Doesn't check the prolog.
SatAddr compositeNextEntry(SatAddr entry);

/// \brief Implementation for LISTfirstOb() etc. Doesn't check the prolog.
SatAddr compositeFirstOb(SatAddr col, SatAddr *iter);

/// \brief Implementation for LISTnextOb() etc. Doesn't check the prolog.
SatAddr compositeNextOb(SatAddr *iter);

/// \brief Implementation for LISTencode() etc. Doesn't check the prolog.
SatAddr compositeEncode(SatAddr dst, va_list argp);

/// \brief Implementation for LISTencodeN() etc. Doesn't check the prolog.
SatAddr compositeEncodeN(int n, SatAddr obs[], SatAddr dst);

/// \brief Implementation for LISTSize() etc. Doesn't check the prolog.
int compositeSize(SatAddr src);

/// @}

// TODO: easy way to encode objects right into a composite object.

#endif
