#include <stdarg.h>
#include <hpgcc49.h>
#include "hpobjects.h"
#include "array.h"
//
// Array routines
// Note that in the C routines, array indexes start at zero.
//

// Arrays are described in section 57.2 of "Introduction to Saturn
// Assembly Language" by Fernandes and Rechlin. These are
// N-dimensional arrays of objects of one type.  The user can only
// create objects of REAL or COMPLEX type, but any type will do.
//
// Since the objects can be of any type, the items can be variable
// length.  The functions here try to be smart about this. If the item
// type is known to be of fixed length, then we use arithmetic to find
// the items. Otherwise it's a linear search.

// Arrays consist of:
//
// Prolog (5 nibbles)
// Size (5 nibbles). Everything except the prolog
// Item prolog (5 nibbles). Prolog of the items within the array.
// number of dimensions (5 nibbles)
// Dimension 1 (5 nibbles) Size of dimension 1
// Dimension 2 (5 nibbles)
// ...
// Dimension N (5 nibbles)
// Body of object 1 (variable size)
// Body of object 2 (variable size)
// ...
// Body of last object (variable size)

//++
// Is this an array?
bool isARRAY(SatAddr src)
{
    return (obProlog(src) == SAT_DOARRY);
}


// Read a bunch of indexes or dimensions from a vargs list and 
// return an array of them, or NULL if any dimension is larger than 0xffff.
// This is shared with lnkarry.c
int *
commonGetDimensions(int numDimensions, int dim1, va_list argp)
{
    int i;
    int *dimensions = malloc(numDimensions * sizeof(int));

    dimensions[0] = dim1;
    for (i = 1; i < numDimensions; ++i) {
	dimensions[i] = va_arg(argp, int);
	if (dimensions[i] != (dimensions[i]&0xffff)) {
	    free(dimensions);
	    return NULL;
	}
    }
    return dimensions;
}

// Create an N-dimensional array with the dimensions given.
// Initialize each item in the array with a copy of "element"
// Returns the array
SatAddr makeARRAY(SatAddr element,
		  int numDimensions,
		  int dim1,
		  ... /* dim2,dim3 etc */ )
{
    va_list argp;
    SatAddr result;		/* the array */
    SatAddr addr;		/* current address in the array */
    int nibbles;		/* size of the array */
    int *dimensions;		/* array of dimensions */
    int numElements=1;		/* total number of elements */
    int elementSize;		/* size of element */
    int i;

    // Get the dimensions from the arg list
    va_start(argp, dim1);
    dimensions = commonGetDimensions(numDimensions, dim1, argp);
    va_end(argp);

    if (dimensions==NULL) return 0;

    // compute the number of elements
    for (i = 0; i < numDimensions; ++i) {
	numElements *= dimensions[i];
    }

    // Compute the size of the element, which is the size of the
    // object, minus its prolog. Since obNibbles returns -1 if it isn't
    // an object, we can check for a valid element here too
    elementSize = obNibbles(element) - 5;
    if (elementSize<0) return 0;

    // Compute the size of the array
    nibbles = 5;		/* prolog */
    nibbles += 5+5+5;		/* size, item prolog, # dimensions */
    nibbles += 5 * numDimensions;
    nibbles += numElements * elementSize;

    // Now make the array
    result = sat_createtemp(nibbles);
    sat_poke(result, SAT_DOARRY, 5);		   /* prolog */
    sat_poke(result+5, nibbles-5, 5);		   /* size */
    sat_poke(result+10, sat_peek(element, 5), 5);  /* kind */
    sat_poke(result+15, numDimensions, 5);	  /* num dimensions */

    addr = result+20;
    for (i = 0; i < numDimensions; ++i) {	  /* each dimension */
	sat_poke(addr, dimensions[i], 5);
	addr += 5;
    }
    
    for (i = 0; i < numElements; ++i) {		   /* N copies of element */
	sat_moveup(element+5, addr, elementSize);
	addr += elementSize;
    }

    free(dimensions);
    return result;
}


// Worker function to find an element body from it's index numbers
// Returns zero on success.
// -10-n means the n'th index value is out of range.
int ARRAYgetElementBodyv(SatAddr array,
		    SatAddr *elementBody,
		    int idx1,
		    va_list argp)
{
    int i;
    int *indices;
    int n=0;				  /* element number */
    int blockSize=1;
    int numDimensions = ARRAYnumDimensions(array);
    SatAddr addr;				   /* address of the element */

    // Get the indices from the arg list
    indices = commonGetDimensions(numDimensions, idx1, argp);

    // Flatten out the indices to get the linear element number.
    for (i=numDimensions-1; i >= 0; --i) {
	int dim = ARRAYgetDimension(array, i);
	if (indices[i] < 0 || indices[i] >= dim) {
	    free(indices);
	    return 10-i-1;
	}
	n += blockSize * indices[i];
	blockSize *= dim;
    }

    // Now get the n'th element. Do this differently depending on
    // whether it's a fixed size or variable sized item
    if (fixedSize(ARRAYgetElProlog(array))) {
	// Get the element size by pretending that the element
	// prolog field is the prolog of a real object
	int elementSize = obNibbles(array+10);
	elementSize -= 5;			   /* no prolog */
	addr = ARRAYfirstElement(array) + elementSize * n;
    } else {
	for (i=0, addr = ARRAYfirstElement(array); i < n; ++i) {
	    addr = ARRAYnextElement(array, addr);
	}
    }

    free(indices);
    *elementBody = addr;
    return 0;
}

// Get a pointer to the body of an element in the array.
// Returns zero on success.
// -10-n means the n'th index value is out of range.
int ARRAYgetElementBody(SatAddr array, SatAddr *elementBody,
			int idx1,
			... /* idx2, idx3, etc */ )
{
    va_list argp;
    int result;

    if (!isARRAY(array)) return -3;
    va_start(argp, idx1);
    result = ARRAYgetElementBodyv(array, elementBody, idx1, argp);
    va_end(argp);
    return result;
}


// Create an HP object corresponding to an item in an array.

// Arrays store the bodies of each object, but only one copy of the
// prolog.  The function creates an HP object with the prolog and body
// together. "Array" is the array.  "dst" is the destination for an
// object that's a copy of the one in the array.  If "dst" is zero
// then a new object will be created. idx1...idxn are the indices,
// which are zero-based.  Returns the object or zero on error.
SatAddr ARRAYgetItem(SatAddr array, SatAddr dst, int idx1,
		     ... /* idx2, idx3 etc */ )
{
    SatAddr body;
    int i;
    va_list argp;

    if (!isARRAY(array)) return 0;
    va_start(argp, idx1);
    i = ARRAYgetElementBodyv(array, &body, idx1, argp);
    va_end(argp);

    if (i!=0) return 0;
    return commonMakeItem(array, body, dst);
}

// Common code for ARRAYelementSize and LNKARRYelementSize
int commonElementSize(SatAddr array, SatAddr addr)
{
    int oldValue;
    int result;

    int prolog = sat_peek(array+10, 5);
    // backup to the location of the fake object;
    addr -= 5;
    oldValue = sat_peek(addr, 5);
    sat_poke(addr, prolog, 5);
    result = obNibbles(addr) - 5;	   /* don't include prolog */
    sat_poke(addr, oldValue, 5);
    return result;
}


// Given an array and a pointer to an item in an array, return the
// size of the item in the array.  This works by temporarily changing
// the 5 nibbles before the item body to the prolog and then calling
// obNibbles().  This only works in a single threaded environment.
int ARRAYelementSize(SatAddr array, SatAddr addr)
{
    if (!isARRAY(array)) return -1;
    return commonElementSize(array, addr);
}

// Put an item into an array.  The item must be the same type
// as the array contents.  Also, it must have the same length as the item
// currently at the position.  Finally, you must specify the right number of
// indexes.  If the elements in the array are fixed length then this will
// find them quickly. Otherwise it must perform a linear search.
//
// Returns zero on success.
// -1 means item is the wrong type
// -2 means the item isn't the same size as the one there.
// -3 means array isn't an array
// -10-n means the n'th index value is out of range.
int ARRAYputItem(SatAddr array,
		 SatAddr item,
		 int idx1,
		 ... /* idx2, idx3, ... */ )
{
    int result;
    int newSize;
    int existingSize;
    SatAddr addr;
    va_list argp;

    if (!isARRAY(array)) return -3;

    if (sat_peek(item, 5) != ARRAYgetElProlog(array)) {
	return -1;				   /* wrong element type */
    }

    va_start(argp, idx1);
    result = ARRAYgetElementBodyv(array, &addr, idx1, argp);
    va_end(argp);
    if (result < 0) {
	return result;
    }

    newSize = obNibbles(item);
    existingSize = ARRAYelementSize(array, addr);
    if (newSize != existingSize+5) {
	return -2;
    }
    sat_moveup(item+5, addr, existingSize);
    return 0;    
}


// Returns pointer to the first element body
SatAddr ARRAYfirstElement(SatAddr array)
{
    SatAddr result;

    if (!isARRAY(array)) return 0;
    result = array + 5*4;	   /* prolog, size, kind, #dims */
    result += 5 * ARRAYnumDimensions(array);
    return (result < skipOb(array) ? result : 0);
}


// Given an array and point to the body of an item in the array,
// return a pointer to the next body in the array, or zero if no
// more elements exist.
SatAddr ARRAYnextElement(SatAddr array, SatAddr item)
{
    SatAddr result;
    if (!isARRAY(array)) return 0;
    result = item + ARRAYelementSize(array, item);
    return (result < skipOb(array) ? result : 0);
}


SatAddr ARRAYfirstItem(SatAddr array, SatAddr dst, SatAddr *iter)
{
    *iter = ARRAYfirstElement(array);
    if (*iter == 0) return 0;
    return commonMakeItem(array, *iter, dst);
}

SatAddr ARRAYnextItem(SatAddr array, SatAddr dst, SatAddr *iter)
{
    *iter = ARRAYnextElement(array, *iter);
    if (*iter == 0) return 0;
    return commonMakeItem(array, *iter, dst);
}

// Size of the Array object
int ARRAYnibbles(SatAddr src)
{
    if (!isARRAY(src)) return -1;
    return sat_peek(src+5, 5) + 5;
}


 /* get prolog of elements in array */
int ARRAYgetElProlog(SatAddr array)
{
    int result;
    if (!isARRAY(array)) return -1;
    result = sat_peek(array+10, 5);
    return result;
}

int ARRAYnumDimensions(SatAddr array)
{
    int result;
    if (!isARRAY(array)) return -1;
    result = sat_peek(array+15, 5);
    return result;
}

// Common method for ARRAY and LNKARRY to get dimension for an array
// or linked array
int commonGetDimension(SatAddr array, int dimensionNumber)
{
    int result;
    int numDimensions = sat_peek(array+15, 5);

    if (dimensionNumber<0 || dimensionNumber >= numDimensions) {
	return -1;
    }

    result = sat_peek(array+20+5*dimensionNumber, 5);
    return result;
}

// Get the size of the n'th dimension (n=1,2,3,...)
int ARRAYgetDimension(SatAddr array, int dimensionNumber)
{
    if (!isARRAY(array)) return -1;
    return commonGetDimension(array, dimensionNumber);
}



// Copy the array item prolog and the body of an object in the array
// to "dst," thus forming a full object that's a copy of the body.
// Note that this does NOT ensure that "array" is an array - all the
// callers do that before calling this.
SatAddr commonMakeItem(SatAddr array, SatAddr body, SatAddr dst)
{
    int size = commonElementSize(array, body);
    int elProlog = sat_peek(array+10, 5);
    if (dst==0) {
	dst = sat_createtemp(size+5);
	if (dst == 0) return 0;
    }
    sat_poke(dst, elProlog, 5);
    sat_moveup(body, dst+5, size);
    return dst;
}

// Given an array and the body of an item within the array, make an
// object for the item at dst.  This just copies the ARRAYelProlog()
// to dst, and copies the array item body to dst+5.  The body's size
// is ARRAYelementSize()
bool ARRAYmakeItem(SatAddr array, SatAddr itemBody, SatAddr dst)
{
    if (!isARRAY(array)) return false;
    return commonMakeItem(array, itemBody, dst);
}
