#include <hpgcc49.h>
#include "hpobjects.h"

//
// COL routines
//

// A Code object is encoded almost the same as a STRING, except that the
// length field is in nibbles, not bytes.

extern int COLSYMBbytes(SatAddr src);
extern bool COLSYMBdecode(SatAddr src, char *dst, int *nibbles);
//++

bool isCOL(SatAddr src)
{
    return (obProlog(src) == SAT_DOCOL);
}

// Common function for COLdecode() and SYMBdecode()
bool COLSYMBdecode(SatAddr src, char *dst, int *nibbles)
{
    if (dst) {
	sat_peek_sat_bytes(dst, src+5, COLSYMBbytes(src));
    }
    if (nibbles) {
	*nibbles = compositeObNibbles(src)-10;
    }
    return true;
}


// Given a COL object at src, copy the code itself (everything between
// the prolog and SEMI to dst. If dst is null, then a new COL will be
// created. Otherwise dst must be big enough to accept it; Note that
// this does NOT copy the prolog or composite themselves.
bool COLdecode(SatAddr src, char *dst, int *nibbles)
{
    if (!isCOL(src)) return false;
    return COLSYMBdecode(src, dst, nibbles);
}

// Construct a COL object at DST and copy "nibbles" nibbles there
// from "src".  The src should not include the prolog or terminating SEMI,
// which will be added automatically. If dst is 0 then space will be
// allcated for the object. OTherwise it must be big enough to hold it
// (nibbles+10)
SatAddr COLencode(const char *src, const int nibbles, SatAddr dst)
{
    if (dst == 0) {
	dst = makeCOL(nibbles);
    }
    sat_poke(dst, SAT_DOCOL, 5);	  /* prolog */
    sat_poke_nibbles(dst+5, src, nibbles);
    sat_poke(dst+5+nibbles, SAT_SEMI, 5);
    return dst;
}


/* Allocate space for a COL object with nibbles nibbles of code, and
   put an empty COL object there. Return the address of the COL.*/
SatAddr makeCOL(unsigned nibbles)
{
    SatAddr result = sat_createtemp(nibbles+10);
    sat_poke(result, SAT_DOCOL, 5);	  /* prolog */
    sat_poke(result+5, SAT_SEMI, 5);	  /* size. This one is empty */
    return result;
}

// Return the size in nibbles of the COL object at src.
int COLnibbles(SatAddr src)
{
    if (!isCOL(src)) return -1;
    return compositeObNibbles(src);
}


int COLSYMBbytes(SatAddr src)
{
    return (compositeObNibbles(src)-9) / 2;
}

// Return the number of bytes necessary to hold the code in the COL object.
// There are COLnibbles()-10 nibbles of code, and to store N nibbles, you need
// (n+1)/2 bytes. So the result is (COLnibbles()-9)/2
int COLbytes(SatAddr src)
{
    if (!isCOL(src)) return -1;
    return COLSYMBbytes(src);
}




// Return the first object in a col, or 0 of the col is nil
SatAddr COLfirstEntry(SatAddr col)
{
    if (!isCOL(col)) return 0;
    return compositeFirstEntry(col);
}


// Given an entry from a col returned by COLfirstEntry or
// COLnextEntry, return the next entry in the col.
SatAddr COLnextEntry(SatAddr entry)
{
    return compositeNextEntry(entry);
}

// Return a pointer to the first object in the col, or zero if "col"
// isn't a col or if the col is empty.  "Iter" is the address of a
// scratch pointer used by COLfirstOb/COLnextOb to keep track of the
// position in the col. Do not modify iter between calls.  This
// dereferences indirect pointers in the col the ensure that it
// returns a pointer to an object
SatAddr COLfirstOb(SatAddr col, SatAddr *iter)
{
    if (!isCOL(col)) return 0;
    return compositeFirstOb(col, iter);
}


// Given an iter that was previously passed to COLfirstOb or
// COLnextOb, return the address of the next object in the col and
// update iter.  This dereferences indirect pointers in the col the
// ensure that it returns a pointer to an object
SatAddr COLnextOb(SatAddr *iter)
{
    return compositeNextOb(iter);
}



// Make a new col consisting of the objects ob1, ob2, ... obn and put
// them in dst. If dst is zero the space is allocated for the new
// col. Otherwise dst must contain sufficient space. Return the new col
SatAddr COLencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ )
{
    va_list argp;
    va_start(argp,dst);
    dst = compositeEncode(dst, argp);
    sat_poke(dst, SAT_DOCOL, 5);
    va_end(argp);
    return dst;
}

// Make a new col consisting of the objects obs[0], obs[1], obs[2],
// ... obs[n-1] and put them in dst. If dst is zero then space is
// allocated for the new col. Otherwise dst must contain sufficient
// space. Return the new col. Basically this is like ->COL on the
// calculator
SatAddr COLencodeN(int n, SatAddr obs[], SatAddr dst)
{
    dst = compositeEncodeN(n,obs,dst);
    if (dst) sat_poke(dst, SAT_DOCOL, 5);
    return dst;    
}
