#include <stdarg.h>
#include <hpgcc49.h>
#include "hpobjects.h"

//
// Implementation of composite routines.  The caller is responsible for
// ensuring that the composite object is the right type.
//
//++
// Returns size of a composite object that uses an composite.
int compositeObNibbles(SatAddr src)
{
    SatAddr end;
    int size;
    for (end = src+5; sat_peek(end, 5) != SAT_SEMI; end += size) {
	size = obNibbles(end);
	if (size == -1) return -1;
    }
    return (end+5) - src;
}


// Return the first object in a col, or 0 of the col is empty
SatAddr compositeFirstEntry(SatAddr col)
{
    if (sat_peek(col+5, 5) == SAT_SEMI) return 0;
    return col+5;
}


// Given an entry from a col returned by compositefirstEntry or
// compositenextEntry, return the next entry in the col.
SatAddr compositeNextEntry(SatAddr entry)
{
    entry = skipOb(entry);
    if (entry == 0) return 0;
    if (sat_peek(entry, 5) == SAT_SEMI) return 0;
    return entry;
}

// Return a pointer to the first object in the col, or zero if "col"
// isn't a col or if the col is empty.  "Iter" is the address of a
// scratch pointer used by compositefirstOb/compositenextOb to keep track of the
// position in the col. Do not modify iter between calls.  This
// dereferences indirect pointers in the col the ensure that it
// returns a pointer to an object
SatAddr compositeFirstOb(SatAddr col, SatAddr *iter)
{
    *iter = compositeFirstEntry(col);
    if (*iter == 0) return 0;
    return directPtr(*iter);
}


// Given an iter that was previously passed to compositefirstOb or
// compositenextOb, return the address of the next object in the col and
// update iter.  This dereferences indirect pointers in the col the
// ensure that it returns a pointer to an object
SatAddr compositeNextOb(SatAddr *iter)
{
    *iter = compositeNextEntry(*iter);
    if (*iter == 0) return 0;
    return directPtr(*iter);
}



// Make a new list-like object consisting of the objects ob1, ob2,
// ... obn and put them in dst. If dst is zero the space is allocated
// for the new object. Otherwise dst must contain sufficient
// space. Return the new object.  NOTE: the caller must insert the
// prolog.
//
// list-like objects can contain objects inline, or they can contain
// pointers to objects.  If an object is a pointer another object,
// then the pointer is inserted.  Otherwise if it points to an address
// less than 0x40000, then a pointer is inserted (because the space
// from 0 to 0x3ffff is ROM.  Otherwise, a copy of the object is
// placed into the list.
SatAddr compositeEncode(SatAddr dst, va_list argp)
{
    SatAddr addr;
    SatAddr result;
    int size, s;

    if (dst == 0) {
	// Compute the size and allocate the object
	va_list argp2 = argp;
	size = 10;		/* prolog & composite */
	while ((addr = va_arg(argp2, SatAddr))) {
	    s = objOrPtrNibbles(addr);
	    if (s == -1) return 0; /* not an object */
	    size += s;
	}
	va_end(argp2);
	dst = sat_createtemp(size);
    }
    // sat_poke(dst, SAT_DOLIST);  // the caller will fill in the prolog
    result = dst;
    dst += 5;
    while ((addr = va_arg(argp, SatAddr))) {
	// No need to check for errors because they would be caught above
	size = copyObjOrPtr(addr, dst);
	if (size < 0) return 0;
	dst += size;
    }
    sat_poke(dst, SAT_SEMI, 5);
    return result;
}

// Make a new col consisting of the objects obs[0], obs[1], obs[2],
// ... obs[n-1] and put them in dst. If dst is zero then space is
// allocated for the new col. Otherwise dst must contain sufficient
// space. Return the new col. Basically this is like ->composite on the
// calculator
SatAddr compositeEncodeN(int n, SatAddr obs[], SatAddr dst)
{
    int i;
    int size, s;
    SatAddr result;

    if (dst == 0) {
	// Compute the size and allocate the object
	size = 10;		/* prolog & semi */
	for (i=0; i<n; ++i) {
	    s = objOrPtrNibbles(obs[i]);
	    if (s == -1) return 0; /* not an object */
	    size += s;
	}
	dst = sat_createtemp(size);
    }
    // sat_poke(dst, SAT_DOLIST);   // the caller will fill in the prolog
    result = dst;
    dst += 5;
    for (i=0; i<n; ++i) {
	// No need to check for errors because they would be caught above
	size = copyObjOrPtr(obs[i], dst);
	if (size < 0) return 0;
	dst += size;
    }
    sat_poke(dst, SAT_SEMI, 5);
    return result;
}


int compositeSize(SatAddr src)
{
    SatAddr end;
    unsigned result = 0;
    for (end = src+5; sat_peek(end, 5) != SAT_SEMI; end += obNibbles(end)) {
	++result;
    }
    return result;
}


bool compositeAdd(SatAddr list, SatAddr obj)
{
    unsigned size = obNibbles(obj);
    SatAddr addr = list + compositeObNibbles(list) - 5;
    if (size == -1) return false;
    sat_moveup(obj, addr, size);
    sat_poke(addr+size, SAT_SEMI, 5);
    return true;
}


SatAddr compositeCat(SatAddr src, SatAddr obj)
{
    unsigned objSize = obNibbles(obj);
    unsigned srcSize = compositeObNibbles(src);
    SatAddr result;

    if (objSize == -1) return 0; /* not an object */
    result = sat_createtemp(srcSize+objSize);
    if (result == 0) return 0;

    sat_poke(result, obProlog(src), 5);

    // Populate the data portion from src (everything minus the prolog
    // and epolog)
    sat_moveup(src+5, result+5, srcSize-10);

    // add the new object
    sat_moveup(obj, result+srcSize-5, objSize);

    // finally add the composite
    sat_poke(result+srcSize+objSize-5, SAT_SEMI, 5);
    return result;
}


SatAddr compositeMerge(SatAddr comp1, SatAddr comp2)
{
    unsigned size1 = compositeObNibbles(comp1);
    unsigned size2 = compositeObNibbles(comp2);
    SatAddr result = sat_createtemp(size1+size2-5);	/* only need 1 composite */
    if (result == 0) {
	return 0;
    }
    sat_poke(result, obProlog(comp1), 5);
    sat_moveup(comp1+5, result+5, size1-10);

    // Now copy the data portion of l2 and l2's composite
    sat_moveup(comp2+5, result+size1-5, size2-5);
    return result;
}


