#include <hpgcc49.h>
#include "hpobjects.h"

// Directory object. This is complex. See section 57.8 of "Saturn Assembly Language Programming."
// And see the very important notes below.

// There are two type of directories, the HOME directory and Subdirectories.

// Home Directory is:
// Prolog
// number of libraries (3 nibbles)
// LIBRARY DESCRIPTION 1
// LIBRARY DESCRIPTION 2
// ...
// LIBRARY DESCRIPTION n
// Offset to "Name length A" field of last ENTRY, or zero if no entries
// ENTRY1
// ENTRY2
// ...
// ENTRYn

// A LIBRARY DESCRIPTION is 13 nibbles:
// 	library number (3 nibbles)
// 	addr of hash table (5 nibbles)
// 	addr of message table (5 nibbles)
// QUESTION: what's up with the note at the bottom of pg 124? I dont understand this

// QUESTION: subdirs have a single library number in place of the
// entire libary part. Where does it find the hash table and the
// message table?
//
// An ENTRY is:
//     - Offset to "Name Length A" field of previous entry, or zero.
//       "Saturn Assembly Language Programming" calls this the "size
//       of object N zone", but it's really the size minus 5.  I find
//       it much easier to think of it as the offset to Name Len A
//	 because it's consistent with the offset at the beginning of
//	 the dir.
//	- Name Length A: the size (in bytes) of the name (2 nibbles)
//	- The name of the directory entry, stored one byte per character
// 	- Name Length B: same as Name Length A.
//	- The object of the directory entry.

// Important Notes:
// This is organized a little differently than how the book presents
// it.  The book groups the "size of object N zone" with the object
// before it, but such a grouping results in two "special cases:"
//  1. The "5 nibbles of all zeros" is not present in an empty
//     directory.
//  2. The "size of object N zone" is not present in the last entry.
// By considering this field as part of the entry AFTER it rather than
// the entry BEFORE it, each entry has the same format.


// Example 1: An empty dir (DIR END) looks like this
// 69A20 prolog
// FF7   lib
// 00000 offset

// 
// Example2: "DIR abc 1. END" looks like this:
// 
// 69A20 PROLOG
// FF7   Lib
// A0000 offset
// 00000 zero (length of previous entry)
// 30    name legth A
// 14 24 34 "ABC"
// 30    name length B
// 33920  REAL prolog
// 000 000000000001 0   1.0
// 
// Example 3: "DIR  1. DEFG 1.23456789012 END"
// 
// 69a20                 DORRP prolog
// ff7                   lib
// 03000                 offset to "name len A" of last object

//                       START OF OBJECT 1
// 00000                 size of previous
// 40                    name len A
// 44 54 64 74           "DEFG"
// 40                    name len B
// 33920                 object 1 prolog (DOREAL)
// 000 210987654321 0    1.23456789012
// 
//                       START OBJECT 2
// 12000		 size of previous object
// 30                    name len A
// 14 24 34              "ABC"
// 30                    name len B
// 33920                 Prolog of obj2 (REAL)
// 000 000000000001 0    1.0

// Given a directory, return the address of the offset to the last
// entry.  This is used in a couple of different places.
extern SatAddr offsetToLast(SatAddr dir);
extern bool isDirEntry(SatAddr entry);

//++
// Given a directory, return the address of the offset to the last
// entry.  This is used in a couple of different places.
SatAddr offsetToLast(SatAddr dir)
{
    SatAddr result;
    // The answer is different for the HOME directory vs. subdirs
    // because HOME contains multiple libraries.
    if (dir == getHOME()) {
	int numLibs = sat_peek(dir+5, 3);
	result = dir + 5 + 3 + numLibs*13;
    } else {
	result = dir + 5 + 3;
    }
    return result;
}


// Free a directory list.
void freeDirItem(struct DirItem *head)
{
    struct DirItem *next;
    while (head) {
	next = head->next;
	free(head->name);
	free(head);
	head = next;
    }
}

// Decode the names and objects in a directory. They will be in the
// list in the same order that they show on the calculator.
struct DirItem *DIRdecode(SatAddr src)
{
    struct DirItem *result = NULL;
    struct DirItem *curLink;
    SatAddr cur;
    // Let DIRfirstEntry() determine if it's a directory
    for (cur = DIRfirstEntry(src); cur; cur = DIRnextEntry(cur)) {
	curLink = malloc(sizeof(struct DirItem));
	curLink->next = result;
	result = curLink;
	curLink->name = malloc(DIRentryNameLen(cur)+1);
	DIRdecodeEntry(cur, curLink->name, &curLink->obj);
    }
    return result;
}


// Create an empty directory at dst, which must have room for it.
void DIRinit(SatAddr dst)
{
    sat_poke(dst, SAT_DORRP, 5);
    sat_poke(dst+5,0x7ff, 3);
    sat_poke(dst+8, 0, 5);
}


// Create a new directory object with the objects in the DirItem
// list. Put it at dst (if set) or in tempob if dst is null. Return the DIR
SatAddr DIRencode(struct DirItem *head, SatAddr dst)
{
    struct DirItem *cur;

    if (dst) {
	DIRinit(dst);
    } else {
	// Compute the size
	int size = 5+3+5+5;  	       /* prolog, library, offset to last obj */
	// Now add space for each object
	for (cur = head; cur; cur = cur->next) {
	    if (!isOb(cur->obj)) return 0;
	    size += 5;			      /* size of previous object zone */
	    size += 2 + 2*strlen(cur->name)+2;/* name len A, name, name len B*/
	    size += obNibbles(cur->obj);	      /* object itself */
	}	
	dst = makeDIR(size);
    }

    for (cur = head; cur; cur = cur->next) {
	DIRadd(dst, cur->name, cur->obj);
    }
    return dst;
}


// Create an empty DIR object in a chuck of memory "nibbles" nibbles long.
// The DIR will be in the subdirectory format.
SatAddr makeDIR(int nibbles)
{
    SatAddr result = sat_createtemp(nibbles);
    DIRinit(result);
    return result;
}


// Add an entry to a DIR. The DIR must have enough space for the new
// entry.
bool DIRadd(SatAddr src, const char *name, SatAddr obj)
{
    SatAddr addr;
    SatAddr offsetAddr;		/* address of offset member */
    int offset;
    int nameLen;			 /* size of the name */
    int size;

    if (!isDIR(src)) return false;
    if (!isOb(obj)) return false;
    offsetAddr = offsetToLast(src);
    offset = sat_peek(offsetAddr, 5);
    nameLen = strlen(name);

    // Set size to the size of the last object zone
    // and set addr to the nibble after the end of the directory
    if (offset) {
	addr = offsetAddr+offset;
	size = 4 + 2*sat_peek(addr,2);		 /* name and 2 copies of name len */
	size += obNibbles(addr+size);		 /* object */
	addr += size;				 /* point to the end */
    } else {
	addr = offsetAddr+5;
	size = 0;
    }
    sat_poke(offsetAddr, addr+5-offsetAddr, 5);  /* the new offset to last object */
    sat_poke(addr, size, 5);			 /* size of previous obj */
    sat_poke(addr+5, nameLen, 2);		 /* name len A */
    sat_poke_nibbles(addr+7, name, nameLen*2);	 /* name */
    sat_poke(addr+7+2*nameLen, nameLen, 2);	 /* name len B */
    sat_moveup(obj, addr+9+2*nameLen, obNibbles(obj)); /* object */
    return true;
}

bool isDIR(SatAddr src)
{
    return (obProlog(src) == SAT_DORRP);
}



//
// ITERATION.
// A directory contains a bunch of "entries" that consist
// of name/value pairs.  Use DIRfirstEntry() and DIRnextEntry() to go
// through the entries.  Then you can use DIRdecode() to extract the
// name and object within each entry.
//



// Basic check to see if an address looks like the Name Length A field
// of a directory entry
// 
bool isDirEntry(SatAddr entry)
{
    int numBytes = sat_peek(entry,2);
    SatAddr nameLenB = entry + 2 + 2*numBytes;
    SatAddr prolog = sat_peek(nameLenB+2, 5);

    if (numBytes != sat_peek(nameLenB,2)) {
	return false;
    }
    if (sat_peek(prolog, 5) != SAT_PRLG) {
	return false;
    }
    return true;
}

// Return the address of the first entry in the directory, or zero if
// there are none.
SatAddr DIRfirstEntry(SatAddr src)
{
    // This returns the address of the the "name length A" field
    SatAddr result;
    int offset;

    if (!isDIR(src)) return 0;
    result = offsetToLast(src);
    offset = sat_peek(result, 5);
    return (offset ? result + offset : 0);
}


// Return the next entry in a directory, or zero if there are no
// more. The parameter "entry" must be the return value from a
// previous call to DIRfirstEntry or DIRnextEntry.
SatAddr DIRnextEntry(SatAddr entry)
{
    int offset;
    if (!isDirEntry(entry)) return 0;
    entry -= 5;		/* backup to size of previous object */
    offset = sat_peek(entry,5);
    return (offset ? entry - offset : 0);
}


// Given the address of a directory entry, extract the name and
// object.  "entry" points to a directory entry returned from
// DIRfirstEntry() or DIRnextEntry. If "name" is non-null, then it
// will be filled with the NULL-terminated name of the directory
// entry. It must be big enough to hold the name and terminating NULL
// (256 bytes will do it).  If "obj" is non-null, then it will be
// populated with the address of the object in the directory.
bool DIRdecodeEntry(SatAddr entry, char *name, SatAddr *obj)
{
    int size;
    if (!isDirEntry(entry)) return false;
    size = DIRentryNameLen(entry);
    if (name) {
	sat_peek_sat_bytes(name, entry+2, size);
	name[size] = 0;
    }
    if (obj) {
	*obj = entry + 2*size + 4;
    }
    return true;
}


// Given a pointer to a directory entry, return the number of
// characters in the entry name.
int DIRentryNameLen(SatAddr entry)
{
    if (!isDirEntry(entry)) return 0;
    return sat_peek(entry,2);
}



// Find the named entry in the directory and return the address of the
// entry, or zero if not found.  Note that this returns the address of
// the directory entry, not the object associated with the name. Use
// DIRfindObjByName to find the object. Unlike RCL, this searches the
// current directory only.
//
SatAddr DIRfindEntryByName(SatAddr dir, const char *name)
{
    SatAddr entry;
    char entryName[256];
    for (entry = DIRfirstEntry(dir); entry; entry = DIRnextEntry(entry)) {
	DIRdecodeEntry(entry, entryName, NULL);
	if (strcmp(name, entryName) == 0) {
	    return entry;
	}
    }
    return 0;
}

// Find an entry in the directory from a global IDNT object.
// Return the entry or 0. Unlike RCL, this searches the current
// directory only.
SatAddr DIRfindEntryByIdnt(SatAddr dir, SatAddr idnt)
{
    char entryName[256];
    if (IDNTdecode(idnt, entryName) == NULL) {
	return 0;		       /* not an idnt */
    }
    return DIRfindEntryByName(dir, entryName);
}


// Find the object associated with the name in a directory.
SatAddr DIRfindObByName(SatAddr dir, const char *name)
{
    SatAddr entry = DIRfindEntryByName(dir, name);
    if (entry == 0) {
	return 0;
    }
    return entry + 4 + 2 * DIRentryNameLen(entry);
}

// Find the object associated with the name in a directory.
SatAddr DIRfindObByIdnt(SatAddr dir, SatAddr idnt)
{
    char entryName[256];
    if (IDNTdecode(idnt, entryName) == NULL) {
	return 0;		       /* not an ident */
    }
    return DIRfindObByName(dir, entryName);
}

// MODIFIERS
// 

// Change the variable "name" to obj. An object for the name must
// exist already and obj must be the same size as it.  Use with
// caution since any pointers to this object (like those on the stack)
// will pick up the new value.
// Return values:
// 0 : success
// -1: no entry for "name" in dir.
// -2: entry for "name" is not the same size as obj
int DIRchangeByName(SatAddr dir, const char *name, SatAddr obj)
{
    SatAddr existing = DIRfindObByName(dir, name);
    if (existing == 0) {
	return -1;
    }
    if (!obCopy(obj, existing)) {
	return -2;
    }
    return 0;
}


// Same as DIRchangeByName, but uses a global identifier object to
// lookup the name
int DIRchangeByIdnt(SatAddr dir, SatAddr idnt, SatAddr obj)
{
    char entryName[256];
    if (!IDNTdecode(idnt, entryName)) {
	return -1;
    }
    return DIRchangeByName(dir, entryName, obj);
}


// Given a directory entry, return a pointer to the directory
// enclosing it.
SatAddr DIRdirForEntry(SatAddr entry)
{
    SatAddr addr;
    SatAddr homeDir;
    // Back up to the first entry
    if (!isDirEntry(entry)) return 0;
    while ((addr = DIRnextEntry(entry))) {
	entry = addr;
    }

    // The encoding for the home directory and subdirs are different.
    // For a subdir, the directory starts 18 nibbles before the
    // first entry, but for the HOME dir, you don't know where it
    // starts.  So we'll go the other way: compute where the first
    // entry for the home directory stars and see if that's where we are.
    homeDir = getHOME();
    if (entry == DIRfirstEntry(homeDir)) {
	return homeDir;
    } else {
	return entry - 18;
    }
}


// Given the address of a directory, return the address of the parent
// directory, or zero if dir is HOME.
SatAddr DIRparentDir(SatAddr dir)
{
    SatAddr addr;
    SatAddr result;

    if (!isDIR(dir)) return 0;

    if (dir == getHOME()) {
	return 0;
    }

    // Dir isn't the home directory, so it must be a subdir.  that
    // means that the directory is itself the object for an entry in a
    // directory.  We'll find this dir's entry in the parent directory
    // and then use DIRparentDir() to get the parent directory itself.

    // Immediately after dir is a 5-nibble value that gives the size
    // of dir's entry in the parent directory.  Find that entry
    addr = dir-2;	/* point to Name Len B */
    addr -= 2*sat_peek(addr,2);	/* back up over the name */
    addr -= 2;			/* back up to start of Name Len A */

    // Now get the directory itself.
    result = DIRdirForEntry(addr);
    return result;
}

// Return the address of the home directory
SatAddr getHOME()
{
    return sat_peek(SAT_USEROB, 5);
}

// Return address of current directory.
SatAddr getCWD()
{
    return sat_peek(SAT_CONTEXT, 5);
}


// Size of a DIR object in nibbles
int DIRnibbles(SatAddr dir)
{
    // walk "addr" down to the end of the directory, then return addr-dir
    SatAddr addr;
    int offset;

    if (!isDIR(dir)) return -1;

    addr = offsetToLast(dir); /* get the "offset to last" field */
    offset = sat_peek(addr, 5); /* get the offset */
    if (offset == 0) {
	// empty directory. Skip offset
	addr += 5;
    } else {
	addr += offset;		/* skip to last entry */
	DIRdecodeEntry(addr, NULL, &addr); /* skip to obj */
	addr = skipOb(addr);	/* skip the object */
    }
    return addr - dir;
}
