#include <hpgcc49.h>
#include "hpobjects.h"

/////////////////////////////////////////////////////////////////////
//
//                         EREL
//
/////////////////////////////////////////////////////////////////////

// Like a REAL, but 15 digits and 5 digit exponent
// 5 nibble prolog
// 5 nibbles exponent
// 15 nibble mantissa
// 1 nibble sign
//++
bool isEREL(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    return (prolog == SAT_DOEREL);
}

// Convert an HP EREL object to a C double. Returns true of false.
bool ERELdecode(SatAddr src, double *dst)
{
    if (!isEREL(src)) return false;
    *dst = ereal2double(src+5);
    return true;
}


// Convert an extended real to double.  Note that src points
// directly to the number, not the prolog of an EREL.
double ereal2double(SatAddr src)
{

#define MAX_IEEE754_EXPO 307

    int expo;
    double val;
    long long mantissa = sat_BCD2LONGLONG(src+5, 15);
    int sgn = sat_peek(src+5+15, 1);

    expo = (int) sat_BCD2LONGLONG (src, 5);
    if (expo > 49999)
	expo -= 100000;

    // Check for IEEE-754 exponent over/underflow...

    if (expo > MAX_IEEE754_EXPO) {
	// return +/- infinity
	return (sgn ? -1.0/0.0 : 1.0/0.0);
    } else if(expo < -MAX_IEEE754_EXPO) {
	// Treat this number as zero...
	return 0.0;
    }

    val = ipow(10.0,expo-14) * mantissa /* mantissa */;
    return  (sgn ? -val : val);
}


// Convert a C double (d) to an HP EREL object at Saturn address dst.
// returns true on success otherwise false.
// 
SatAddr ERELencode(double d, SatAddr dst)
{
    if (dst == 0) {
	dst = makeEREL();
    }
    sat_poke(dst, SAT_DOEREL, 5);
    double2hpereal(d, dst+5);
    return dst;
}

// Create an uninitialized HP real object in temp space and returns
// its address.
SatAddr makeEREL()
{
    SatAddr result = sat_createtemp(26);
    sat_poke(result, SAT_DOEREL, 5);
    return result;
}


//
// Convert a C double to an HP 84 bit ereal number. Note that this
// does NOT convert it to the 28 nibble DOEREL object, it just converts
// the binary formats.  This is the inverse of ereal2double(),
// except it always writes to the Saturn address space
void double2hpereal(double number, SatAddr dst)
{
    int exponent;
    double m;			/* normalized mantissa */
    LONGLONG mantissa;

    doubledecode(number, &exponent, &m);

    // DMH - not sure why it's doing this since the IEEE double can never
    // have an exponent outside this range.
    if(exponent<-49999){ //make sure we're not outside the HP's range
	exponent=0;
	mantissa=0;
    } else if(exponent>49999) {
	exponent=49999;
	mantissa = 999999999999999ll;
    } else {
	// scale it to a 15 digit mantissa and round it off.
	m *= 1.0E14;
	if (m<0.0) {
	    m -= 0.5;
	} else {
	    m += 0.5;
	}
	mantissa = m;
    }

	
    if (mantissa < 0) {
	mantissa = -mantissa;	/* get absolute value */
    }

    if (exponent<0){
	exponent = 100000-abs(exponent); /* this is how negative exponents are stored */
    }

    //========================

    //Here we've broken the double into a suitable mantissa and exponent
    //now we have to convert it into HP's memory format.

    /*

    ereal format:      EEEE EMMM MMMM MMMM MMMM S
    E is exponent. M is mantissa. S is sign (0 for positive, 9 for negative
    1.23456789012E12345 =  5432 1000 2109 8765 4321 0
    */
    ll2bcd(exponent, dst, 5);
    ll2bcd(mantissa, dst+5, 15);
    sat_poke(dst+20, (number < 0 ? 9 : 0), 1);
}




