#include <hpgcc49.h>
#include "hpobjects.h"

//
// GROB routines
//

// A grob is:
// Prolog SAT_DOGROB
// Size ( 5 nibbles, size of object minus size of prolog)
// rows ( 5 nibbles giving the number of rows)
// width (5 nibbles giving the width of each row in nibbles)
// rows*width nibbles of data

int GROBnumBytes(int rows, int columns);


//++
/* is it an GROB? */
bool isGROB(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    return (prolog == SAT_DOGROB);
}

// Return the number of grob data bytes used to encode a GROB with the
// given number of rows and columns.  Each row in encoded with an
// whole number of bytes.
int GROBnumBytes(int rows, int columns)
{
    int result = (columns+7)/8 * rows;
    return result;
}

int GROBnibbles(SatAddr src)
{
    int result;
    if (!isGROB(src)) return -1;
    result = sat_peek_sat_addr(src+5) + 5;
    return result;
}

// Decode the grob at src.  If "rows" is non-null, it will be
// populated with the number of rows. If "columns" is non-null it will
// be populated with the number of columns. If "pixels" is non-null it
// will get the pixel data and must contain enough space for the
// pixels. The space requred is available via the GROBbytes() function.
bool GROBdecode(SatAddr src, int *rows, int *columns, char *pixels)
{
    if (!isGROB(src)) return false;
    if (rows) {
	*rows = sat_peek(src+10,5);
    }
    if (columns) {
	*columns = sat_peek(src+15,5);
    }
    if (pixels) {
	sat_peek_sat_bytes(pixels, src+20, GROBbytes(src));
    }
    return true;
}


// Create a GROB object with "rows" x "columns" pixels. The pixel data
// in in "pixels", or if "pixels" is null, it's filled with zeros.  If
// "dst" is non-zero then the GROB will be placed there in which case
// it must contain enough space to hold the GROB.  Otherwise a new
// GROB is created.  In either case the GROB is returned.
SatAddr GROBencode(int rows, int columns, char *pixels, SatAddr dst)
{
    int bytes = GROBnumBytes(rows, columns);
    if (dst == 0) {
	dst = sat_createtemp(20+2*bytes);
    }
    sat_poke(dst, SAT_DOGROB, 5);
    sat_poke(dst+5, 2*bytes+15, 5);
    sat_poke(dst+10, rows, 5);
    sat_poke(dst+15, columns, 5);
    if (pixels) {
	sat_poke_nibbles(dst+20, pixels, bytes*2);
    } else {
	int i;
	for (i=0; i < bytes; ++i) {
	    sat_poke(dst+20+i, 0, 1);
	}
    }
    return dst;
}


// Return the number of bytes of graphics data in the GROB at src.
// Returns -1 on error.
int GROBbytes(SatAddr src)
{
    if (!isGROB(src)) return -1;
    return GROBnumBytes(sat_peek(src+10,5), sat_peek(src+15,5));
}

SatAddr makeGROB(int rows, int columns)
{
    return GROBencode(rows, columns, NULL, 0);
}

// Set the pixel to 1 or zero depending on the value of "val"
bool GROBsetPixel(SatAddr src, int row, int col, bool val)
{
    int numRows, numCols;	/* number of rows & cols in GROB */
    int pixelNumber;		/* pixel number to modify */
    SatAddr addr;		/* address of pixel's nibble */
    int nibble;			/* the nibble containing the pixel */
    int mask;			/* bit mask for the pixel being
				   changed within nibble */

    if (!isGROB(src)) return false;
    numRows = sat_peek(src+10, 5);
    numCols = sat_peek(src+15, 5);
    if (row<0 || row>=numRows) return 0;
    if (col<0 || col>=numCols) return 0;

    // Get the actual number of columns in the grob.
    numCols = (numCols+7) / 8;
    numCols *= 8;

    // Now get the nibble that you're changing and the mask for the bit
    pixelNumber = row*numCols + col;
    addr = src+20+pixelNumber/4;
    mask = 1 << (pixelNumber & 3);
    nibble = sat_peek(addr,1);
    if (val) {
	nibble |= mask;
    } else {
	nibble &= ~mask;
    }
    sat_poke(addr, nibble, 1);
    return true;
}


