#include <hpgcc49.h>
#include "hpobjects.h"

//
// HXS (hex string) routines
//

// A hex string (hxs) is:
// Prolog SAT_DOHXS
// Size ( 5 nibbles, size of object minus size of prolog)
// hex digits, lsb first


//++

/* is it an HXS? */
bool isHXS(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    return (prolog == SAT_DOHXS);
}

// Return the size in nibbles of the HXS object at src.
int HXSnibbles(SatAddr src)
{
    int result;
    if (!isHXS(src)) return -1;
    result = sat_peek_sat_addr(src+5) + 5;
    return result;
}


// Convert HP HXS to C long long . No type checking is done on src.
// This assumes that src is short enough to fit
bool HXSdecodell(SatAddr src, unsigned long long *dst)
{
    int i;
    int digits;
    unsigned long long val;

    if (!isHXS(src)) return false;
    digits = sat_peek_sat_addr(src+5) - 5;
    if (digits > 16) return false;

    val = 0;
    for (i = digits - 1; i >= 0; i--) {
	val = (val << 4) | sat_peek(src+10+i,1);
    }
    *dst = val;
    return true;
}


// Convert a longlong to HXS
SatAddr HXSencodell(unsigned long long src, SatAddr dst)
{
    int digits = 0;
    unsigned long long tmp = src;
    int i;

    while (tmp) {
	++digits;
	tmp >>= 4;
    }
    if (dst==0) {
	dst = sat_createtemp(10+digits);
    }
    sat_poke(dst, SAT_DOHXS, 5);
    sat_poke(dst+5, 5+digits, 5);
    for (i=10, tmp = src; tmp; tmp >>= 4) {
	sat_poke(dst + i++, tmp&0xf, 1);
    }
    return dst;
}


// Decode a HXS. If "buf is non-null, then the hex string will be
// written to buf, which must contain sufficient space to hold it. If
// "nibbles" is non-null, then it will be populated with the size of
// the hex string in nibbles. Note that this is the size of the hex
// string date itself and excludes the prolog and "size" fields of the
// HXS object that contains it.
bool HXSdecode(SatAddr src, char *buf, int *nibbles)
{
    if (!isHXS(src)) return false;
    if (buf) {
	sat_peek_sat_bytes(buf, src+10, HXSbytes(src));
    }
    if (nibbles) {
	*nibbles = sat_peek(src+5, 5) - 5;
    }
    return true;
}


// Encode "nibbles" nibbles beginning at "buf" into a hex string at
// dst.  If dst is zero then create a new HXS. Returns the resulting
// HXS.
SatAddr HXSencode(char *buf, int nibbles, SatAddr dst)
{
    if (dst == 0) {
	dst = makeHXS(nibbles);
    }
    sat_poke(dst, SAT_DOHXS, 5);
    sat_poke(dst+5, 5+nibbles, 5);
    sat_poke_nibbles(dst+10, buf, nibbles);
    return dst;
}



// Given a HXS, return the number of bytes needed to hold the data within it.
int HXSbytes(SatAddr src)
{
    int result;
    if (!isHXS(src)) return -1;

    // The number of data nibbles is sat_peek(src+5)-5;
    // If there are N nibbles, then you need (N+1)/2 bytes.
    // Simplifying you get...
    result = (sat_peek(src+5, 5)-4)/2;
    return result;
}



// create a HXS in HP address space with room for the given number of nibbles
// and return it.  Contents of the HXS are undefined.
SatAddr makeHXS(int nibbles)
{
    SatAddr result;
    if ((result = sat_createtemp(10+nibbles))) {
	sat_poke(result, SAT_DOHXS, 5);
	sat_poke(result+5, nibbles+5, 5);
	sat_memset(result+10, 0, nibbles-10);
    }
    return result;
}

