#include <hpgcc49.h>
#include "hpobjects.h"
#include "LAMIDNT.h"

/////////////////////////////////////
//   LAM and IDNT
/////////////////////////////////////


//++
int LAMIDNTnumChars(SatAddr src)
{
    unsigned result = sat_peek(src+5, 2);
    return result;
}


// An identifier is a prolog (5 nibbles)

// Then 2 nibbles of size.  This gives the number of bytes (not
// nibbles) that follow.
// Then SIZE bytes, each containing the character.
//
// Note that the size field is quite different from STRINGs.
// Here, the size is the number of BYTES in the name.  In a STRING,
// it's the number of NIBBLES excluding the prolog.

// Is the "src" address a IDNT
bool isIDNT(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    bool result = (prolog == SAT_DOIDNT);
    return result;
}

// Is the "src" address a LAM
bool isLAM(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    bool result = (prolog == SAT_DOLAM);
    return result;
}

// Convert from an HP IDNT or LAM pointed at by "src" into a null-terminated
// string located in the buffer pointed to by dst.  Dst must have
// enough space for the string and the null terminated buffer.
char * LAMIDNTtoStr(SatAddr src, char *dst)
{
    int chars = LAMIDNTnumChars(src);
    if (dst==0) {
	dst = malloc(chars+1);
    }
    sat_peek_sat_bytes(dst, src+7, chars);
    dst[chars] = 0;
    return dst;
}


// Convert from an HP IDNT pointed at by "src" into a null-terminated
// string located in the buffer pointed to by dst.  Dst must have
// enough space for the string and the null terminated buffer.
char *IDNTdecode(SatAddr src, char *dst)
{
    if (!isIDNT(src)) return NULL;
    return LAMIDNTtoStr(src, dst);
}

// Same as IDNT2str, but for a LAM
char *LAMdecode(SatAddr src, char *dst)
{
    if (!isLAM(src)) return NULL;
    return LAMIDNTtoStr(src, dst);
}

// Convert from a NULL-terminated string to an HP IDNT or HP LAM.
// "dst" must have enough room to hold the identifier. This writes
// all data for the object, including the prolog and size fields
SatAddr LAMIDNTstr2(const char *src, SatAddr dst, int prolog)
{
    int bytes = strlen(src);
    if (dst == 0) {
	int nibbles = bytes * 2 + 7;
	dst = sat_createtemp(nibbles);
    }
    sat_poke(dst, prolog, 5);	               /* prolog */
    sat_poke(dst+5, bytes, 2);		       /* size */
    sat_poke_nibbles(dst+7, src, bytes*2);     /* data */
    return dst;
}

SatAddr IDNTencode(const char *src, SatAddr dst)
{
    return LAMIDNTstr2(src,dst,SAT_DOIDNT);
}

SatAddr LAMencode(const char *src, SatAddr dst)
{
    return LAMIDNTstr2(src,dst,SAT_DOLAM);
}

/* Allocate space for an IDNT or LAM with numChars characters, and put
   an empty IDNT at the beginning of the space. Return the address of
   the IDNT. You can add data to the identifier with IDNTadd(), as
   long as the total number of characters added doesn't exceed
   numChars. */
SatAddr LAMIDNTmake(size_t size, int prolog)
{
    int nibbles = size * 2 + 7;
    SatAddr result = sat_createtemp(nibbles);
    sat_poke(result, prolog, 5);	  /* prolog */
    sat_poke(result+5, 0, 2);		  /* size. This one is empty */
    return result;
}

SatAddr makeIDNT(size_t size)
{
    return LAMIDNTmake(size, SAT_DOIDNT);
}

SatAddr makeLAM(size_t size)
{
    return LAMIDNTmake(size, SAT_DOLAM);
}


int IDNTnumChars(SatAddr src)
{
    if (!isIDNT(src)) return -1;
    return LAMIDNTnumChars(src);
}
int LAMnumChars(SatAddr src)
{
    if (!isLAM(src)) return -1;
    return LAMIDNTnumChars(src);
}


// Return the size in nibbles of the IDNT object at src.  Returns -1
// if the object isn't a identifier
int LAMIDNTnibbles(SatAddr src)
{
    int result = 2 * LAMIDNTnumChars(src) + 7;
    return result;
}

int IDNTnibbles(SatAddr src)
{
    if (!isIDNT(src)) return -1;
    return LAMIDNTnibbles(src);
}
int LAMnibbles(SatAddr src)
{
    if (!isLAM(src)) return -1;
    return LAMIDNTnibbles(src);
}


