#include <hpgcc49.h>
#include "hpobjects.h"

/////////////////////////////////////////////////////////////////////
// LNGCOMP (infinite precision complex) routines

// An infinite precision complex is a prolog followed by the bodies of
// two infinite precision reals.  The first is the real part and the
// second is the imaginary part.


bool lngrealDecodell(SatAddr src, long long *mantissa, long long *exponent);
extern bool ZINTdecodell(SatAddr src, long long *dst);
extern bool lngrealDecoded(SatAddr src, double *d);
extern bool lngrealEncoded(double d,  SatAddr src);
extern void doubledecodell(double number, long long *mantissa, long long *exponent);
extern bool lngrealDecodeStr(SatAddr src, char *mantissa, char *exponent);
extern int lngrealNibbles(SatAddr src);
extern bool lngrealBytes(SatAddr src, int *mBytes, int *eBytes);

extern SatAddr LNGCOMPimAddr(SatAddr src);

//++


// Given the address of a LNGCOMP, return the address 5 nibbles before the 
// size of the imaginary mantissa.  The imaginary part looks just like
// a long real starting at this address, except that the prolog isn't
// there.  Nevertheless, we can use many of the LNGREAL functions to
// operate on the imaginary part this way.
SatAddr LNGCOMPimAddr(SatAddr src)
{
    SatAddr result = src + lngrealNibbles(src)-5;
    return result;
}

bool isLNGCOMP(SatAddr src)
{
    return (obProlog(src) == SAT_DOLNGCOMP);
}

bool LNGCOMPdecoded(SatAddr src, double *re, double *im)
{
    bool result = isLNGCOMP(src);
    result = result && lngrealDecoded(src, re);
    result = result && lngrealDecoded(LNGCOMPimAddr(src), im);
    return result;
}


// Convert the doubles to a LNGCOMP. re and im are the real and imaginary
// parts of the number to convert.  If dst is zero then a new LNGCOMP will be
// allocated.  Otherwise it must point to enough space to hold the
// LNGCOMP.  Returns the SatAddr of the LNGCOMP.
SatAddr LNGCOMPencoded(double re, double im, SatAddr dst)
{
    long long reEx, imEx;
    long long reMant, imMant;
    extern char * ll2s(long long);

    doubledecodell(re, &reMant, &reEx);
    doubledecodell(im, &imMant, &imEx);
    dst = LNGCOMPencodell(reMant, reEx, imMant, imEx, dst);
    return dst;
}


// Decode a LNGCOMP into mantissa and exponent parts. No over flow checking
// is done.  Use LNGCOMPdigits() to check the number of digits.
bool LNGCOMPdecodell(SatAddr src,
		     long long *re_mantissa, long long *re_exponent,
		     long long *im_mantissa, long long *im_exponent)
{
    bool result = isLNGCOMP(src);
    result = result && lngrealDecodell(src, re_mantissa, re_exponent);
    result = result && lngrealDecodell(LNGCOMPimAddr(src), im_mantissa, im_exponent);
    return result;
}


// Encode a LNGCOMP from long long mantissa and exponent values of
// real & imaginary parts If dst is zero, then a new LNGCOMP will be
// allocated to hold the value.  Otherwise it must point to enough
// space to hold it.  In either case the resulting LNGCOMP will be
// returned
SatAddr LNGCOMPencodell(long long re_mantissa, long long re_exponent,
			long long im_mantissa, long long im_exponent,
			SatAddr dst)
{
    SatAddr im;
    int old;

    if (dst == 0) {
	int mReDigits = llDigits(llabs(re_mantissa));
	int eReDigits = llDigits(llabs(re_exponent));
	int mImDigits = llDigits(llabs(im_mantissa));
	int eImDigits = llDigits(llabs(im_exponent));
	dst = makeLNGCOMP(mReDigits, eReDigits, mImDigits, eImDigits);
    }
    // Encode using LNGREALencode and then clean up
    if (!LNGREALencodell(re_mantissa, re_exponent, dst)) return 0;
    im = LNGCOMPimAddr(dst);
    old = sat_peek(im, 5);
    LNGREALencodell(im_mantissa, im_exponent, im);
    sat_poke(im, old, 5);
    sat_poke(dst, SAT_DOLNGCOMP, 5);
    return dst;
}



// Populate a LNGCOMP from null-terminated strings representing the
// mantissa and exponent.  If dst is zero then space will be allocated
// to hold the LNGCOMP.  Otherwise it must contain enought space for
// it.  The resulting LNGCOMP is returned.
SatAddr LNGCOMPencodeStr(const char *re_mantissa, const char *re_exponent,
			 const char *im_mantissa, const char *im_exponent,
			 SatAddr dst)
{
    bool mPositive, ePositive;
    int mReDigits, eReDigits, mImDigits, eImDigits;
    SatAddr im;
    int old;

    if (dst == 0) {
	cleanupZINTstr(re_mantissa, &mReDigits, &mPositive);
	cleanupZINTstr(re_exponent, &eReDigits, &ePositive);
	cleanupZINTstr(im_mantissa, &mImDigits, &mPositive);
	cleanupZINTstr(im_exponent, &eImDigits, &ePositive);
	dst = makeLNGCOMP(mReDigits, eReDigits, mImDigits, eImDigits);
    }
    LNGREALencodeStr(re_mantissa, re_exponent, dst);
    im = LNGCOMPimAddr(dst);
    old = sat_peek(im, 5);
    LNGREALencodeStr(im_mantissa, im_exponent, im);
    sat_poke(im, old, 5);
    sat_poke(dst, SAT_DOLNGCOMP, 5);
    return dst;
}


SatAddr LNGCOMPdecodeStr(SatAddr src,
			 char *re_mantissa, char *re_exponent,
			 char *im_mantissa, char *im_exponent)
{
    if (!isLNGCOMP(src)) return false;
    if (!lngrealDecodeStr(src, re_mantissa, re_exponent)) return false;
    if (!lngrealDecodeStr(LNGCOMPimAddr(src), im_mantissa, im_exponent)) return false;
    return true;
}


// Make a LNGCOMP with room for the number of digits in the different parts
// given in the parameters. Populate it with zero

SatAddr makeLNGCOMP(int numReMdigits, int numReEdigits,
		    int numImMdigits, int numImEdigits)
{
    // 5 for the prolog. Then each digit as a size field (5), the
    // digits, and a sign field (1)
    SatAddr result = sat_createtemp(5+
				    5+numReMdigits+1+
				    5+numReEdigits+1+
				    5+numImMdigits+1+
				    5+numImEdigits+1);
    return result;
}



// Number of nibbles taken up by the object
int LNGCOMPnibbles(SatAddr src)
{
    int result;
    if (!isLNGCOMP(src)) return -1;
    result = lngrealNibbles(src);
    result += lngrealNibbles(LNGCOMPimAddr(src));
    result -= 5;		/* double counted prolog area of IM
				   part */
    return result;
}


// Get the number of digits in the mantissa and exponent.  If mdigits
// is non-null, it will be populated with the number of digits in the
// mantissa.  Edigits is the same, but uses the exponent.
bool LNGCOMPdigits(SatAddr src, int *mReDigits, int *eReDigits,
		   int *mImDigits, int *eImDigits)
{
    if (!isLNGCOMP(src)) return false;
    lngrealDigits(src, mReDigits, eReDigits);
    lngrealDigits(LNGCOMPimAddr(src), mImDigits, eImDigits);
    return true;
}

// Get the number of bytes required to store a null-terminated string
// for each part of the number.  If any pointer is null then no
// decoding is done for that part.
bool LNGCOMPbytes(SatAddr src, int *mReBytes, int *eReBytes,
		   int *mImBytes, int *eImBytes)
{
    if (!isLNGCOMP(src)) return false;
    lngrealBytes(src, mReBytes, eReBytes);
    lngrealBytes(LNGCOMPimAddr(src), mImBytes, eImBytes);
    return true;
}



