#include <hpgcc49.h>
#include "hpobjects.h"

/////////////////////////////////////////////////////////////////////
// LNGREAL (infinite precision real) routines

// An infinite precision real is just the bodies of two contiguous
// infinite precision integers (ZINTs). The mantissa first and then
// the exponent.  Here is the explanation of the ZINT. I'll use many
// of the zint functions here.

// A ZINT is:
// prolog (02614) 5 nibbles
// size (minus prolog? or number of digits?) 5 nibbles
// digit 1 (1 nibble)
// ...
// digit n (1 nibble)
// sign: 0/9 for +/- (1 nibble)
// 
// Note that zero has no sign nibble (or no digit nibbles, depending on how
// you look at it.  Since client code will likely want to know how many
// digits are needed to represent the number, the ZINTdigits() function
// returns 1 when the number encodes zero.


// We make heavy use of implementation of code in ZINT
extern bool zintDecodell(SatAddr src, long long *dst);
bool lngrealDecoded(SatAddr src, double *d);
char * zintDecodeStr(SatAddr src, char *str);
extern int zintDigits(SatAddr src);
extern int zintNibbles(SatAddr src);
extern int zintBytes(SatAddr src);

bool lngrealDecodeStr(SatAddr src, char *mantissa, char *exponent);
bool lngrealDecodell(SatAddr src, long long *mantissa, long long *exponent);
bool lngrealBytes(SatAddr src, int *mBytes, int *eBytes);

//++


// Is this a LNGREAL?
bool isLNGREAL(SatAddr src)
{
    return (obProlog(src) == SAT_DOLNGREAL);
}


//////////////////////////////////////////////////////////
//      Encode/decode to double
//////////////////////////////////////////////////////////

// Implementation of LNGREALdecoded(). This doesn't check
// the prolog and is used by several other objects.
bool lngrealDecoded(SatAddr src, double *d)
{
    long long mantissa, exponent;
    double result;

    if (!zintDecodell(src, &mantissa)) return false;
    if (!zintDecodell(src+zintNibbles(src)-5, &exponent)) return false;
    result = mantissa;
    result *= ipow(10, exponent);
    *d = result;
    return true;
 }

// Convert a LNGREAL to/from a double
bool LNGREALdecoded(SatAddr src, double *d)
{
    if (!isLNGREAL(src)) return false;
    return lngrealDecoded(src, d);
}

// Convert a double to a LNGREAL.  If dst is zero then a new LNGREAL
// will be created. Otherwise it must contain enough space for the
// object.  Return value is the LNGREAL
SatAddr LNGREALencoded(double src, SatAddr dst)
{
    long long exponent;
    long long mantissa;

    doubledecodell(src, &mantissa, &exponent);
    dst = LNGREALencodell(mantissa, exponent, dst);
    return dst;
}


//////////////////////////////////////////////////////////
//      Encode/decode to long long mantissa and exponent
//////////////////////////////////////////////////////////

// Decode a LNGREAL into mantissa and exponent. No over flow checking
// is done.  Use LNGREALdigits() to check the number of digits.  "mantissa"
// or "exponent" can be NULL, which means don't decode that part.
bool lngrealDecodell(SatAddr src, long long *mantissa, long long *exponent)
{
    if (mantissa) {
	if (!zintDecodell(src, mantissa)) return false;
    }
    if (exponent) {
	if (!zintDecodell(src+zintNibbles(src)-5, exponent)) return false;
    }
    return true;
 }

bool LNGREALdecodell(SatAddr src, long long *mantissa, long long *exponent)
{
    if (!isLNGREAL(src)) return false;
    return lngrealDecodell(src, mantissa, exponent);
}


// Encode a LNGREAL from long long mantissa and exponent values If Dst
// is zero then space will be allocated for the object. Otherwise Dst
// must point to enough space to hold the object.  The resulting
// LNGREAL is returned
SatAddr LNGREALencodell(long long mantissa, long long exponent, SatAddr dst)
{
    int mDigits = llDigits(llabs(mantissa));
    if (dst == 0) {
	int eDigits = llDigits(llabs(exponent));
	dst = makeLNGREAL(mDigits, eDigits);
    }
    // Jam them in using ll2ZINT and the clean up.
    ZINTencodell(exponent, dst+10+mDigits+1-5);
    ZINTencodell(mantissa, dst);	/* overwrites the DOZINT prolog created above */
    sat_poke(dst, SAT_DOLNGREAL, 5); /* fix the prolog */
    return dst;
}


//////////////////////////////////////////////////////////
//      Encode/decode to strings
//////////////////////////////////////////////////////////

// Get the mantissa and exponent of a long real expressed as null-terminated
// strings.  If the mantissa is null, then a new string is malloc'ed to
// hold it. Otherwise it must point to sufficient space for the
// string.  The exponent is similar for the exponent part.  Necessary
// size of the strings is available via LNGREALbytes()
bool lngrealDecodeStr(SatAddr src, char *mantissa, char *exponent)
{
    if (!zintDecodeStr(src, mantissa)) return false;
    if (!zintDecodeStr(src+zintNibbles(src)-5, exponent)) return false;
    return true;
 }

bool LNGREALdecodeStr(SatAddr src, char *mantissa, char *exponent)
{
    if (!isLNGREAL(src)) return false;
    return lngrealDecodeStr(src, mantissa, exponent);
}


// Populate a LNGREAL from null-terminated strings representing the
// mantissa and exponent.  If dst is zero then space will be allocated
// to hold the LNGREAL.  Otherwise it must contain enought space for
// it.  The resulting LNGREAL is returned.
SatAddr LNGREALencodeStr(const char *mantissa,
			 const char *exponent,
			 SatAddr dst)
{
    bool mPositive, ePositive;
    int mDigits, eDigits;

    mantissa = cleanupZINTstr(mantissa, &mDigits, &mPositive);
    exponent = cleanupZINTstr(exponent, &eDigits, &ePositive);

    if (dst == 0) {
	dst = makeLNGREAL(mDigits, eDigits);
    }
    ZINTencodeStr(exponent, dst+10+mDigits+1-5);
    ZINTencodeStr(mantissa, dst);
    sat_poke(dst, SAT_DOLNGREAL, 5);
    return dst;
}

// Make a LNGREAL with space for mDigits in mantissa and edigits in
// exponent.  Populate it with 0.0
SatAddr makeLNGREAL(int mDigits, int eDigits)
{
    SatAddr result = sat_createtemp(5+		 /* prolog */
				    5+		 /* mantissa size */
				    mDigits+1+   /* mantissa digits and sign */
				    5+eDigits+1	 /* exponent size, digits and sign */
				    );
    sat_poke(result, SAT_DOLNGREAL, 5);
    sat_poke(result+5, 6, 5);
    sat_poke(result+6, 0, 1);
    sat_poke(result+7, 6, 5);
    sat_poke(result+13, 0, 1);
    return result;
}



// Number of nibbles taken up by the object
int lngrealNibbles(SatAddr src)
{
    int mSize;
    int eSize;

    mSize = sat_peek(src+5, 5);
    eSize = sat_peek(src+5+mSize, 5);
    return 5 + mSize + eSize;
 }

int LNGREALnibbles(SatAddr src)
{
    if (!isLNGREAL(src)) return -1;
    return lngrealNibbles(src);
}



// Get the number of digits in the mantissa and exponent.  If mdigits
// is non-null, it will be populated with the number of digits in the
//  mantissa.  Edigits is the same, but uses the exponent.
bool lngrealDigits(SatAddr src, int *mDigits, int *eDigits)
{
    if (mDigits) {
	*mDigits = zintDigits(src);
    }
    if (eDigits) {
	*eDigits = zintDigits(src+zintNibbles(src)-5);
    }
    return true;
 }

bool LNGREALdigits(SatAddr src, int *mDigits, int *eDigits)
{
    if (!isLNGREAL(src)) return false;
    return lngrealDigits(src, mDigits, eDigits);
}

// Implemtation for LNGREALbytes(). Also used for LNGREALCOMP
bool lngrealBytes(SatAddr src, int *mBytes, int *eBytes)
{
    if (mBytes) {
	*mBytes = zintBytes(src);
    }
    if (eBytes) {
	*eBytes = zintBytes(src+zintNibbles(src)-5);
    }
    return true;
 }

// Get the number of bytes required to store a LNGREAL's mantissa and
// exponent as strings.  If mBytes is non-null it will be populated
// with the required bytes for teh mantissa. If eBytes is non-null, it
// will be populated with the required bytes for the exponent.
// Returns false if src isn't a LNGREAL.
bool LNGREALbytes(SatAddr src, int *mBytes, int *eBytes)
{
    if (!isLNGREAL(src)) return false;
    return lngrealBytes(src, mBytes, eBytes);
}


