#include <stdarg.h>
#include <hpgcc49.h>
#include "hpobjects.h"
#include "../array/array.h"
//
// Linked Array routines
// N-dimensional arrays of objects of one type.  These contain
// pointers to each element for quick access.
// Note that in the C routines, array indexes start at zero.
//
// Functions follow the same semantics as for ARRAY
//

// Linked arrays are described in section 57.12 of "Introduction to Saturn
// Assembly Language" by Fernandes and Rechlin, but that description
// appears incorrect. Comparing to the RPL Manual, the "size of objects" field
// is really the Object's Prolog.  So the layout would seem to be:
// Prolog (02A0A)        5 nibbles
// Size (minus prolog)   5 nibbles
// Element Prolog        5 nibbles
// Number of dimensions  5 nibbles
// Dimension 1           5 nibbles
// ...
// Dimension n           5 nibbles
// Offset to 1st object  5 nibbles
// Offset to 2nd object  5 nibbles
// ...
// Offset to last object 5 nibbles
// Body of obj1          variable size
// Body of obj2          variable size
// ...
// Body of last obj      variable size

// There is no guarantee that the number of objects match the total
// dimensions in the array, or the order.  IF the same obj appears
// twice in the array, it might have a single copy in the contents
// part.
//

//
// Since the objects can be of any type, the items can be variable
// length.

int LNKARRYgetElementBodyv(SatAddr array,
			   SatAddr *elementBody,
			   int idx1,
			   va_list argp);
//++

// Is this a linked array?
bool isLNKARRY(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    return (prolog == SAT_DOLNKARRY);
}


// Create an N-dimensional linked array with the dimensions given.
// Initialize each item in the array with a copy of "element"
// Returns the array. Note that this puts separate copies of
// "element" for each item
SatAddr makeLNKARRY(SatAddr element,
		  int numDimensions,
		  int dim1,
		  ... /* dim2,dim3 etc */ )
{
    va_list argp;
    SatAddr result;		/* the array */
    SatAddr addr;		/* current address in the array */
    SatAddr addr2;		/* another addr */
    int nibbles;		/* size of the array */
    int *dimensions;		/* array of dimensions */
    int numElements=1;		/* total number of elements */
    int elementSize;		/* size of element */
    int i;

    // Get the dimensions from the arg list
    va_start(argp, dim1);
    dimensions = commonGetDimensions(numDimensions, dim1, argp);
    va_end(argp);
    if (dimensions==NULL) return 0;

    // compute the number of elements
    for (i = 0; i < numDimensions; ++i) {
	numElements *= dimensions[i];
    }

    // Compute the size of the element, which is the size of the
    // object, minus its prolog
    elementSize = obNibbles(element) - 5;
    if (elementSize<0) return 0;

    // Compute the size of the array
    nibbles = 5;		/* prolog */
    nibbles += 5+5+5;		/* size, item prolog, # dimensions */
    nibbles += 5 * numDimensions; /* dimension table */
    nibbles += 5 * numElements;	/* pointer table */
    nibbles += numElements * elementSize; /* elements */

    // Now make the array
    result = sat_createtemp(nibbles);
    sat_poke(result, SAT_DOLNKARRY, 5);		   /* prolog */
    sat_poke(result+5, nibbles-5, 5);		   /* size */
    sat_poke(result+10, sat_peek(element, 5), 5);  /* kind */
    sat_poke(result+15, numDimensions, 5);	  /* num dimensions */

    addr = result+20;
    for (i = 0; i < numDimensions; ++i) {	  /* each dimension */
	sat_poke(addr, dimensions[i], 5);
	addr += 5;
    }
    
    // Now do the pointer table and content table at the same
    // time.  addr is a pointer table entry.  Addr2 is a content table
    // entry
    addr2 = addr + 5*numElements;	          /* addr of first obj */
    for (i = 0; i < numElements; ++i) {	          /* N copies of element */
	sat_poke(addr, addr2-addr, 5);
	sat_moveup(element+5, addr2, elementSize);
	addr +=5;
	addr2 += elementSize;
    }

    free(dimensions);
    return result;
}


// Worker function to find an element body from it's index numbers
// Returns zero on success.
// -10-n means the n'th index value is out of range.
int LNKARRYgetElementBodyv(SatAddr array,
			   SatAddr *elementBody,
			   int idx1,
			   va_list argp)
{
    int i;
    int *indices;
    int n=0;				  /* element number */
    int blockSize=1;
    int numDimensions = LNKARRYnumDimensions(array);
    SatAddr addr;			  /* address of the element */

    if (!isLNKARRY(array)) return -3;

    // Get the indices from the arg list
    indices = commonGetDimensions(numDimensions, idx1, argp);
    if (indices==NULL) return -4;

    // Flatten out the indices to get the linear element number.
    for (i=numDimensions-1; i >= 0; --i) {
	int dim = LNKARRYgetDimension(array, i);
	if (indices[i] < 0 || indices[i] >= dim) {
	    free(indices);
	    return 10-i-1;
	}
	n += blockSize * indices[i];
	blockSize *= dim;
    }

    // Now get the n'th element.

    addr = array + 5*4;	   /* prolog, size, kind, #dims */
    addr += 5 * numDimensions;	/* skip dimensions */
    addr += n*5;		/* N'th element in table */
    addr += sat_peek(addr,5);	/* N'th element in array */
    free(indices);
    *elementBody = addr;
    return 0;
}

// Get a pointer to the body of an element in the array.
// Returns zero on success.
// -10-n means the n'th index value is out of range.
int LNKARRYgetElementBody(SatAddr array, SatAddr *elementBody,
			int idx1,
			... /* idx2, idx3, etc */ )
{
    va_list argp;
    int result;

    va_start(argp, idx1);
    result = LNKARRYgetElementBodyv(array, elementBody, idx1, argp);
    va_end(argp);
    return result;
}



// Given a prolog and a pointer to an item in an array, return the
// size of the item in the array.  This works by temporarily changing
// the 5 nibbles before the item body to the prolog and then calling
// obNibbles().  This only works in a single threaded environment.
// This is defined in array.c
int LNKARRYelementSize(SatAddr array, SatAddr addr)
{
    if (!isLNKARRY(array)) return -1;
    return commonElementSize(array, addr);
}



// Put an item into a linked array.  The item must be the same type
// as the array contents.  Also, it must have the same length as the item
// currently at the position.  Finally, you must specify the right number of
// indexes.
//
// Returns zero on success.
// -1 means item is the wrong type
// -2 means the item isn't the same size as the one there.
// -3 means array isn't an array.
// -10-n means the n'th index value is out of range.
int LNKARRYputItem(SatAddr array,
		   SatAddr item,
		   int idx1,
		   ... /* idx2, idx3, ... */ )
{
    int result;
    SatAddr addr;
    int newSize;
    int existingSize;

    va_list argp;

    if (!isLNKARRY(array)) return -3;

    if (sat_peek(item, 5) != LNKARRYgetElProlog(array)) {
	return -1;				   /* wrong element type */
    }

    va_start(argp, idx1);
    result = LNKARRYgetElementBodyv(array, &addr, idx1, argp);
    va_end(argp);
    if (result < 0) {
	return result;				  /* bad index */
    }

    newSize = obNibbles(item);
    existingSize = LNKARRYelementSize(array, addr);
    if (newSize != existingSize+5) {
	return -2;
    }
    sat_moveup(item+5, addr, existingSize);
    return 0;    
}


// Returns pointer to the first element body
SatAddr LNKARRYfirstElement(SatAddr array)
{
    SatAddr result;
    if (!isLNKARRY(array)) return 0;
    result = array + 5*4;	   /* prolog, size, kind, #dims */
    result += 5 * LNKARRYnumDimensions(array);
    result += sat_peek(result,5);
    return result;
}


// Given an array and point to the body of an item in the array,
// return a pointer to the next body in the array, or zero if no
// more elements exist.
SatAddr LNKARRYnextElement(SatAddr array, SatAddr item)
{
    SatAddr result;
    if (!isLNKARRY(array)) return 0;
    result = item + LNKARRYelementSize(array, item);
    
    if (result >= skipOb(array)) result = 0;
    return result;
}


// Size of the Array object
int LNKARRYnibbles(SatAddr src)
{
    if (!isLNKARRY(src)) return -1;
    return sat_peek(src+5, 5) + 5;
}


 /* get prolog of elements in array */
int LNKARRYgetElProlog(SatAddr array)
{
    int result;
    if (!isLNKARRY(array)) return -1;
    result = sat_peek(array+10, 5);
    return result;
}

int LNKARRYnumDimensions(SatAddr array)
{
    int result;
    if (!isLNKARRY(array)) return -1;
    result = sat_peek(array+15, 5);
    return result;
}

// Get the size of the n'th dimension (n=1,2,3,...)
int LNKARRYgetDimension(SatAddr array, int dimensionNumber)
{
    if (!isLNKARRY(array)) return -1;
    return commonGetDimension(array, dimensionNumber);
}



// Given an array and the body of an item within the array, make an
// object for the item at dst.  This just copies the ARRAYelProlog()
// to dst, and copies the array item body to dst+5.  The body's size
// is LNKARRYelementSize()
bool LNKARRYmakeItem(SatAddr array, SatAddr itemBody, SatAddr dst)
{
    if (!isLNKARRY(array)) return false;
    return commonMakeItem(array, itemBody, dst);
}


SatAddr LNKARRYfirstItem(SatAddr lnkarry, SatAddr dst, SatAddr *iter)
{
    *iter = LNKARRYfirstElement(lnkarry);
    if (*iter == 0) return 0;
    return commonMakeItem(lnkarry, *iter, dst);
}

SatAddr LNKARRYnextItem(SatAddr lnkarry, SatAddr dst, SatAddr *iter)
{
    *iter = LNKARRYnextElement(lnkarry, *iter);
    if (*iter == 0) return 0;
    return commonMakeItem(lnkarry, *iter, dst);
}

SatAddr LNKARRYgetItem(SatAddr lnkarry, SatAddr dst, int idx1,
		     ... /* idx2, idx3 etc */ )
{
    SatAddr body;
    int i;
    va_list argp;

    if (!isLNKARRY(lnkarry)) return 0;
    va_start(argp, idx1);
    i = LNKARRYgetElementBodyv(lnkarry, &body, idx1, argp);
    va_end(argp);

    if (i!=0) return 0;
    return commonMakeItem(lnkarry, body, dst);
}
