#include <hpgcc49.h>
#include "hpobjects.h"

//////////////////////////////////////////////////////////////////////
//                Symbolic MATRIX
//////////////////////////////////////////////////////////////////////

// A symbolic matrix is like a list, but with a different prolog and
// restrictions on what objects can go into it.

//++

// Is it a matrix?
bool isMATRIX(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    bool result = (prolog == SAT_DOMATRIX);
    return result;
}


// Put a matrix at dst with a copy of obj.
void MATRIXinit(SatAddr obj, SatAddr dst)
{
    int size = obNibbles(obj);

    sat_poke(dst, SAT_DOMATRIX, 5);
    sat_moveup(obj, dst+5, size);
    sat_poke(dst+5+size, SAT_SEMI, 5);
}

// Allocate space for a MATRIX with NIBBLES data in it, and put a copy
// of "obj" into it. Return the address of the MATRIX. You can add
// data to the matrix with MATRIXadd(), as long as the total size of
// the items added doesn't exceed NIBBLES.  Note that if obj is a pointer
// to an object, then the pointer will be copied into the matrix
SatAddr makeMATRIX(SatAddr obj, unsigned nibbles)
{
    SatAddr result;
    result = sat_createtemp(nibbles);
    MATRIXinit(obj, result);
    return result;
}



// Returns size of the MATRIX in nibbles.
int MATRIXnibbles(SatAddr src)
{
    if (!isMATRIX(src)) return -1;
    return compositeObNibbles(src);
}


// Returns the number of elements in the matrix at src
int MATRIXsize(SatAddr src)
{
    if (!isMATRIX(src)) return -1;
    return compositeSize(src);
}


/* Make a copy of obj and append it to the matrix. This assumes that
   there is space in the matrix for the object. */
bool MATRIXadd(SatAddr dst, SatAddr obj)
{
    if (!isMATRIX(dst)) return false;
    return compositeAdd(dst, obj);
}


/* Concatenate an object to a matrix. This creates a new matrix
   consisting of the old one and the new object. Returns the address
   of the new matrix */
SatAddr MATRIXcat(SatAddr src, SatAddr obj)
{
    if (!isMATRIX(src)) return 0;
    return compositeCat(src, obj);
}


/* Merge two matrixs into one.  This creates a new matrix consisting of
   all the items in MATRIX1 followed by all the items in MATRIX2. For
   example, if matrix1 is {1 2 3} and matrix2 is {4 5 6} then
   MATRIXmerge(matrix1, matrix2) creates {1 2 3 4 5 6} */
SatAddr MATRIXmerge(SatAddr matrix1, SatAddr matrix2)
{
    return compositeMerge(matrix1, matrix2);
}

// Return the first object in a matrix, or 0 of the matrix is nil
SatAddr MATRIXfirstEntry(SatAddr matrix)
{
    if (!isMATRIX(matrix)) return 0;
    return compositeFirstEntry(matrix);
}


// Given an entry from a matrix returned by MATRIXfirstEntry or
// MATRIXnextEntry, return the next entry in the matrix.
SatAddr MATRIXnextEntry(SatAddr entry)
{
    return compositeNextEntry(entry);
}

// Return a pointer to the first object in the matrix, or zero if "matrix"
// isn't a matrix or if the matrix is empty.  "Iter" is the address of a
// scratch pointer used by MATRIXfirstOb/MATRIXnextOb to keep track of the
// position in the matrix. Do not modify iter between calls.  This
// dereferences indirect pointers in the matrix the ensure that it
// returns a pointer to an object
SatAddr MATRIXfirstOb(SatAddr matrix, SatAddr *iter)
{
    if (!isMATRIX(matrix)) return 0;
    return compositeFirstOb(matrix, iter);
}


// Given an iter that was previously passed to MATRIXfirstOb or
// MATRIXnextOb, return the address of the next object in the matrix and
// update iter.  This dereferences indirect pointers in the matrix the
// ensure that it returns a pointer to an object
SatAddr MATRIXnextOb(SatAddr *iter)
{
    return compositeNextOb(iter);
}



// Make a new matrix consisting of the objects ob1, ob2, ... obn and put
// them in dst. If dst is zero the space is allocated for the new
// matrix. Otherwise dst must contain sufficient space. Return the new matrix
SatAddr MATRIXencodeV(SatAddr dst, ... /* ob1, ob2, ob3, ..., obn , 0 */ )
{
    va_list argp;
    va_start(argp,dst);
    dst = compositeEncode(dst, argp);
    sat_poke(dst, SAT_DOMATRIX, 5);
    va_end(argp);
    return dst;
}

// Make a new matrix consisting of the objects obs[0], obs[1], obs[2],
// ... obs[n-1] and put them in dst. If dst is zero then space is
// allocated for the new matrix. Otherwise dst must contain sufficient
// space. Return the new matrix. Basically this is like ->MATRIX on the
// calculator
SatAddr MATRIXencodeN(int n, SatAddr obs[], SatAddr dst)
{
    dst = compositeEncodeN(n,obs,dst);
    if (dst) sat_poke(dst, SAT_DOMATRIX, 5);
    return dst;    
}


