#include <hpgcc49.h>
#include "hpobjects.h"

///////////////////////////////////////////////////////////
//   Miscellaneous functions for HPObjects library
///////////////////////////////////////////////////////////

struct Table {
    SatAddr prolog;
    char *name;
};
extern struct Table prologTab[];

//++
// This file contains functions to convert native HP objects to/from
// ARM equivalents
///
// Extract the base-10 exponent and mantissa from a double.
// Store the exponent in ex.  1 <= abs(mantissa) < 10.
void
doubledecode(double number, int *ex, double *m)
{
    int exponent;
    double mantissa;

    if (abs(number)<=1E-505) {
	// It's too small. Make it zero
	mantissa = exponent = 0;
    } else {  //decent sized number
	/////////////
	//Following fix by Wesley Loewer
 
	double lognumber = log10(abs(number));
	if (abs(number) >= 1.0)
	    exponent = (int)lognumber;   //get exponent - its positive, so round down
	else if (lognumber == floor(lognumber)) {   // is the log an exact integer?
	    exponent = (int)lognumber;          // if so, just get exponent
	} else {
	    exponent = (int)lognumber - 1;      // if not, it's 1 less
	}
	//////////////

	//now extract the mantissa
	
	if(exponent>=0) {
	    number /= ipow(10.0, exponent);
	} else { //negative exponent
	    number *= ipow(10.0, abs(exponent));
	}
	mantissa = number;
    }
    *ex = exponent;
    *m = mantissa;
}

// This is like doubledecode() but it results in long long mantissa
// and exponent.  Although I could have made the exponent an int, or
// even a short, I'm using a long long so the programmer doesn't have
// to remember the detail of what pointer is what type

void
doubledecodell(double number, long long *mantissa, long long *exponent)
{
    double m;
    int e;
    // This extracts the mantissa and exponent as integer values
    // and then calls LNGREALencode()

    doubledecode(number, &e, &m);

    // doubledecode returns 1 <= abs(m) < 10.  We want an integer
    // mantissa, so adjust it.  A double has about 15 digits of
    // mantissa so we'll start there.
    m *= 1.0E15;
    if (m>0) {			/* round off */
	m += 0.5;
    } else {
	m -= 0.5;
    }
    *mantissa = m;
    if (*mantissa) {		/* no sense getting weird exponents
				   with zero */
	e -= 15;		/* exponent and mantissa now match */

	// adjust for trailing zeros.  e.g., convert 123000E6 to 123E9
	while (*mantissa%10 == 0) {
	    *mantissa /= 10;
	    ++e;
	}
    }
    *exponent = e;
}


//
// Convert a C double to an HP 64 bit real number. Note that this
// does NOT convert it to the 21 nibble DOREAL object, it just converts
// the binary formats.  This is the inverse of sat_real2double(),
// except it always writes to the Saturn address space
void double2hp(double number, SatAddr dst)
{
    unsigned long long real = sat_double2real(number);

    sat_poke(dst,(unsigned int)real,8);
    sat_poke(dst+8,(unsigned int)(real>>32),8);
}




// Are objects with this prolog fixed sized?
bool fixedSize(int prolog)
{
    switch (prolog) {
    case SAT_DOBINT:
    case SAT_DOCHAR:
    case SAT_DOREAL:
    case SAT_DOCSTR:
    case SAT_DOLIST:
    case SAT_DOCMP:
    case SAT_DOXLIB:
    case SAT_DOEREL:
    case SAT_DOMINIFONT:
    case SAT_DOEXT2:		/* FONT */
    case SAT_DOACPTR:
    case SAT_DOFLASHP:
	return true;

    case SAT_DOIDNT:
    case SAT_DOLAM:
    case SAT_DOHXS:
    case SAT_DOGROB:
    case SAT_DOCODE:
    case SAT_DOLIB:
    case SAT_DOINT:
    case SAT_DOARRY:
    case SAT_DORRP:
    case SAT_DOLNGREAL:
    default:
	return false;
    }
}


// Return the address of the object after src.  Src can point directly
// to an object, or it can be a pointer to an object.
// Return the size in nibbles of object or pointer at src.
// TODO: this and all the XYZnibble() routines should really return 0 on
// error. That way they'd still work if int and address size was the same.
int obNibbles(SatAddr src)
{
    SatAddr p1 = sat_peek_sat_addr(src);
    SatAddr prolog = sat_peek_sat_addr(p1);

    if (prolog != SAT_PRLG) {
	// src is an indirect pointer.  Skip the pointer
	return 5;
    }

    switch (p1) {
    case SAT_DOBINT:
	return 10;
    case SAT_DOCHAR:
	return 7;
    case SAT_DOREAL:
	return 5+3+12+1;
    case SAT_DOCSTR:
	return STRINGnibbles(src);
    case SAT_DOLIST:
	return LISTnibbles(src);
    case SAT_DOCOL:
    case SAT_DOSYMB:		/* symb is the same as col */
    case SAT_DOUNIT:		/* UNIT is the same as COL */
	return compositeObNibbles(src);
    case SAT_DOCMP:
	return 5+16+16;
    case SAT_DOIDNT:
	return IDNTnibbles(src);
    case SAT_DOLAM:
	return LAMnibbles(src);
    case SAT_DOHXS:
	return HXSnibbles(src);
    case SAT_DOGROB:
	return GROBnibbles(src);
    case SAT_DOXLIB:
	return 5+3+3;
    case SAT_DOEREL:
	return 5+5+15+1;
    case SAT_DOCODE:
    case SAT_DOINT:
    case SAT_DOMINIFONT:
    case SAT_DOEXT2:		/* FONT */
    case SAT_DOARRY:
    case SAT_DOBAK:
    case SAT_DOLIB:
	// These all begin with a 5 nibble "size" field that's
	// the size of the object minus the prolog
	return 5 + sat_peek(src+5, 5);
    case SAT_DOACPTR:
	return 5+5+5;
    case SAT_DOTAG:
	return TAGnibbles(src);
    case SAT_DORRP:
	return DIRnibbles(src);
    case SAT_DOECMP:
	return 5+21+21;
    case SAT_DOLNGREAL:
	return LNGREALnibbles(src);
    case SAT_DOLNGCOMP:
	return LNGCOMPnibbles(src);
    case SAT_DOMATRIX:
	return MATRIXnibbles(src);
    case SAT_DOLNKARRY:
	return LNKARRYnibbles(src);
    case SAT_DOFLASHP:
	return 5+3+4;
    default:
	;
    }
    return -1;
}


// Return the address of the object after src.  Src can point directly
// to an object, or it can be a pointer to an object.
SatAddr skipOb(SatAddr src)
{
    int size = obNibbles(src);
    if (size>= 0) return src+size;
    else return 0;
}


// return the number of digits in a long long.  Note that
// llDigits(0)=0  to be consistent with ZINT and the objects derived
// from it (LNGREAL, LNGCOMP)
int llDigits(unsigned long long src)
{
  int result=0;

  while (src) {
    ++result;
    src /= 10;
  }
  return result;
}

// Convert a long long to BCD, one digit per nibble, LSB stored first.
// "src" is the input number. "dst" is where we want to store it.
// numDigits" is the number of digits to print (-1 means as many as
// required).  Note that if src==0 and numDigits==-1, this will
// output NO DIGITS at all, but that's the behavior we usually want
// because zero is stored with no digits in a ZINT
int ll2bcd(unsigned long long src, SatAddr dst, int numDigits)
{
    int digit;
    int digitsUsed = 0;
    while ((numDigits < 0 && src) || numDigits-- > 0) {
	digit = src%10;
	sat_poke(dst++, digit, 1);
	src /= 10;
	++digitsUsed;
    }
    return digitsUsed;
}


// Convert a null terminated string of digits BCD, one digit per
// nibble, LSB stored first.  "src" is the input number. "dst" is
// where we want to store it.  numDigits" is the number of digits to
// print (-1 means as many as required).
int str2bcd(const char *src, SatAddr dst, int numDigits)
{
    int digitsUsed = 0;
    int i;
    for (i = strlen(src); --i>=0 && numDigits-- > 0;) {
	sat_poke(dst++, src[i] - '0', 1);
	++digitsUsed;
    }
    while(numDigits-- > 0) {
	sat_poke(dst++, 0, 1);
	++digitsUsed;
    }
	
    return digitsUsed;
}


// Overwrite the source object with a copy of the destination
// object, if they are the same size.  Returns true/false depending
// on whether it overwrote the object.  If the source and dest are
// pointers to objects, then it overwrites the source pointer with
// the destination pointer.  CAUTION: If anything in the calculator
// points to the source object (like a value on the stack), then you'll
// change that also!
bool obCopy(SatAddr src, SatAddr dst)
{
    int srcSize = obNibbles(src);
    int dstSize = obNibbles(dst);
    if (srcSize != dstSize || srcSize<0) {
	return false;
    }
    sat_moveup(src, dst, srcSize);
    return true;
}

// Is there a recognized object at src?  This is a quick&dirty check.
// All prologs point to PRLG
bool isOb(SatAddr src)
{
    SatAddr prolog = obProlog(src);
    bool result = sat_peek(prolog,5) == SAT_PRLG;
    return result;
}



// Absolute value of a long long
long long llabs(long long src)
{
    return (src < 0 ? -src : src);
}


// Given an address that contains an object or a pointer to an object,
// return the pointer to the object itself.
SatAddr directPtr(SatAddr ob)
{
    // RPLMAN.txt says that the first thing at a prolog address is
    // a pointer to the prolog itself. This lets you do direct or indirect
    // execution of objects. But it looks like the 49 series uses
    // a common prolog (PRLG) as the first thing in the prologs
    SatAddr a1 = sat_peek(ob,5);
    SatAddr prolog = sat_peek(a1,5);
    // return (a1 == prolog ? ob : a1);
    return (prolog == SAT_PRLG ? ob : a1);
}

struct Table prologTab[] = {
    { SAT_DOINT, "DOINT" },
    { SAT_DOCSTR, "DOCSTR" },
    { SAT_DOREAL, "DOREAL" },
    { SAT_DOEREL, "DOEREL" },
    { SAT_DOHXS, "DOHXS" },
    { SAT_DOBINT, "DOBINT" },
    { SAT_DOCMP, "DOCMP" },
    { SAT_DOECMP, "DOECMP" },
    { SAT_DOCHAR, "DOCHAR" },
    { SAT_DOARRY, "DOARRY" },
    { SAT_DOLNKARRY, "DOLNKARRY" },
    { SAT_DOIDNT, "DOIDNT" },
    { SAT_DOLAM, "DOLAM" },
    { SAT_DOROMP, "DOROMP" },
    { SAT_DORRP, "DORRP" },
    { SAT_DOSYMB, "DOSYMB" },
    { SAT_DOLIST, "DOLIST" },
    { SAT_DOCOL, "DOCOL" },
    { SAT_DOGROB, "DOGROB" },
    { SAT_DOMINIFONT, "DOMINIFONT" },
    { SAT_DOBAK, "DOBAK" },
    { SAT_DOCODE, "DOCODE" },
    { SAT_DOEXT, "DOEXT" },
    { SAT_DOFLASHP, "DOFLASHP" },
    { SAT_DOLIB, "DOLIB" },
    { SAT_DOMATRIX, "DOMATRIX" },
    { SAT_DOTAG, "DOTAG" },
    { SAT_DOAPLET, "DOAPLET" },
    { SAT_DOEXT0, "DOEXT0" },
    { SAT_DOACPTR, "DOACPTR" },
    { SAT_DOEXT1, "DOEXT1" },
    { SAT_DOEXT2, "DOEXT2" },
    { SAT_DOEXT3, "DOEXT3" },
    { SAT_DOEXT4, "DOEXT4" },
    { SAT_DOLNGREAL, "DOLNGREAL" },
    { SAT_DOLNGCOMP, "DOLNGCOMP" },
    { 0,0}
};

char *prolog2Name(SatAddr prolog)
{
    struct Table *p;
    for (p = prologTab; p->name; ++p) {
	if (p->prolog == prolog) {
	    return p->name;
	}
    }
    return "unknown";
}

// Including string.h causes problems when compiling on the PC, so I'll just
// declare this here.
int strcasecmp(const char *s1, const char *s2)
{
    char c1, c2;
    while ((c1 = *s1)) {
	c2 = *s2;
	if (c2 == 0) {
	    return 1;
	} else if (c1 != c2) {
	    if (isalpha(c1)) c1 = tolower(c1);
	    if (isalpha(c2)) c2 = tolower(c2);
	    if (c1 < c2) {
		return -1;
	    } else if (c1 > c2) {
		return 1;
	    }
	}
	++s1;
	++s2;
    }
    if ((c2 = *s2) == 0) {
	return 0;
    }
    return -1;
}


// Return the sat addr for the named prolog, or 0 if none.  The name
// can be case insensitive and may optionally skip the "DO" prefix.
SatAddr name2Prolog(const char *name)
{
    extern int strcasecmp(const char *s1, const char *s2);
    struct Table *p;
    for (p = prologTab; p->name; ++p) {
	if (strcasecmp(p->name, name) == 0 ||
	    strcasecmp(p->name+2, name) == 0) {
	    return p->prolog;
	}
    }
    return 0;
}

// Like memset(), but works in the saturn domain and sets a specified
// number of nibbles, not bytes
void sat_memset(SatAddr dst, int val, int nibbles)
{
    while (nibbles--) {
	sat_poke(dst++, val, 1);
    }
}


// In most places where you can store an object, you can also store
// a pointer.  So when do you store which?  These functions deal with
// this problem.

// return true if the object should be stored as a pointer.
// Currently, anything at an address below 0x40000 will be stored as a
// pointer because 00000-40000 is ROM.
bool obStorePtr(SatAddr obj)
{
    return (obj < 0x40000);
}


// Given an object, return the number of nibbles required to store it
// as an object or pointer.
int objOrPtrNibbles(SatAddr obj)
{
    if (obStorePtr(obj)) {
	return 5;
    } else {
	return obNibbles(obj);
    }
}

// Copy src, or a pointer to src, to dst. Returns the number of
// nibbles copied or -1 on error
int copyObjOrPtr(SatAddr src, SatAddr dst)
{
    int size = objOrPtrNibbles(src);
    if (size == 5) {
	// Must be a pointer
	sat_poke(dst, src, 5);
    } else if (size > 0) {	
	// copy object
	sat_moveup(src, dst, size);
    }
    return size;
}
