#include <hpgcc49.h>
#include "hpobjects.h"

//
// Routines that deal with HP STRING
//

// A STRING is:
// prolog (5 nibbles)
// size   (5 nibbles)
// 2 nibbles per character

//++

// Is the "src" address a STRING
bool isSTRING(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    bool result = (prolog == SAT_DOCSTR);
    return result;
}

// Convert from an HP STRING pointed at by "src" into a null-terminated
// string located in the buffer pointed to by dst.  Dst must have
// enough space for the string and the null terminated buffer.
char *STRINGdecode(SatAddr src, char *dst)
{
    int chars;
    if (!isSTRING(src)) return NULL;
    chars = STRINGnumChars(src);
    if (dst == 0) {
	dst = malloc(chars+1);
    }
    sat_peek_sat_bytes(dst, src+10, chars);
    dst[chars] = 0;
    return dst;
}

// Convert from a NULL-terminated string to an HP STRING.
// If "dst" is zero then a new STRING will be created, otherwise
// "dst" must have enough room to hold the string. This writes
// all data for the object, including the prolog and size fields.
SatAddr STRINGencode(const char *src, SatAddr dst)
{
    int bytes = strlen(src);
    if (dst == 0) {
	dst = makeSTRING(bytes);
    }
    sat_poke(dst, SAT_DOCSTR, 5);	       /* prolog */
    sat_poke(dst+5, bytes*2+5, 5);	       /* size */
    sat_poke_nibbles(dst+10, src, bytes*2);    /* data */
    return dst;
}


// Add the null-terminated string at "src" to the STRING at "dst".
// There must be enough space allocated at dst for the combined
// strings.  CAUTION: The string must eventually be exactly the
// length that you allocated or it must be built in a variable
// length buffer (such as within another string) and then copied out
bool STRINGadd(const char *src, SatAddr dst)
{
    SatAddr target;
    int bytes;

    if (!isSTRING(dst)) return false;
    // advance dst past the prolog, size, and data
    target = dst + STRINGnibbles(dst);
    bytes = strlen(src);
    sat_poke_nibbles(target, src, bytes*2);
    sat_poke(dst+5, sat_peek(dst+5, 5) + bytes * 2, 5);
    return true;
}


/* Allocate space for a STRING with numChars characters, and put an empty
   STRING at the beginning of the space. Return the address of the
   STRING. You can add data to the string with STRINGadd(), as long as the
   total number of characters added doesn't exceed numChars. */
SatAddr makeSTRING(unsigned numChars)
{
    int nibbles = numChars * 2 + 10;
    SatAddr result = sat_createtemp(nibbles);
    sat_poke(result, SAT_DOCSTR, 5); /* prolog */
    sat_poke(result+5, nibbles-5, 5); /* size. Must take up all space */
    return result;
}


// Return the size in nibbles of the STRING object at src.  Returns -1
// if the object isn't a string
int STRINGnibbles(SatAddr src)
{
    int result;

    if (!isSTRING(src)) {
	return -1;
    }
    // The "size" field in the STRING does not include the prolog itself.
    result = 5 + sat_peek_sat_addr(src+5);
    return result;
}


// Return the number of characters in the STRING at SatAddr, or -1
// src doesn't point to a STRING
int STRINGnumChars(SatAddr src)
{
    int nibbles;

    if (!isSTRING(src)) {
	return -1;
    }
    nibbles = STRINGnibbles(src) - 10;
    return nibbles/2;
}
