#include <hpgcc49.h>
#include "hpobjects.h"

// ZINT routines - Infinite precision integer
//
// A ZINT is:
// prolog (02614) 5 nibbles
// Number of digits 5 nibbles
// digit 1 (1 nibble, LSB)
// ...
// digit n (1 nibble, MSB)
// sign: 0/9 for +/- (1 nibble)
// 
// Note that zero has no sign nibble (or no digit nibbles, depending on how
// you look at it.  Since client code will likely want to know how many
// digits are needed to represent the number, the ZINTdigits() function
// returns 1 when the number encodes zero.

int zintDigits(SatAddr src);
int zintNibbles(SatAddr src);

//++

bool isZINT(SatAddr src)
{
    int prolog = sat_peek_sat_addr(src);
    return (prolog == SAT_DOINT);
}

// ZINT2ll without parameter checking. This is used by other objects.
bool zintDecodell(SatAddr src, long long *dst)
{
    int numDigits = sat_peek(src+5, 5) - 6;
    int sgn = sat_peek(src+10+numDigits, 1);
    long long val = 0;
    int i;

    for (i = numDigits-1; i >=0; --i) {
	int digit = sat_peek(src+10+i, 1);
	long long newval = val * 10 + digit;
	if (newval < 0 || (newval-digit)/10 != val) return false; /* overflow */
	val = newval;
    }
    if (sgn) {
	val = -val;
    }
    *dst = val;
    return true;
 }

bool ZINTdecodell(SatAddr src, long long *dst)
{
    if(!isZINT(src)) return false;
    return zintDecodell(src, dst);
}


SatAddr ZINTencodell(long long src, SatAddr dst)
{
    int sgn;
    int digits;

    if (src < 0) {
	sgn = 9;
	src = -src;
    } else {
	sgn = 0;
    }

    // sgn is set.  src is positive
    if (dst == 0) {
	digits = llDigits(src);
	dst = makeZINT(digits);
    }
    sat_poke(dst, SAT_DOINT, 5);	/* prolog */
    digits = ll2bcd(src, dst+10, -1);	/* digits */
    sat_poke(dst+5, digits+5+1, 5);     /* size */
    sat_poke(dst+10+digits, sgn, 1);	/* sign */
    return dst;
}

// Clean up a string representation of an integer.  This skips +/-
// signs and leading zeros, and figures out how many digits are
// present.  "src" is the source string.  "digits" is an output param
// that is set to the number of digits in the string (which might be
// zero) the scan stops at the first non-digit character.  isPostive
// will be set to true/false depending on whether the number is
// positive or negative.  Returns a pointer to the first non-zero
// digit of the number.
//
// The following examples show the result of
// int digits;
// bool isPositive
// const char *result = cleanUpZINTstr(input, &digits, &isPositive);
//
// Examples:
// input               digits       isPositive        return value
// "1234"              4            1                 "1234"
// "+1234"             4            1                 "1234"
// "-1234"             4            1                 "1234"
// "---+1234ab"        4            0                 "1234"
// "-0"                0            1                 ""
const char *cleanupZINTstr(const char *src, int *digits, bool *isPositive)
{
    bool done = false;
    int i;
    *isPositive = true;

    // scan through +/-/0
    while (!done) {
	switch (*src) {
	case '-':
	    *isPositive = ! *isPositive;
	    // FALL THROUGH
	case '+':
	case '0':
	    ++src;
	    break;
	default:
	    done = true;
	}
    }

    for (i=0; isdigit(src[i]); ++i)
	;
    if (i ==0) {
	*isPositive = true;	/* "-0" is positive */
    }
    *digits = i;
    return src;
}

// Convert a null-terminated ascii string to a ZINT.  The string
// contains the digits, as ascii (e.g. 123 would be "123").  If dst is
// zero then a new ZINT will be created. Otherwise dst must point to
// enough space to hold the number.  The resulting ZINT is returned.
SatAddr ZINTencodeStr(const char *src, SatAddr dst)
{
    bool isPositive = true;
    int digits;
    int i;
    src = cleanupZINTstr(src, &digits, &isPositive);

    if (dst == 0) {
	dst = makeZINT(digits);
    }
    // src now points to first non-zero character in number.
    // isPositive tells whether it's positive and numDigits
    // tells the number of digits in the number.
    sat_poke(dst, SAT_DOINT, 5);	/* prolog */
    sat_poke(dst+5, digits+5+1, 5);     /* size */
    for (i=0; i<digits; ++i) {
	sat_poke(dst+10+i, src[digits-i-1]-'0', 1);
    }
    sat_poke(dst+10+digits, (isPositive ? 0 : 9), 1);	/* sign */
    return dst;
}


char * zintDecodeStr(SatAddr src, char *str)
{
    int i = 0;
    char *result;
    int numDigits;
    int sgn;

    numDigits = zintDigits(src);
    sgn = sat_peek(src+zintNibbles(src)-1, 1);
    if (str == NULL) {
	str = malloc(numDigits + (sgn ? 2 : 1));
    }
    result = str;
    if (sgn) {
	*str++ = '-';
    }
    for (i = numDigits-1; i >= 0; --i) {
	*str++ = sat_peek(src+10+i, 1) + '0';
    }
    *str = 0;
    return result;
 }

char * ZINTdecodeStr(SatAddr src, char *str)
{
    if (!isZINT(src)) return NULL;
    return zintDecodeStr(src, str);
}

SatAddr makeZINT(digits)
{
    SatAddr result = sat_createtemp(5+5+digits+1);
    sat_poke(result, SAT_DOINT, 5);
    sat_poke(result+5, 5+digits+1, 5);
    sat_poke(result+5+digits, 0, 1);	/* set the sign to something valid */
    return result;
}

int zintNibbles(SatAddr src)
{
    int result = sat_peek(src+5, 5) + 5;
    return result;
 }

// Number of nibbles taken up by the object
int ZINTnibbles(SatAddr src)
{
    if (!isZINT(src)) return -1;
    return zintNibbles(src);
}


int zintDigits(SatAddr src)
{
    int result = sat_peek(src+5,5) - 6;	/* subtract size and sign fields */
    if (result == 0) result = 1; /* zero is encoded without digits */
    return result;
 }


// Number of digits in the zint
int ZINTdigits(SatAddr src)
{
    if (!isZINT(src)) return -1;
    return zintDigits(src);
}


// Implementation for ZINTbytes and similar for other objects. This
// does NOT check the prolog of the source
int zintBytes(SatAddr src)
{
    int size = sat_peek(src+5,5);	    // we'll need the size field
    int result = size-6;		    // each digit
    if (result == 0) result = 1;	    // zero is encoded without digits
    if (sat_peek(src+4+size, 1)) {	    // add 1 if negative
	++result;
    }
    return result+1;			    // 1 more for the terminating null
 }


// Return the number of bytes needed to store the zint as a string.
int ZINTbytes(SatAddr src)
{
    if (!isZINT(src)) return -1;
    return zintBytes(src);
}




